theory Affset imports "~~/src/HOL/Old_Number_Theory/Primes"

begin

definition
  affset :: "[nat, nat] \<Rightarrow> nat set" where
  "affset a b == {k. \<exists> n. k = a * n + b}"
definition 
  setadd :: "[nat set, nat] \<Rightarrow> nat set" where
  "setadd A m == {k. \<exists> n. n \<in> A \<and> k = m + n}"

definition
  setmult :: "[nat set, nat] \<Rightarrow> nat set" where
  "setmult A m == {k. \<exists> n. n \<in> A \<and> k = m * n}"
definition
  symdiff :: "[nat, nat] \<Rightarrow> nat" where
  "symdiff m n == (max m n) - (min m n)"


lemma setadd_image: "setadd A m = (\<lambda> x. x + m) ` A"
by (fastforce simp add: setadd_def)

lemma setadd_empty [simp]: "setadd {} m = {}"
by (simp add: setadd_def)

lemma setadd_union [simp]: "setadd (A \<union> B) m = ((setadd A) m \<union> (setadd B) m)"
by (fastforce simp add: setadd_def)

lemma setadd_inter [simp]: "setadd (A \<inter> B) m = ((setadd A) m \<inter> (setadd B) m)"
by (fastforce simp add: setadd_def)


lemma range_plus: "range (\<lambda>x. (x::nat) + m) = - {0 ..< m}"
apply (simp add: image_def Compl_eq)
apply (rule Collect_cong)
apply arith
done

lemma setadd_UNIV [simp]: "setadd UNIV m = - {0 ..< m}"
apply (simp add: setadd_def Compl_eq)
apply (rule Collect_cong)
apply arith
done

lemma setadd_diff [simp]: "setadd (A - B) m = setadd A m - setadd B m"
by (fastforce simp add: setadd_def)

lemma setadd_Compl_aux: "- {0 ..< m} - setadd A m = setadd (UNIV - A) m"
by (simp add: setadd_image image_set_diff range_plus)


lemma UNIV_diff [simp]: "UNIV - A = - A"
by fast

lemma setadd_Compl [simp]: "setadd (- A) m = - {0 ..< m} - setadd A m"
by (simp add: setadd_Compl_aux)

lemma setadd_UN [simp]: "(setadd (\<Union>c\<in>C. (F c)) m) = (\<Union>c\<in>C. (setadd (F c) m))"
by (fastforce simp add: setadd_def)

lemma setadd_subseteq [simp]:
  "(setadd A m \<subseteq> setadd B m) = (A \<subseteq> B)"
by (fastforce simp add: setadd_def subset_eq)


  (* some special affsets *)

lemma affset_UNIV [simp]: "affset (Suc 0) 0 = UNIV"
by (simp add: affset_def)

lemma affset_const: "affset 0 c = {c}"
by (simp add: affset_def)

lemma affset_setadd_shift: 
  "affset a b = setadd (affset a 0) b"
by (fastforce simp add: affset_def setadd_def)

  (* as rewrite rule; precondition to avoid looping during simplification *)
lemma affset_setadd_shiftRW: 
  "0 < b \<Longrightarrow> affset a b = setadd (affset a 0) b"
by (fastforce simp add: affset_def setadd_def)

lemma setadd_affset: "setadd (affset a b) m = (affset a (b + m))"
by (fastforce simp add: affset_def setadd_def)


lemma setmult_empty [simp]: "setmult {} m = {}"
by (simp add: setmult_def)

lemma setmult_zero [simp]: "setmult A 0 = (if A={} then {} else {0})"
by (auto simp add: setmult_def)

lemma setmult_union [simp]: "setmult (A \<union> B) m = setmult A m \<union> setmult B m"
by (fastforce simp add: setmult_def)

lemma setmult_inter [simp]: "m \<noteq> 0 \<Longrightarrow> setmult (A \<inter> B) m = setmult A m \<inter> setmult B m"
by (fastforce simp add: setmult_def)

lemma setmult_subseteq [simp]:
  "m \<noteq> 0 \<Longrightarrow> (setmult A m \<subseteq> setmult B m) = (A \<subseteq> B)"
by (fastforce simp add: setmult_def subset_eq)


lemma "affset a 0 = setmult UNIV a"
by (fastforce simp add: affset_def setmult_def)

lemma affset_0_b [simp]: "affset 0 b = {b}"
by (fastforce simp add: affset_def)

lemma setmult_affset: " setmult (affset a b) m = (affset (a * m) (b * m))"
by (fastforce simp add: affset_def setmult_def add_mult_distrib2)


lemma affset_Int_setadd:
  "\<lbrakk> b0 \<le> b1 \<rbrakk> \<Longrightarrow>
  ((affset a0 b0) \<inter> (affset a1 b1)) = 
  (setadd ((affset a0 0) \<inter> (affset a1 (b1 - b0))) b0)"
by (simp add: setadd_affset)

lemma affset_Int_common_factor: "m \<noteq> 0 \<Longrightarrow> 
  (affset (a0 * m) (b0 * m)) \<inter> (affset (a1 * m) (b1 * m)) = 
  (setmult ((affset a0 b0) \<inter> (affset a1 b1)) m)"
by (simp add: setmult_affset)


lemma dvd_div_mult [simp]: "(m::nat) dvd a \<Longrightarrow> (a div m * m) = a"
by (simp add: dvd_eq_mod_eq_0 div_mod_equality')

lemma setmult_affset_div [simp]:
  "\<lbrakk> m dvd a; m dvd b\<rbrakk> \<Longrightarrow> setmult (affset (a div m) (b div m)) m = affset a b"
by (simp add: setmult_affset)


lemma affset_Int_no_common_factor:
  "c mod m \<noteq> 0
  \<Longrightarrow> (affset (a0 * m) 0) \<inter> (affset (a1 * m) c) = {}"
apply (erule contrapos_pp)
apply (clarsimp simp add: affset_def Int_def)
apply (rename_tac n0 n1)
apply (subgoal_tac "c = (a0 * n0 - a1 * n1) * m")
apply simp
apply (clarsimp simp add: diff_mult_distrib)
apply (subgoal_tac "a0 * m * n0 = a0 * n0 * m")
apply (simp only:)
apply fastforce
apply (simp (no_asm_use))
done


definition
  affset_solvable :: "[nat, nat, nat] \<Rightarrow> bool" where
  "affset_solvable a0 a1 b1 == \<exists> n0 n1. a0 * n0 = a1 * n1 + b1"

lemma affset_solvable_mod:
  "affset_solvable a0 a1 b1 = (\<exists> n. (a1 * n + b1) mod a0 = 0)"
apply (simp add: affset_solvable_def)
apply (rule iffI)
apply (clarsimp)
apply (drule sym) 
apply (rule_tac x=n1 in exI)
apply simp

apply clarsimp
apply (drule sym)
apply (intro exI)
apply assumption
done

lemma affset_solvable_gcd_dvd:
  "affset_solvable a0 a1 b1 \<Longrightarrow> ((gcd a0 a1) dvd b1)"
apply (clarsimp simp add: affset_solvable_def)
apply (rule gcd_bezout [THEN iffD1])
apply (subgoal_tac "b1 = a0 * n0 - a1 * n1") prefer 2 apply arith
apply fastforce
done


lemma affset_solvable_a0_zero [simp]: "affset_solvable 0 a1 b1 = (b1 = 0)"
by (clarsimp simp add: affset_solvable_def)

lemma affset_solvable_a1_zero [simp]: "affset_solvable a0 0 b1 = (b1 mod a0 = 0)"
by (rule iffI) (clarsimp simp add: affset_solvable_def)+

lemma affset_solvable_b1_zero [simp]: "affset_solvable a0 a1 0"
apply (clarsimp simp add: affset_solvable_def)
apply (rule_tac x=a1 in exI)
apply (rule_tac x=a0 in exI)
apply simp
done


lemma affset_solvable_gcd_one:
  "\<lbrakk> gcd a0 a1 = 1; a0 \<noteq> 0; a1 \<noteq> 0 \<rbrakk> \<Longrightarrow> affset_solvable a0 a1 b1"
apply (case_tac "b1 = 0")
apply simp
apply (clarsimp simp add: affset_solvable_def)
apply (subgoal_tac "(gcd a0 a1) dvd b1") prefer 2 apply simp
apply (drule gcd_bezout [THEN iffD2])
apply clarsimp
apply (erule disjE)

apply (rule_tac x=x in exI)
apply (rule_tac x=y in exI)
apply arith
apply clarsimp
apply (subgoal_tac "\<exists>n0 n1. a0 * n0 +  a0 * y = a1 * n1 + a1 * x")
apply clarsimp 
apply (rule_tac x=n0 in exI)
apply (rule_tac x=n1 in exI)
apply arith
apply (rule_tac x="(max x y) * a1 - y" in exI)
apply (rule_tac x="(max x y) * a0 - x" in exI)
apply (simp add: diff_mult_distrib2)
apply (subgoal_tac "a0 * (max x y * a1) \<ge> a0 * y")
apply (subgoal_tac "a1 * (max x y * a0) \<ge> a1 * x")
apply simp_all
apply (clarsimp simp add: max_def)
apply (subgoal_tac "y  \<le> y * a0") 
apply (erule le_trans) apply assumption apply simp

apply (clarsimp simp add: max_def)
apply (subgoal_tac "x \<le> x * a1") 
apply (rule le_trans) prefer 2 apply assumption apply simp+
done


definition
  affset_is_solution :: "[nat, nat, nat, nat] \<Rightarrow> bool" where
  "affset_is_solution a0 a1 b1 n == \<exists> n0. a0 * n0 = (a1 * n + b1)"
definition
  affset_is_solution_mod :: "[nat, nat, nat, nat] \<Rightarrow> bool" where
  "affset_is_solution_mod a0 a1 b1 n == ((a1 * n + b1) mod a0 = 0)"

lemma affset_is_solution_mod_adequate:
  "affset_is_solution a0 a1 b1 n = affset_is_solution_mod a0 a1 b1 n"
apply (simp add: affset_is_solution_def affset_is_solution_mod_def)
apply (rule iffI)
apply (clarsimp)
apply (drule sym)
apply simp
apply clarsimp
done


definition
  is_least_with :: "[nat \<Rightarrow> bool, nat] \<Rightarrow> bool" where
  "is_least_with P n == ((P n) \<and> (\<forall> n'. P n' \<longrightarrow> n \<le> n'))"


lemma "P n' \<Longrightarrow> \<exists> n. is_least_with P n"
by (fastforce simp add: is_least_with_def ex_has_least_nat)

definition
  affset_is_least_solution :: "[nat, nat, nat, nat] \<Rightarrow> bool" where
  "affset_is_least_solution a0 a1 b1 n == 
    is_least_with (affset_is_solution a0 a1 b1) n"


lemma affset_is_solution_b_mod:
  "affset_is_solution a0 a1 b1 n = affset_is_solution a0 a1 (b1 mod a0) n"
apply (simp add: affset_is_solution_mod_adequate zmod_simps)
apply (subgoal_tac "nat (int ((a1 * n + b1 mod a0) mod a0)) = nat (int ((a1 * n + b1) mod a0))")
apply (simp add: affset_is_solution_mod_def)
apply (simp add: zmod_int mod_add_right_eq [THEN sym])
done

lemma affset_is_solution_n_mod:
  "affset_is_solution a0 a1 b1 n = affset_is_solution a0 a1 b1 (n mod a0)"
apply (simp add: affset_is_solution_mod_adequate)
apply (subgoal_tac "((a1 * n) +  b1) = ((a1 * (a0 * (n div a0) + n mod a0)) + b1)")
    prefer 2 apply (simp add: mod_div_equality [THEN sym])
apply (simp only: affset_is_solution_mod_def)

apply (subgoal_tac "(a1 * (a0 * (n div a0) + n mod a0) + b1) = (a1 * (n mod a0) + b1 + (a1 * a0 * (n div a0)))")
  prefer 2
  apply (simp add: mult_assoc add_mult_distrib2 [THEN sym] mod_div_equality [THEN sym])

apply (simp only:)
apply (subgoal_tac "a1 * a0 * (n div a0) = a0 * a1 * (n div a0)") prefer 2 apply simp
apply (simp only: mod_mult_self2 mult_assoc)
done

lemma affset_solvable_is_solution: 
  "affset_solvable a0 a1 b1  = (\<exists> n. affset_is_solution a0 a1 b1 n)"
by (fastforce simp add: affset_solvable_def affset_is_solution_def)

lemma affset_solvable_is_solution_leq: "0 < a0 \<Longrightarrow>
  affset_solvable a0 a1 b1  = (\<exists> n < a0. affset_is_solution a0 a1 b1 n)"
apply (simp add: affset_solvable_is_solution)
apply (rule iffI)
apply clarsimp
apply (rename_tac n)
apply (rule_tac x="n mod a0" in exI)
apply (drule affset_is_solution_n_mod [THEN iffD1]) 
apply simp
apply fastforce
done

lemma affset_is_least_solution_le_aux:
  "\<lbrakk> a1 \<noteq> 0; affset_is_solution a0 a1 b1 n; n < a0; gcd a0 a1 = 1 \<rbrakk> 
  \<Longrightarrow> (affset_is_least_solution a0 a1 b1 n)"
apply (clarsimp simp add: affset_is_least_solution_def is_least_with_def)
apply (clarsimp simp add: affset_is_solution_def)
apply (rename_tac n1 n0a n0b)
apply (case_tac "n0a \<le> n0b")

apply (subgoal_tac "a0 * (n0b - n0a) = a1 * (n1 - n)") 
   prefer 2  apply (simp add: diff_mult_distrib2)
apply (case_tac "n \<le> n1") apply simp+

apply (subgoal_tac "a0 * (n0a - n0b) = a1 * (n - n1)")
   prefer 2  apply (simp add: diff_mult_distrib2) 
   apply (case_tac "n \<le> n1")  apply simp+

   apply (subgoal_tac "a0 dvd a0 * (n0a - n0b)") prefer 2 apply (rule dvd_triv_left)
   apply (subgoal_tac "\<not> a0 dvd a1 * (n - n1)") 
   apply simp
   apply clarify
   apply (subgoal_tac "(a0 dvd ((n - n1) * a1)) = (a0 dvd (n - n1))") 
     prefer 2 apply (simp add: relprime_dvd_mult_iff)
   apply (subgoal_tac " a1 * (n - n1) = (n - n1) * a1") prefer 2 apply simp
   apply (simp only:)
     (* 
     Argument requires combination of arithmetic and order: 
     n < a0 and \<not> n \<le> n1, thus 0 < n - n1 < a0, therefore \<not> a0 dvd n - n1 
     *)
   apply (subgoal_tac "n - n1 < a0")  prefer 2 apply simp
   apply (clarsimp simp add: dvd_def) 
done


lemma affset_is_least_solution_a0_zero [simp]:
  "(affset_is_least_solution 0 a1 b1 n) = (b1 = 0 \<and> n = 0)"
by (fastforce simp add: affset_is_least_solution_def is_least_with_def affset_is_solution_mod_adequate affset_is_solution_mod_def)

lemma affset_is_least_solution_le:
  "\<lbrakk> gcd a0 a1 = 1; a0 \<noteq> 0 \<rbrakk>
  \<Longrightarrow> (affset_is_least_solution a0 a1 b1 n) = 
   (affset_is_solution a0 a1 b1 n \<and> n < a0)"
apply (case_tac "a1 = 0")
apply (fastforce simp add: affset_is_least_solution_def is_least_with_def affset_is_solution_def)
apply (case_tac "a0 = 0")
apply (simp add: affset_is_least_solution_def is_least_with_def affset_is_solution_def)

apply (rule iffI)

apply (clarsimp simp add: affset_is_least_solution_def is_least_with_def)
apply (case_tac "n < a0") apply simp
apply (subgoal_tac "affset_is_solution a0 a1 b1 (n mod a0)")
  prefer 2 apply (simp add: affset_is_solution_n_mod [THEN sym])

apply (subgoal_tac "n mod  a0 < a0") prefer 2 apply simp
apply fastforce

apply (simp add: affset_is_least_solution_le_aux)
done

lemma affset_Int_not_solvable:
  "\<not> affset_solvable a0 a1 b1 
  \<Longrightarrow> (affset a0 0) \<inter> (affset a1 b1) = {}"
by (fastforce simp add: affset_def Int_def affset_solvable_def)



  (* Justification: a0 and a1 have no prime factors in common.
Therefore, prime_factors(a0) \<subseteq> prime_factors(n1), so progressively
divide by the prime factors of a0 to isolate n0.
 *)

lemma affset_Int_solvable_aux1: "\<lbrakk>gcd a0 a1 = Suc 0; a0 * n0 = a1 * n1; 0 < a0\<rbrakk>
       \<Longrightarrow> n0 = a1 * (n1 div a0)"
apply (subgoal_tac "a0 * n0 = a0 * a1 * (n1 div a0)") apply simp
apply (subgoal_tac "a0 dvd n1")
apply (frule dvd_mult_div_cancel)
apply simp
apply (rule relprime_dvd_mult) apply simp
apply (subgoal_tac "(a0 * n0) dvd (n1 * a1)") prefer 2 apply simp
apply (erule dvd_mult_left)
done

  (* 
gcd a0 a1 = 1 in a first time to make life easy. Should work without that
as well, by repeatedly applying affset_Int_common_factor
 *)

lemma affset_Int_solvable:
  "\<lbrakk> affset_is_least_solution a0 a1 b1 n; gcd a0 a1 = 1 \<rbrakk>
  \<Longrightarrow> (affset a0 0) \<inter> (affset a1 b1) = affset (a0 * a1) (a1 * n + b1)"
apply (simp add: affset_def Int_def)
apply (rule Collect_cong)
apply (rule iffI)
apply clarsimp
apply (rename_tac n0 n1)
apply (simp add: affset_is_least_solution_def is_least_with_def)
apply (clarsimp simp add: affset_is_solution_def)
apply (rule_tac x="(n0 - n0a) div a1" in exI)
apply (subgoal_tac "\<exists> k. (n0 - n0a) = a1 * k")
apply clarsimp
apply (subgoal_tac "a1 * n + a0 * a1 * k = a1 * n + a0 * (n0 - n0a)")
   prefer 2 apply simp
apply (simp (no_asm_use) only: diff_mult_distrib2)
apply simp
apply (subgoal_tac "a1 * n \<le> a1 * n1")
  prefer 2 apply fastforce
apply arith

  (* subgoal \<exists>k. n0 - n0a = a1 * k *)
apply (subgoal_tac "a0 * n0 - a1 * n1 = a0 * n0a - a1 * n")
  prefer 2 apply arith
apply (subgoal_tac "a0 * n0 - a0 * n0a = a1 * n1 - a1 * n")
  prefer 2 apply arith 
apply (simp (no_asm_use) only: diff_mult_distrib2 [THEN sym])

apply (case_tac "a0 = 0") apply simp
apply (rule exI)
apply (erule affset_Int_solvable_aux1)
apply assumption
apply simp

  (* other direction of iff *)
apply (clarsimp simp add: affset_is_least_solution_def is_least_with_def 
  affset_is_solution_def)
apply (subgoal_tac "a1 * n = a0 * n0 - b1") prefer 2 apply arith
apply simp
apply (simp add: add_mult_distrib2 [THEN sym] mult_assoc)
apply (rule_tac x="n + a0 *na" in exI)
apply (simp add: add_mult_distrib2 mult_assoc)
apply (simp add: mult_assoc[THEN sym])
apply (subgoal_tac "a0 * a1 = a1 * a0") apply (simp (no_asm_simp))
apply fastforce
done

  (* doesnt permit a presentation as finite union of affsets *)
lemma "- (affset 0 b) = - {b}"
by (simp add: affset_def)


lemma mod_exists_prod: "0 < q \<Longrightarrow>
  (((p::nat) mod q = r)) = ((\<exists> k. (p = k * q + r)) \<and> r < q)"
apply (rule iffI)
apply (drule sym)
apply simp
apply (rule exI) 
apply (rule sym)
apply (fastforce intro: mod_div_equality)
apply fastforce
done



lemma Compl_affset_sum0:
  "a =0 \<Longrightarrow> - (affset a b)= {0..<b} \<union> (affset (Suc 0) (Suc b))"
apply (subgoal_tac "(affset 1 (Suc b)) = {(Suc b)..}")
apply (simp add: Un_def)
apply (simp add: Compl_eq)
apply (fastforce)
apply (subgoal_tac "affset 1 (Suc b)=(setadd UNIV (Suc b))")
prefer 2
apply (fastforce simp add: affset_def setadd_def)
apply (fastforce)
done


lemma Compl_affset_sum: "0 < a \<Longrightarrow>
  - (affset a b) = {0 ..< b} \<union> (\<Union>c\<in>{1 ..< a}. (affset a (b + c)))"
apply (simp add: affset_def)
apply (simp add: Collect_bex_eq [THEN sym] Compl_eq Un_def)
apply (rule Collect_cong)
apply (rename_tac x)
apply (rule iffI)
apply (case_tac "x < b") apply simp apply simp
apply (subgoal_tac "x mod a \<noteq> b") 
   prefer 2 apply (rule notI) apply (drule mod_eqD) apply fastforce
apply (rule_tac x="(x - b) mod a" in bexI)
apply (rule_tac x="(x - b) div a" in exI)
apply fastforce
apply (simp add: atLeastAtMost_def) 
apply (case_tac "(x - b) mod a = 0")
apply simp
apply (clarsimp simp add: mod_exists_prod)
apply (rename_tac x k)
apply (subgoal_tac "x = k * a + b") prefer 2 apply simp
apply fastforce
apply fastforce

apply clarsimp
apply (rename_tac n c n0)
apply (subgoal_tac "c = a * n - a * n0") prefer 2 apply arith
apply (simp add: diff_mult_distrib2 [THEN sym])
done

  (* The following transformations are for simplification of affset unions *)

lemma affset_Union_UNIV: "0 < a \<Longrightarrow>
  (\<Union>c\<in>{0 ..< a}. (affset a c)) = (affset 1 0)"
apply (simp add: affset_def)
apply (rule equalityI)
apply clarsimp+
apply (rename_tac x)
apply (rule_tac x="x mod a" in bexI)
apply (rule_tac x="x div a" in exI)
apply (simp add:  mod_div_equality [THEN sym])
apply simp
done


lemma affset_inj_snd: 
  "\<lbrakk>affset a b0 = affset a b1; b0 < a; b1 < a\<rbrakk>
       \<Longrightarrow> b0 = b1"
apply (erule contrapos_pp)
apply (simp add: affset_def set_eq_iff)
apply (rule_tac x=b0 in exI)
apply (rule iffI)
apply clarsimp
apply (subgoal_tac "a * n \<ge> a") apply simp+
done


lemma ex_in_affset_snd_eq: "\<lbrakk>x \<in> affset a b0; b0 < a; b1 < a; x \<in> affset a b1\<rbrakk>
       \<Longrightarrow> b0 = b1"
apply (clarsimp simp add: affset_def)
apply (rename_tac n0 n1)
apply (case_tac "b0 < b1")
apply (subgoal_tac "a * n0 - a * n1 = b1 - b0") prefer 2 apply arith
apply (simp add: diff_mult_distrib2 [THEN sym])
apply (subgoal_tac "a dvd a * (n0 - n1)") prefer 2 apply (simp (no_asm))
apply (fastforce dest: dvd_imp_le)

apply (subgoal_tac "a * n1 - a * n0 = b0 - b1") prefer 2 apply arith
apply (simp add: diff_mult_distrib2 [THEN sym])
apply (subgoal_tac "a dvd a * (n1 - n0)") prefer 2 apply (simp (no_asm))
apply (fastforce dest: dvd_imp_le)
done

lemma Compl_affset: "\<lbrakk> b < a \<rbrakk> \<Longrightarrow>
  - (affset a b) = (\<Union>c\<in>({0 ..< a} - {b}). (affset a c))"
apply (subgoal_tac "0 < a") prefer 2 apply simp
apply (subgoal_tac "- (affset a b) = (affset 1 0) - (affset a b)")
  prefer 2 apply simp
apply (simp only: affset_Union_UNIV [THEN sym])
apply (rule set_eqI)
apply (rule iffI)
apply fastforce

apply simp
apply (rule conjI)
apply clarsimp
apply (clarsimp simp add: ex_in_affset_snd_eq)
done

lemma affset_shift_incl: "affset a (b + k * a) \<subseteq> affset a b"
apply (clarsimp simp add: affset_def)
apply (rule_tac x="n + k" in exI)
apply (fastforce simp add: add_mult_distrib2)
done

lemma affset_shift_simp: 
  "affset a (b + k * a) \<union> affset a b = affset a b"
apply (insert affset_shift_incl)
apply fast
done

lemma affset_subseteq_dvd_b: 
  "affset a0 b0 \<subseteq> (affset a1 b1) \<Longrightarrow> a1 dvd (b0 - b1)"
apply (erule contrapos_pp)
apply (simp add: affset_def subset_eq)
apply (rule_tac x=b0 in exI)
apply fastforce
done

lemma affset_subseteq_impos: 
  "(affset a0 b0 \<subseteq> affset a1 b1) \<Longrightarrow> b1 \<le> b0"
apply (clarsimp simp add: affset_def subset_eq)
apply (simp del: all_simps add: all_simps [THEN sym])
apply (drule_tac x=0 in spec)
apply clarsimp
done


lemma affset_subseteq_dvd_a:
  "affset a0 b0 \<subseteq> (affset a1 b1) \<Longrightarrow>  (a1 dvd a0)"
apply (frule affset_subseteq_impos)
apply (frule affset_subseteq_dvd_b)
apply (simp add: affset_def subset_eq)
apply (simp del: ex_simps add: ex_simps [THEN sym])
apply (simp del: all_simps add: all_simps [THEN sym])
apply (erule allE)
apply (erule exE)
apply (erule_tac x=1 in allE)
apply (erule impE)
apply (rule refl)
apply clarsimp
apply (subgoal_tac "(\<exists>k. b0 - b1 = a1 * k)") prefer 2 apply (simp add: dvd_def)
apply clarsimp
apply (subgoal_tac "b0 = a1 * k + b1") prefer 2 apply arith
apply simp
apply (subgoal_tac "a0 = a1 * (n -  k)") prefer 2 apply (simp add: diff_mult_distrib2)
apply simp
done



lemma affset_subset_setadd:
  "b1 \<le> b0 \<Longrightarrow>
  (affset a0 b0 \<subseteq> affset a1 b1) = 
  (affset a0 (b0 - b1) \<subseteq> affset a1 0)"
apply (subgoal_tac "affset a0 b0 = setadd (affset a0 (b0 - b1)) b1")
apply (subgoal_tac "affset a1 b1 = setadd (affset a1 0) b1")
apply (simp only: setadd_subseteq)
apply (simp add: setadd_affset)+
done

lemma affset_subset_setmult:
  "m \<noteq> 0 \<Longrightarrow>
  (affset (a0 * m) (b0 * m) \<subseteq> affset (a1 * m) (b1 * m)) = 
  (affset a0 b0 \<subseteq> affset a1 b1)"
by (simp add: setmult_affset [THEN sym])


lemma affset_subseteq_triv: "a1 = 0 \<Longrightarrow>  
  (affset a0 b0 \<subseteq> affset a1 b1) = (a0 = 0 \<and> b0 = b1)"
apply simp
apply (rule iffI)
apply (simp add: affset_def subset_eq)

apply (simp del: all_simps add: all_simps [THEN sym])
apply (case_tac "a0 = 0")
apply simp
apply (drule_tac x="b1 +1" in spec)
apply (subgoal_tac "a0 * b1 \<ge> b1")
apply simp+
done

lemma affset_subseteq_nontriv: "a1 \<noteq> 0 \<Longrightarrow> 
  (affset a0 b0 \<subseteq> affset a1 b1) = 
  (b1 \<le> b0 \<and> a1 dvd (b0 - b1) \<and> a1 dvd a0)"
apply(rule iffI)
apply (frule affset_subseteq_dvd_a, simp)
apply (frule affset_subseteq_dvd_b, simp)
apply (fastforce dest: affset_subseteq_impos) 

apply (elim conjE)
apply (erule affset_subset_setadd [THEN iffD2])
apply (simp add: dvd_def)
apply (elim exE)
apply simp
apply (rename_tac kb ka)
apply (subgoal_tac "affset (ka * a1) (kb * a1) \<subseteq> affset (1 * a1) (0 * a1)")
prefer 2
apply (rule affset_subset_setmult [THEN iffD2])
apply simp
apply simp
apply (subgoal_tac "(ka * a1) = (a1 * ka)")
apply (subgoal_tac "(kb * a1) = (a1 * kb)")
apply simp+
done

end

