theory Affset_calc imports Affset 

(* "~~/src/HOL/NumberTheory/IntPrimes" 
might be useful for a more efficient implementation, see below
*)

  (* Calculational version of affine sets: 
  affsets represented as data type *)

begin

datatype affset_tp 
  = Affset nat nat
  | AffsetEmpty

primrec 
  interp_affset :: "affset_tp \<Rightarrow> nat set"
where
  "interp_affset (Affset a b) = affset a b"|
  "interp_affset AffsetEmpty = {}"



primrec 
  search_downward :: "[nat \<Rightarrow> bool, nat] \<Rightarrow> nat" where
  "search_downward P 0 = 0" |
  "search_downward P (Suc n) = 
     (if (P (Suc n)) then (Suc n) else search_downward P n)"

lemma search_downward_correct [rule_format]:
  "(\<exists> n' \<le> n. P n') \<longrightarrow> P (search_downward P n)"
by (induct n) (auto simp add: le_Suc_eq)

lemma search_downward_sol_bound [simp]: "(search_downward P n \<le> n)"
by (induct n) auto



  (* A computationaly more efficient way of computing an n such that
  affset_is_solution a0 a1 b1 n
  might be to use the algorithm  xzgcd in 
  NumberTheory/IntPrimes/
  as follows, assuming gcd(a0, a1) = 1
  calculate xzgcd(a0, a1) = (1, k0, k1). From xzgcd_linear, one concludes
  1 = a0 * k0 + a1 * k1
  thus
  b1 = a0 * b1 * k0 + a1 * b1 * k1
  which permits to find an n such that
  \<exists> n0. a0 * n0 = (a1 * n + b1)
  and thus affset_is_solution a0 a1 b1 n
  *)

definition
  affset_least_solution_calc :: "[nat, nat, nat] \<Rightarrow> nat" where
  "affset_least_solution_calc a0 a1 b1 == 
  search_downward (affset_is_solution_mod a0 a1 b1) (a0 - 1)"



lemma affset_is_least_solution_by_calc:
  "\<lbrakk> gcd a0 a1 = 1; affset_solvable a0 a1 b1; a0 \<noteq> 0\<rbrakk> \<Longrightarrow>
  affset_is_least_solution a0 a1 b1 (affset_least_solution_calc a0 a1 b1)"
apply (simp add: affset_least_solution_calc_def)
apply (simp add: affset_is_least_solution_le)
apply (rule conjI)
apply (simp add: affset_is_solution_mod_adequate)
apply (rule search_downward_correct)
apply (simp add: affset_solvable_is_solution_leq affset_is_solution_mod_adequate)
apply clarsimp 
apply (intro exI conjI) prefer 2 apply assumption apply simp
apply (clarsimp simp add: affset_solvable_def)
apply (rule le_less_trans) apply (rule search_downward_sol_bound) apply simp
done


definition
  affset_solvable_cases :: "nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> bool" where
  "affset_solvable_cases a0 a1 b1 == 
  ((a0 \<noteq> 0 \<and> a1 \<noteq> 0) \<or>
  (a0 = 0 \<and> b1 = 0) \<or>
  (a1 = 0 \<and> b1 mod a0 = 0))"


lemma affset_solvable_cases_adequate:
  "gcd a0 a1 = 1 \<Longrightarrow>
  affset_solvable_cases a0 a1 b1 = affset_solvable a0 a1 b1"
apply (rule iffI)
apply (fastforce intro: affset_solvable_gcd_one simp add: affset_solvable_cases_def)

apply (simp add: affset_solvable_cases_def)
apply (case_tac "a0 = 0") apply simp
apply (case_tac "a1 = 0") apply simp+
done

definition
  affset_inter_0 :: "nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> affset_tp" where
  "affset_inter_0 a0 a1 b1 == 
  if affset_solvable_cases a0 a1 b1
  then 
  (let n = affset_least_solution_calc a0 a1 b1 in
  (Affset (a0 * a1) (a1 * n + b1)))
  else AffsetEmpty"

primrec
  affset_setmult_calc :: "[affset_tp, nat] \<Rightarrow> affset_tp"
where
  "affset_setmult_calc (Affset a b) m = (Affset (a * m) (b * m))" |
  "affset_setmult_calc AffsetEmpty m = AffsetEmpty"

lemma interp_affset_setmult_calc [simp]:
  "interp_affset (affset_setmult_calc s m)  = setmult (interp_affset s) m"
by (case_tac s) (simp add: setmult_affset)+

primrec
  affset_setadd_calc :: "[affset_tp, nat] \<Rightarrow> affset_tp"
where
  "affset_setadd_calc (Affset a b) m = (Affset a (b + m))" |
  "affset_setadd_calc AffsetEmpty m = AffsetEmpty"

lemma interp_affset_setadd_calc [simp]:
  "interp_affset (affset_setadd_calc s m)  = setadd (interp_affset s) m"
by (case_tac s) (simp add: setadd_affset)+



lemma interp_affset_gcd_calc:
  "gcd a0 a1 = 1 \<Longrightarrow>
  (interp_affset (Affset a0 0)) \<inter> (interp_affset (Affset a1 b1)) = 
  interp_affset (affset_inter_0 a0 a1 b1)"
apply (case_tac "a0 = 0")
apply (simp add: affset_inter_0_def affset_solvable_cases_def Let_def
  affset_least_solution_calc_def affset_def) 

apply (clarsimp simp add: affset_inter_0_def Let_def)
apply (intro conjI impI)

apply (rule affset_Int_solvable)
apply (rule affset_is_least_solution_by_calc) 
apply (simp add: affset_solvable_cases_adequate)+

apply (rule affset_Int_not_solvable)
apply (simp add: affset_solvable_cases_adequate)
done

definition
  affset_inter_calc_sorted :: "[nat, nat, nat, nat] \<Rightarrow> affset_tp" where
  "affset_inter_calc_sorted a0 b0 a1 b1 ==
       let m = gcd a0 a1 in
       if m = 0 
       then if b0 = b1 
            then (Affset 0 b0)
            else AffsetEmpty
       else 
            let d = b1 - b0 in
            if d mod m = 0 
            then 
             (affset_setadd_calc 
               (affset_setmult_calc 
                 (affset_inter_0 (a0 div m) (a1 div m) (d div m)) m) b0)
            else AffsetEmpty"


lemma affset_Int_no_common_factor_dvd:
  "\<lbrakk> c mod m \<noteq> 0; m dvd a0; m dvd a1 \<rbrakk>
  \<Longrightarrow> (affset a0 0) \<inter> (affset a1 c) = {}"
apply (simp add: dvd_def)
apply clarify
apply (rename_tac a0 a1)
apply (subgoal_tac "(m * a0) = (a0 * m)")
apply (subgoal_tac "(m * a1) = (a1 * m)")
apply (simp only:)
apply (rule affset_Int_no_common_factor) 
apply simp+
done

lemma affset_Int_no_common_factor_dvd':
assumes g: "\<not> (gcd a0 a1) dvd c"
  shows "(affset a0 0) \<inter> (affset a1 c) = {}"
proof-
  let ?g01 = "gcd a0 a1"
  from g have g': "c mod (gcd a0 a1) \<noteq> 0"
    unfolding dvd_eq_mod_eq_0 .
  from affset_Int_no_common_factor_dvd[ OF g' gcd_dvd1[of a0 a1] gcd_dvd2[of a0 a1]] show ?thesis .
qed



lemma affset_inter_diff: "b0 \<le> b1
    \<Longrightarrow> affset a0 b0 \<inter> affset a1 b1 = setadd (affset a0 0) b0 \<inter> setadd (affset a1 (b1 - b0)) b0"
by (simp add: setadd_affset)


lemma mod_zero_div: "\<lbrakk> (a::nat) mod m = 0\<rbrakk> \<Longrightarrow> a div m * m = a"
by (simp add: dvd_eq_mod_eq_0)


lemma affset_inter_calc_sorted_correct:
  "\<lbrakk> (b0 \<le> b1) \<rbrakk> \<Longrightarrow>
  (interp_affset (Affset a0 b0)) \<inter>  (interp_affset (Affset a1 b1)) =
  interp_affset (affset_inter_calc_sorted a0 b0 a1 b1)"

apply (case_tac "gcd a0 a1 = 0")
apply (simp add: gcd_zero)
apply (simp add: affset_inter_calc_sorted_def)


apply (simp add: affset_inter_calc_sorted_def Let_def)
apply (intro conjI impI)

apply (subgoal_tac "gcd (a0 div (gcd a0 a1)) (a1 div (gcd a0 a1)) = 1")
   prefer 2
   apply (rule div_gcd_relprime)
   apply (case_tac "gcd a0 a1 =0")  apply simp apply (simp add: gcd_zero) 
apply (simp add: interp_affset_gcd_calc [THEN sym])
apply (simp add: affset_inter_diff setmult_affset mod_zero_div)

apply (subgoal_tac "setadd (affset a0 0 \<inter> affset a1 (b1 - b0)) b0 = {}")
apply (simp add: setadd_affset)
apply (subgoal_tac "(affset a0 0 \<inter> affset a1 (b1 - b0)) = {}")
apply simp
apply (rule_tac m="gcd a0 a1" in affset_Int_no_common_factor_dvd)
apply simp+
done


fun affset_inter_calc :: "[affset_tp, affset_tp] \<Rightarrow> affset_tp"
where
  "affset_inter_calc (Affset a0 b0) (Affset a1 b1) = 
     (if (b0 \<le> b1)
     then affset_inter_calc_sorted a0 b0 a1 b1
     else affset_inter_calc_sorted a1 b1 a0 b0)"
  | "affset_inter_calc _ _ = AffsetEmpty"

lemma affset_inter_calc_correct: 
  "interp_affset (affset_inter_calc affs0 affs1) =
  (interp_affset affs0) \<inter> (interp_affset affs1)"
apply (case_tac affs0)
apply (case_tac affs1)
apply (simp add: affset_inter_calc_sorted_correct [THEN sym])
apply fastforce
apply simp+
done

(*************************************************************)
(*                  compl transformation                     *)
(*************************************************************)

primrec
  interp_affset_union :: "affset_tp list \<Rightarrow> nat set"
where
"interp_affset_union []  = {}" |
  "interp_affset_union (head#tail)=((interp_affset head) \<union> interp_affset_union (tail))"


primrec
gen_affset_singleton :: "nat \<Rightarrow> affset_tp list"
where
  "gen_affset_singleton 0 = [(Affset 0 0)]" |
  "gen_affset_singleton (Suc n) = ((Affset 0 (Suc n) )#(gen_affset_singleton n ))"

(*normal_form "tl (gen_affset_singleton (Suc (Suc 0)))"*)
  (* 
"[Affset 0 (Suc 0), Affset 0 0]" :: "affset_tp list"
 *)

primrec
  tl_gen_affset_singleton :: "nat \<Rightarrow> affset_tp list"
where
  "tl_gen_affset_singleton 0 = []"|
  "tl_gen_affset_singleton (Suc n) = ((Affset 0 n)#(tl_gen_affset_singleton n ))"

(*normal_form "tl_gen_affset_singleton (Suc (Suc 0))"*)


lemma interp_affset_singleton_1 :
  "interp_affset_union (gen_affset_singleton n)={0..<(Suc n)}"
apply (induct n)
apply (simp)
apply (fastforce)
done

lemma tl_interp_affset_singleton_1 :
  "interp_affset_union (tl_gen_affset_singleton n)={0..<n}"
apply (induct n)
apply (simp)
apply (fastforce)
done

lemma interp_affset_singleton :
  "interp_affset_union  (tl (gen_affset_singleton n))={0..< n}"
apply (induct n)
apply (simp)
apply (case_tac "n=0")
apply (simp)
apply (simp add: interp_affset_singleton_1)
done

primrec
  gen_union_list :: "affset_tp \<Rightarrow> nat \<Rightarrow> affset_tp list"
where
  "gen_union_list a 0 = []"|
  "gen_union_list a (Suc n) =((affset_setadd_calc a   (Suc n) )#(gen_union_list a n ))"

(*normal_form "tl (gen_union_list (Affset a b) (Suc (Suc (Suc 0))))"*)
  (* "[Affset a (b + Suc (Suc 0)), Affset a (b + Suc 0)]"
 *)
primrec
  tl_gen_union_list :: "affset_tp \<Rightarrow> nat \<Rightarrow> affset_tp list"
where
  "tl_gen_union_list a 0 = []"|
  "tl_gen_union_list a (Suc n) =((affset_setadd_calc a n)#(tl_gen_union_list a n ))"

(*normal_form "tl_gen_union_list (Affset a b) (Suc (Suc (Suc 0)))"*)

lemma UN_affset_un:
  "0 < n \<Longrightarrow>
  affset a (Suc (b +  n)) \<union> (\<Union> c \<in> {(Suc 0)..<(Suc n)} . affset a (b+c))
  = (\<Union> c \<in> {(Suc 0)..< (Suc (Suc n ))} . affset a (b + c))"
apply (subgoal_tac "{1..< (n + 2)} = {1..< (n + 1)}  \<union> {(n + 1)}")
apply fastforce+
done

lemma gen_union_list_to_set [rule_format]:
  "0 < n \<longrightarrow>
  interp_affset_union (gen_union_list (Affset a b) n) = 
  (\<Union> c\<in> {(Suc 0)..<(Suc n)} . affset a (b+c))"
apply (induct n)
apply (simp)
(*
apply (subgoal_tac "(Suc (b + n))=b + (Suc n)")
prefer 2
apply (simp)
apply (simp)
*)
apply (case_tac "0<n")
apply (simp add: UN_affset_un)
apply (simp)
done 


definition
  gen_union_list_1:: "affset_tp \<Rightarrow> nat \<Rightarrow> affset_tp list" where
  "gen_union_list_1 a n == (tl  (gen_union_list a n))"


lemma interp_union_gen_list_1_aux [rule_format]:
  "0< n \<longrightarrow>
  interp_affset_union (tl (gen_union_list (Affset a b) n))
  = (\<Union> c\<in> {1..<n} . affset a (b+c))"
apply (induct n)
apply (fastforce)
apply (case_tac "n=0")
apply (simp)
apply (simp add:  gen_union_list_to_set)
done


lemma interp_union_gen_list_1:
  "0 < n \<Longrightarrow> 
  interp_affset_union (gen_union_list_1 (Affset a b) n) = (\<Union> c\<in> {1..<n} . affset a (b+c))"
by (simp add: gen_union_list_1_def interp_union_gen_list_1_aux)

primrec
  compl_affset_calc :: "affset_tp \<Rightarrow> affset_tp list"
where
  "compl_affset_calc AffsetEmpty = [(Affset 1 0)]"|
  "compl_affset_calc (Affset a b) = 
  (tl_gen_affset_singleton b @ 
    (if (a = 0) 
    then 
      [Affset 1 (Suc b)]
    else 
      (gen_union_list_1 (Affset a b) a)))"

lemma interp_affset_union_concat:
  "interp_affset_union (l1@l2) = (interp_affset_union l1) \<union>  (interp_affset_union l2)"
by (induct l1, fastforce+)


lemma interp_compl1:
  "interp_affset_union (compl_affset_calc (Affset a b)) = - (affset a b)"
by (simp_all  add: Compl_affset_sum  interp_union_gen_list_1 
  tl_interp_affset_singleton_1 interp_affset_union_concat Compl_affset_sum0 [THEN sym])+



lemma interp_compl2:
  "interp_affset_union (compl_affset_calc a)= - (interp_affset a)"
apply (case_tac a)
apply (simp add: add: Compl_affset_sum  interp_union_gen_list_1 
  tl_interp_affset_singleton_1 interp_affset_union_concat Compl_affset_sum0 [THEN sym])+
done
end

