theory Affset_constr imports Setconstr Compos_constr Affset_calc

begin


definition
  swap :: "'a \<times> 'b \<Rightarrow> 'b \<times> 'a" where
  "swap = (\<lambda> (x,y). (y, x))"
definition
  pair_disj :: "'a set \<times> 'a set \<Rightarrow> bool" where
  "pair_disj = (\<lambda> (x,y). (x \<inter> y = {}))"
definition
  pair_union :: "('a set \<times> 'a set) \<Rightarrow> ('a set \<times> 'a set) \<Rightarrow> ('a set \<times> 'a set)" where
  "pair_union = (\<lambda> (p,n) (p',n'). (p \<union> p', n \<union> n'))"
definition
  pair_dom :: "('a set \<times> 'a set) \<Rightarrow> 'a set" where
  "pair_dom = (\<lambda> (p,n). (p \<union> n))"

  (* positive and negative variables of a set_rep *)

fun pnvars :: "('n, 'a) set_rep \<Rightarrow> 'n set * 'n set"
where
  "pnvars (Atom a) = ({}, {})"
| "pnvars (Bconst b) = ({}, {})"
| "pnvars (Var n) = ({n}, {})"
| "pnvars (Comp e) = swap (pnvars e)"
| "pnvars (Binop bop e e') = pair_union (pnvars e) (pnvars e')"


  (* all variables *)
fun vars :: "('n, 'a) set_rep \<Rightarrow> 'n set"
  where "vars t = (pair_dom (pnvars t))"

fun
  vars_constr :: "('n, 'a) setconstr \<Rightarrow> 'n set"
where "vars_constr (Lesseq l r) = (vars l \<union> vars r)"

lemma vars_constr_true_constr [simp]: "vars_constr true_constr = {}"
by (simp add: true_constr_def pair_dom_def)

lemma swap_swap [simp]: "swap (swap p) = p"
by (simp add: swap_def split_def)

lemma pair_disj_swap [simp]:
  "pair_disj (swap p) = pair_disj p"
by (fastforce simp add: swap_def pair_disj_def split_def)

lemma pair_disj_union: 
  "pair_disj (pair_union p p') \<Longrightarrow> (pair_disj p \<and> pair_disj p')"
by (fastforce simp add: pair_union_def pair_disj_def split_def)

lemma pair_union_swap [simp]:
  "pair_union (swap p) (swap p') = swap (pair_union p p')"
by (simp add: pair_union_def swap_def split_def)

lemma pair_dom_swap [simp]: "pair_dom (swap p) = pair_dom p"
by (fastforce simp add: pair_dom_def swap_def split_def)

lemma pair_dom_union [simp]:
  "pair_dom (pair_union p p') = (pair_dom p) \<union> (pair_dom p')"
by (fastforce simp add: pair_dom_def pair_union_def split_def)

definition
  subst_elim :: "'b \<Rightarrow> 'b \<Rightarrow> ('a set) \<times> ('a set) \<Rightarrow> 'a \<rightharpoonup> 'b" where
  "subst_elim pc nc  = (\<lambda> (pl, nl) x. 
            if x \<in> pl
            then Some pc
            else if x \<in> nl 
            then Some nc
            else None)"

definition
  subst_elim_res :: "'b \<Rightarrow> 'b \<Rightarrow> ('a set) \<Rightarrow> 'a \<Rightarrow> 'b" where
  "subst_elim_res pc nc = (\<lambda> pl x. 
            if x \<in> pl
            then pc
            else nc)"


definition
  simp_constr_vars :: 
  "(('n, 'a) set_rep) \<Rightarrow> ('n set * 'n set) \<Rightarrow> 
   (('n, 'a) set_rep) \<Rightarrow> ('n set * 'n set) \<Rightarrow> ('n, 'a) setconstr" where
  "simp_constr_vars l vl r vr =
  (if (pair_disj vl) \<and> (pair_disj vr)
   then 
    Lesseq 
     (subst_set_rep (subst_elim (Bconst Top) (Bconst Bot) vl) l)
     (subst_set_rep (subst_elim (Bconst Bot) (Bconst Top) vr) r)
   else true_constr)"


fun
  simp_elementary_constr :: "('n, 'a) setconstr \<Rightarrow> ('n, 'a) setconstr"
where 
  "simp_elementary_constr (Lesseq l r) = 
  (let (pl, nl) = (pnvars l) in
   let (pr, nr) = (pnvars r) in
   (if pl \<inter> pr = {} \<and> nl \<inter> nr = {} 
    then simp_constr_vars l (pl, nl) r (pr, nr)
    else true_constr))"

(*
fun
  simp_elementary_constr :: "('n, 'a) setconstr \<Rightarrow> ('n, 'a) setconstr"
where 
  "simp_elementary_constr (Lesseq l r) = 
  (let (pl, nl) = (pnvars l) in
   let (pr, nr) = (pnvars r) in
   (if pl \<inter> pr = {} \<and> nl \<inter> nr = {} 
    then simp_constr_vars l (pnvars l) r (pnvars r)
    else true_constr))"
*)

lemma subst_elim_swap [simp]: 
  "pair_disj p \<Longrightarrow>
  (subst_elim a b (swap p)) = 
  (subst_elim b a p)"
apply (case_tac p)
apply (clarsimp simp add: subst_elim_def swap_def fun_eq_iff)
apply (simp add: pair_disj_def)
apply blast
done


lemma interp_set_rep_vars_indiff_le [rule_format]: 
  "\<forall> vi vi'. 
  (\<forall> x \<in> fst (pnvars sr). vi x \<le> vi' x) \<longrightarrow> 
   (\<forall> x \<in> snd (pnvars sr). vi' x \<le> vi x) \<longrightarrow> 
  (interp_set_rep ai vi sr \<le> interp_set_rep ai vi' sr)"
apply (induct sr)
apply (clarsimp simp add: swap_def split_def)+
apply (clarsimp simp add: pair_union_def split_def ball_Un)
apply (case_tac bop)
apply simp
apply (blast intro: le_supI1 le_supI2)
apply simp
apply (blast intro: le_infI1 le_infI2)
done


lemma interp_set_rep_vars_indiff [rule_format]: 
  "(\<forall> x \<in> vars sr. vi x = vi' x) \<Longrightarrow>
  (interp_set_rep ai vi sr = interp_set_rep ai vi' sr)"
apply (rule antisym)
apply (fastforce simp add: pair_dom_def split_def
  intro: interp_set_rep_vars_indiff_le)+
done

lemma interp_set_rep_subst: "interp_set_rep ai vi (subst_set_rep s e) =
  interp_set_rep ai 
  (\<lambda> n. (case s n of None \<Rightarrow> vi n | Some t \<Rightarrow> interp_set_rep ai vi t)) e"
by (induct e) (clarsimp split add: option.splits)+


lemma subst_elim_interp_set_rep [rule_format]: "\<forall> n \<in> (vars sr). 
(case subst_elim (Bconst a) (Bconst b) (pnvars sr) n of None \<Rightarrow> vi n
          | Some t \<Rightarrow> interp_set_rep ai vi t) =
subst_elim_res (interp_bconst a) (interp_bconst b) (fst (pnvars sr)) n"
by (induct sr) (clarsimp simp add: pair_dom_def subst_elim_def subst_elim_res_def)+


lemma interp_set_rep_subst_set_rep: 
  "interp_set_rep ai vi
           (subst_set_rep (subst_elim (Bconst a) (Bconst b) (pnvars sr)) sr)
  = interp_set_rep ai
     (\<lambda> n. subst_elim_res (interp_bconst a) (interp_bconst b) (fst (pnvars sr)) n)
    sr"
by (fastforce simp add: interp_set_rep_subst
  intro: interp_set_rep_vars_indiff subst_elim_interp_set_rep)




lemma interp_set_rep_interpol [rule_format]: 
  "pair_disj (pnvars t) \<longrightarrow> 
  interp_set_rep ai vi t 
    \<le> interp_set_rep ai vi (subst_set_rep (subst_elim (Bconst Top) (Bconst Bot) (pnvars t)) t)
  \<and> interp_set_rep ai vi (subst_set_rep (subst_elim (Bconst Bot) (Bconst Top) (pnvars t)) t) 
    \<le> interp_set_rep ai vi t"
apply (clarsimp simp add: interp_set_rep_subst_set_rep)
apply (fastforce intro: interp_set_rep_vars_indiff_le 
  simp add: subst_elim_res_def pair_disj_def split_def)
done



lemma interp_elem_of_decompose_Inf_fst [rule_format]:
  "is_elementary_composition Inf sr \<longrightarrow>
  a \<in> fst (pnvars sr) \<longrightarrow>
  interp_set_rep ai vi sr \<le> interp_set_rep ai vi (Var a)"
apply (induct sr)
apply simp_all
apply (clarsimp simp add: swap_def split add: set_rep.splits )+
apply (simp add: pair_union_def split_def)
apply (fastforce elim: le_infI1 le_infI2)
done

lemma interp_elem_of_decompose_Sup_fst [rule_format]:
  "is_elementary_composition Sup sr \<longrightarrow>
  a \<in> fst (pnvars sr) \<longrightarrow>
  interp_set_rep ai vi (Var a) \<le> interp_set_rep ai vi sr"
apply (induct sr)
apply simp_all
apply (clarsimp simp add: swap_def split add: set_rep.splits )+
apply (simp add: pair_union_def split_def)
apply (fastforce elim: le_supI1 le_supI2)
done

lemma interp_elem_of_decompose_Inf_snd [rule_format]:
  "is_elementary_composition Inf sr \<longrightarrow>
  a \<in> snd (pnvars sr) \<longrightarrow>
  interp_set_rep ai vi sr \<le> interp_set_rep ai vi (Comp (Var a))"
apply (induct sr)
apply simp_all
apply (clarsimp simp add: swap_def split add: set_rep.splits )+
apply (simp add: pair_union_def split_def)
apply (fastforce elim: le_infI1 le_infI2)
done

lemma interp_elem_of_decompose_Sup_snd [rule_format]:
  "is_elementary_composition Sup sr \<longrightarrow>
  a \<in> snd (pnvars sr) \<longrightarrow>
  interp_set_rep ai vi (Comp (Var a)) \<le> interp_set_rep ai vi sr"
apply (induct sr)
apply simp_all
apply (clarsimp simp add: swap_def split add: set_rep.splits )+
apply (simp add: pair_union_def split_def)
apply (fastforce elim: le_supI1 le_supI2)
done

lemma interp_set_rep_Inf_Sup_nonempty:
  "\<lbrakk> is_elementary_composition Inf sr1; is_elementary_composition Sup sr2;
  a \<in> fst (pnvars sr1) \<inter> fst (pnvars sr2) \<or>
        a \<in> snd (pnvars sr1) \<inter> snd (pnvars sr2) \<rbrakk> \<Longrightarrow>
  interp_set_rep ai vi sr1 \<le> interp_set_rep ai vi sr2"
apply (elim disjE)
apply (rule order_trans)
apply (erule interp_elem_of_decompose_Inf_fst) apply fast
apply (erule interp_elem_of_decompose_Sup_fst) apply fast
apply (rule order_trans)
apply (erule interp_elem_of_decompose_Inf_snd) apply fast
apply (erule interp_elem_of_decompose_Sup_snd) apply fast
done

lemma interp_set_rep_elementary [simp]:
  "\<lbrakk> is_elementary sr; x \<in> fst (pnvars sr); x \<in> snd (pnvars sr)\<rbrakk>
  \<Longrightarrow> interp_set_rep ai vi sr = bot"
by (simp split add: set_rep.splits)

lemma is_elementary_is_elementary_composition:
  "is_elementary sr \<Longrightarrow> is_elementary_composition bop sr"
by (simp split add: set_rep.splits)



lemma neg_interp_leq_interp_Sup:
  "\<lbrakk>is_elementary_composition Sup sr1; is_elementary_composition Sup sr2;
   x \<in> fst (pnvars sr1); x \<in> snd (pnvars sr2)\<rbrakk> 
  \<Longrightarrow>  - (interp_set_rep ai vi sr2) \<le> interp_set_rep ai vi sr1"
apply (drule interp_elem_of_decompose_Sup_fst) apply assumption
apply (drule interp_elem_of_decompose_Sup_snd) apply assumption
apply (rule order_trans) 
apply (rule compl_le_impl)
apply simp+
done

lemma neg_interp_leq_interp_Inf:
  "\<lbrakk>is_elementary_composition Inf sr1; is_elementary_composition Inf sr2;
   x \<in> fst (pnvars sr1); x \<in> snd (pnvars sr2)\<rbrakk>
       \<Longrightarrow> interp_set_rep ai vi sr1 \<le> - interp_set_rep ai vi sr2"
apply (drule interp_elem_of_decompose_Inf_fst) apply assumption
apply (drule interp_elem_of_decompose_Inf_snd) apply assumption
apply (rule order_trans) apply assumption+
apply (rule compl_le_impl)
apply simp
done


lemma interp_set_rep_bconst_neutral [rule_format]: "
  x \<in>fst (pnvars sr) \<longrightarrow> x \<in> snd (pnvars sr)  \<longrightarrow>
  is_elementary_composition bop sr \<longrightarrow> 
  interp_set_rep ai vi sr = - interp_bconst (neutral_bop bop)"
apply (case_tac bop)

  (* Sup *)
apply (induct sr)
apply (clarsimp simp add: swap_def split_def Let_def pair_union_def)+
apply (intro conjI impI)

apply (rule sup_ge_compl_top) 
apply (rule compl_le_impl) apply simp
apply (fast intro: neg_interp_leq_interp_Sup) 

apply (rule sup_ge_compl_top)
apply (fast intro: neg_interp_leq_interp_Sup)

  (* Inf *)
apply (induct sr)
apply (clarsimp simp add: swap_def split_def Let_def pair_union_def 
  split add: set_rep.splits)+
apply (intro conjI impI)

apply (rule inf_ge_compl_bot)
apply (fast intro: neg_interp_leq_interp_Inf) 

apply (rule inf_ge_compl_bot)
apply (rule compl_le_impl) apply simp
apply (fast intro: neg_interp_leq_interp_Inf) 
done



lemma interp_simp_elementary_constr: "is_elementary_constr c \<Longrightarrow>
  interp_constr ai vi (simp_elementary_constr c) \<Longrightarrow>
  interp_constr ai vi c"
apply (case_tac c)
apply (clarsimp simp add: split_def Let_def)
apply (clarsimp split add: split_if_asm)

  (* case set (decompose bop.Inf set_rep1) \<inter> set (decompose bop.Sup set_rep2) = {}; *)
apply (rename_tac sr1 sr2)
apply (clarsimp simp add: simp_constr_vars_def split add: split_if_asm)

  (* case         pair_disj (pnvars sr1); pair_disj (pnvars sr2); *)
apply (drule interp_set_rep_interpol [where ai=ai and vi=vi])
apply (drule interp_set_rep_interpol [where ai=ai and vi=vi])
apply clarify
apply (rule order_trans) apply assumption
apply (rule order_trans) apply assumption+ 

  (* case  \<not> pair_disj (pnvars sr1) \<or> \<not> pair_disj (pnvars sr2); *)
apply (fastforce simp add: interp_set_rep_bconst_neutral pair_disj_def)

  (* case set (decompose bop.Inf set_rep1) \<inter> set (decompose bop.Sup set_rep2) \<noteq> {}; *)
apply (fast intro: interp_set_rep_Inf_Sup_nonempty)

done


lemma vars_subst_set_rep:
  "vars (subst_set_rep (subst_elim a b p) e) \<subseteq>
  (((vars e) - (pair_dom p)) \<union> (vars a) \<union> (vars b))"
apply (case_tac p)
apply (induct e)
apply (simp add: pair_dom_def split_def)+
apply (simp add: subst_elim_def)
apply fastforce+
done

lemma vars_subst_set_rep_const [simp]:
  "vars (subst_set_rep (subst_elim (Bconst c) (Bconst c') (pnvars e)) e) = {}"
apply (insert vars_subst_set_rep [of "(Bconst c)" "(Bconst c')" "(pnvars e)" e])
apply (simp add: pair_dom_def)
done

lemma vars_simp_elementary_constr [simp]:
  "vars_constr (simp_elementary_constr c) = {}"
by (case_tac c) 
   (clarsimp simp add: simp_constr_vars_def split_def Let_def
     simp del: vars.simps)

lemma valid_simp_elementary_constr:
  "is_elementary_constr c \<Longrightarrow>
  valid_constr ai (simp_elementary_constr c) = valid_constr ai c"
apply (simp add: valid_constr_def)
apply (rule iffI)
apply (clarsimp simp add: interp_simp_elementary_constr)

apply (case_tac c)
apply (clarsimp simp add: simp_constr_vars_def split_def Let_def)
apply (simp add: interp_set_rep_subst_set_rep)
apply (drule_tac x="(\<lambda>n. subst_elim_res \<top> \<bottom> 
 (fst (pnvars set_rep1) \<union> snd (pnvars set_rep2)) n)" in spec)

apply (subgoal_tac "interp_set_rep ai
     (\<lambda>n. subst_elim_res \<top> \<bottom> (fst (pnvars set_rep1) \<union> snd (pnvars set_rep2)) n) set_rep1 
    = interp_set_rep ai
   (\<lambda>n. subst_elim_res \<top> \<bottom> (fst (pnvars set_rep1)) n) set_rep1")
apply (subgoal_tac "interp_set_rep ai
           (\<lambda>n. subst_elim_res \<top> \<bottom>
                 (fst (pnvars set_rep1) \<union>  (snd (pnvars set_rep2))) n)
           set_rep2 = interp_set_rep ai
             (\<lambda>n. subst_elim_res \<bottom> \<top> (fst (pnvars set_rep2)) n) set_rep2")
apply simp
apply (fastforce intro: interp_set_rep_vars_indiff
simp add: pair_dom_def subst_elim_res_def pair_disj_def)+
done


end
