theory Compos_constr imports Setconstr

  (* Based on the modification of Setconstr_flat *)

  (* Composite constraints: conjunction of equalities and 'Let var in constraint'.
    - elim_let_subst eliminates the Let constructor
  *)

begin

  (* 'n is a variable type, 'a the type of expressions *)
datatype ('n, 'a) compos_constr 
  = CnTrue
  | CnEq 'a 'a
  | CnConj "(('n, 'a) compos_constr)" "(('n, 'a) compos_constr)"
  | CnLet 'n 'a "(('n, 'a) compos_constr)"

fun
  interp_compos_constr :: 
  "['n \<Rightarrow> 'b, ('n \<Rightarrow> 'b) \<Rightarrow> 'a \<Rightarrow> 'b, ('n, 'a) compos_constr ] \<Rightarrow> bool"
where
  "interp_compos_constr vari expri CnTrue = True"
| "interp_compos_constr vari expri (CnEq e e') = ((expri vari e) = (expri vari e'))"
| "interp_compos_constr vari expri (CnConj c c') = 
     (interp_compos_constr vari expri c \<and> interp_compos_constr vari expri c')"
| "interp_compos_constr vari expri (CnLet v e c) = interp_compos_constr (vari(v:=expri vari e)) expri c"


fun
  elim_let_subst :: "( ('n \<Rightarrow> 'a option) \<Rightarrow> 'a \<Rightarrow> 'a ) \<Rightarrow> ('n \<Rightarrow> 'a option) 
  \<Rightarrow> ('n,'a) compos_constr \<Rightarrow>  ('n,'a) compos_constr"
where
    "elim_let_subst subst s CnTrue = CnTrue"
  | "elim_let_subst subst s (CnEq e e') = (CnEq (subst s e)(subst s e'))"
  | "elim_let_subst subst s (CnConj c c') = 
         (CnConj (elim_let_subst subst s c) (elim_let_subst subst s c'))"
  | "elim_let_subst subst s (CnLet v a c) = (elim_let_subst subst (s(v\<mapsto>subst s a)) c)"


definition
valid_compos_constr :: "[('n \<Rightarrow> 'b) \<Rightarrow> 'a \<Rightarrow> 'b, ('n, 'a) compos_constr ] \<Rightarrow> bool" where
  "valid_compos_constr expri c = (\<forall> vari. interp_compos_constr vari expri c)"


fun 
  cnlet_free :: "[('n,'a) compos_constr ]\<Rightarrow>bool"
where
  "cnlet_free CnTrue = True"
| "cnlet_free (CnEq a a') = True"
| "cnlet_free (CnConj c c') = (cnlet_free c \<and> cnlet_free c')"
| "cnlet_free (CnLet n v c) = False"

lemma cnlet_free_elim_let_subst [rule_format,simp]: 
  "\<forall> s. cnlet_free (elim_let_subst subst s c)"
by (induct c) simp+


definition
  override_fun_with_map :: "('a \<Rightarrow> 'b) \<Rightarrow> ('a \<Rightarrow> 'b option) \<Rightarrow> 'a \<Rightarrow> 'b"
where
  "override_fun_with_map f m = (\<lambda>a. 
       case (m a) of
          None \<Rightarrow> f a
        | Some v \<Rightarrow> v)"

lemma override_fun_with_map_empty [simp]: "override_fun_with_map f empty = f"
by (simp add: override_fun_with_map_def)

lemma override_fun_with_map_upd:
  "(override_fun_with_map vari (s(n \<mapsto> a))) = ((override_fun_with_map vari s)(n := a))"
by (simp add: override_fun_with_map_def fun_eq_iff)

(* For testing only:
datatype 'a var_only = Var 'a
fun
  expri_var_only :: "('n \<Rightarrow> 'b) \<Rightarrow> 'n var_only \<Rightarrow> 'b"
where
  "expri_var_only vari (Var v) = (vari v)"


fun
  subst_var_only :: "('a \<Rightarrow> 'a var_only option) \<Rightarrow> 'a var_only \<Rightarrow> 'a var_only"
where
  "subst_var_only s (Var v) = (case (s v) of None \<Rightarrow> (Var v) | Some t \<Rightarrow> t)"

lemma "(subst_var_only (s(v \<mapsto> Var a)) (Var v)) = Var a"
by simp


lemma "(interp_compos_constr vari expri_var_only 
  (elim_let_subst subst_var_only s (CnLet v (Var a) (CnEq (Var v) (Var a)))))"
by (clarsimp split add: option.splits)
*)

lemma "(\<lambda>x. case s x of None \<Rightarrow> None | Some t \<Rightarrow> Some (expri vari t)) =
 ((Some \<circ> (expri vari)) o_m s)"
by (auto simp add: fun_eq_iff map_comp_def comp_def)


lemma override_fun_with_map_subst: "
  (override_fun_with_map vari ((Some \<circ> f) o_m (s(n \<mapsto> v)))) =
  (override_fun_with_map vari ((Some \<circ> f) o_m s))(n := f v)"
by (simp add: override_fun_with_map_def fun_eq_iff map_comp_def)


definition
  "subst_invariant expri vari subst =
  (\<forall> s e. 
    (expri (override_fun_with_map vari ((Some \<circ> (expri vari)) o_m s)) e)
  = (expri vari (subst s e)))"

lemma interp_compos_constr_elim_let_subst_override [rule_format]: 
  "subst_invariant expri vari subst
  \<longrightarrow>
  (\<forall> s. 
  (interp_compos_constr vari expri (elim_let_subst subst s c)) = 
  (interp_compos_constr
     (override_fun_with_map vari ((Some \<circ> (expri vari)) o_m s))
  expri c))"
apply (unfold subst_invariant_def)
apply (induct c)
apply (clarsimp simp only: 
  interp_compos_constr.simps elim_let_subst.simps
  override_fun_with_map_subst)+
done

lemma interp_compos_constr_elim_let_subst:
  "subst_invariant expri vari subst \<Longrightarrow> 
  (interp_compos_constr vari expri (elim_let_subst subst empty c)) = 
  (interp_compos_constr vari expri c)"
by (simp add: interp_compos_constr_elim_let_subst_override)



fun
  compos_constr_to_constr_list ::
  "[('n,(('n, 'a) set_rep)) compos_constr ] \<Rightarrow> ('n, 'a) setconstr list"
where 
  "compos_constr_to_constr_list CnTrue = []"
| "compos_constr_to_constr_list (CnEq a a')=[(Lesseq a a') , (Lesseq a' a)]"
| "compos_constr_to_constr_list (CnConj c c')=
     (compos_constr_to_constr_list c @ compos_constr_to_constr_list c')"


lemma compos_constr_to_setconstr: 
  "cnlet_free c \<Longrightarrow>
  (\<forall> x \<in> set (compos_constr_to_constr_list c).  interp_constr ai vi x) =
  interp_compos_constr vi (interp_set_rep ai) c"
by (induct c) (auto simp add: ball_Un)


(* the same as further below, using compos_constr_to_setconstr_list
lemma setconstr_list_decomp: "cnlet_free c \<Longrightarrow>
  (\<forall> x \<in> set (decomp_setconstr_list (compos_constr_to_constr_list c)) . interp_constr atomi x) =
  interp_compos_constr atomi interp_set_rep c"
apply (clarsimp simp add: decomp_setconstr_list_def)
apply (auto simp add: compos_constr_to_setconstr interp_constr_setconstr_to_list)
done
*)

definition
  compos_constr_to_setconstr_list :: 
  "('n, ('n, 'a) set_rep) compos_constr \<Rightarrow> ('n, 'a) setconstr list" where
  "compos_constr_to_setconstr_list c =
  (decomp_setconstr_list (compos_constr_to_constr_list c))"


lemma interp_setconstr_list_decomp: "cnlet_free c \<Longrightarrow>
  (\<forall> x \<in> set (compos_constr_to_setconstr_list c). interp_constr ai vi x)=
  interp_compos_constr vi (interp_set_rep ai) c"
apply (clarsimp simp add: compos_constr_to_setconstr_list_def decomp_setconstr_list_def)
apply (auto simp add: compos_constr_to_setconstr interp_constr_setconstr_to_list)
done


lemma is_elementary_setconstr_list_decomp: "
  (\<forall> x \<in> set (compos_constr_to_setconstr_list c). is_elementary_constr x)"
apply (clarsimp simp add: compos_constr_to_setconstr_list_def)
apply (erule decomp_setconstr_list_is_elementary_constr)
done

end

