theory Gencode imports Affset_constr "../SignalStreams_procedure_based/List_extended"


begin

  (* redefinition of executable versions of functions *)


definition
  pair_union_exec :: "('a list \<times> 'a list) \<Rightarrow> ('a list \<times> 'a list) 
  \<Rightarrow> ('a list \<times> 'a list)" where
  "pair_union_exec = (\<lambda> (p,n) (p',n'). (list_union p p', list_union n  n'))"

definition
  pair_disj_exec :: "'a list \<times> 'a list \<Rightarrow> bool" where
  "pair_disj_exec = (\<lambda> (x,y). (list_inter x y = []))"


definition set_pair_list :: "('a list \<times> 'a list) \<Rightarrow> ('a set \<times> 'a set)" where
  "set_pair_list  = (\<lambda> (p, n). (set p, set n))"

lemma set_pair_list_pair_union [simp]:
  "set_pair_list (pair_union_exec pl pl') = 
  pair_union (set_pair_list pl) (set_pair_list pl')"
by (simp add: pair_union_exec_def pair_union_def set_pair_list_def split_def)

lemma set_list_empty: "(xs = []) = (set xs = {})"
by (case_tac xs) simp+

lemma set_pair_disj_exec [simp]:
  "pair_disj_exec p = pair_disj (set_pair_list p)"
by (simp add: pair_disj_exec_def pair_disj_def set_pair_list_def split_def)
 (simp add: set_list_empty)

fun pnvars_exec :: "('n, 'a) set_rep \<Rightarrow> 'n list * 'n list"
where
  "pnvars_exec (Atom a) = ([], [])"
| "pnvars_exec (Bconst b) = ([], [])"
| "pnvars_exec (Var n) = ([n], [])"
| "pnvars_exec (Comp e) = swap (pnvars_exec e)"
| "pnvars_exec (Binop bop e e') = pair_union_exec (pnvars_exec e) (pnvars_exec e')"

lemma set_pair_list_swap [simp]: 
  "set_pair_list (swap p) = swap (set_pair_list p)"
by (simp add: set_pair_list_def swap_def split_def)

lemma set_pair_list_pnvars_exec [simp]:
  "set_pair_list (pnvars_exec sr) = pnvars sr"
apply (induct sr)
apply (simp add: set_pair_list_def)
apply (simp add: set_pair_list_def)
apply (simp add: set_pair_list_def)
apply simp+
done

definition
  subst_elim_exec :: "'b \<Rightarrow> 'b \<Rightarrow> ('a list) \<times> ('a list) \<Rightarrow> 'a \<rightharpoonup> 'b" where
  "subst_elim_exec pc nc  = (\<lambda> (pl, nl) x. 
            if x : set pl
            then Some pc
            else if x : set nl 
            then Some nc
            else None)"

lemma subst_elim_exec_adeq [simp]:
  "subst_elim_exec pc nc p = 
  subst_elim pc nc  (set_pair_list p)"
by (simp add: fun_eq_iff subst_elim_exec_def
 subst_elim_def set_pair_list_def split_def)

definition
  simp_constr_vars_exec :: 
  "(('n, 'a) set_rep) \<Rightarrow> ('n list * 'n list) \<Rightarrow> 
   (('n, 'a) set_rep) \<Rightarrow> ('n list * 'n list) \<Rightarrow> ('n, 'a) setconstr" where
  "simp_constr_vars_exec l vl r vr ==
  (if (pair_disj_exec vl) \<and> (pair_disj_exec vr)
   then 
    Lesseq 
     (subst_set_rep (subst_elim_exec (Bconst Top) (Bconst Bot) vl) l)
     (subst_set_rep (subst_elim_exec (Bconst Bot) (Bconst Top) vr) r)
   else true_constr)"

lemma simp_constr_vars_exec_adeq [simp]:
  "simp_constr_vars_exec l vl r vr =
  simp_constr_vars l (set_pair_list vl) r (set_pair_list vr)"
by (simp add: simp_constr_vars_exec_def simp_constr_vars_def)

fun
  simp_elementary_constr_exec :: "('n, 'a) setconstr \<Rightarrow> ('n, 'a) setconstr"
where 
  "simp_elementary_constr_exec (Lesseq l r) = 
  (let (pl, nl) = (pnvars_exec l) in
   let (pr, nr) = (pnvars_exec r) in
   (if list_inter pl pr = [] \<and> list_inter nl nr = []
    then simp_constr_vars_exec l (pl, nl) r (pr, nr)
    else true_constr))"

lemma fst_set_pair_list [simp]:
  "set (fst p) = fst (set_pair_list p)"
by (case_tac p) (simp add: set_pair_list_def)
lemma snd_set_pair_list [simp]:
  "set (snd p) = snd (set_pair_list p)"
by (case_tac p) (simp add: set_pair_list_def)

lemma simp_elementary_constr_exec_adeq:
  "simp_elementary_constr_exec c = simp_elementary_constr c"
by (case_tac c)
 (simp add: split_def Let_def set_list_empty)

definition
  simp_elementary_constr_list_exec :: "('n, 'a) setconstr list \<Rightarrow> ('n, 'a) setconstr list "
where
  "simp_elementary_constr_list_exec l = (map (\<lambda> x . simp_elementary_constr_exec x) l)"



lemma valid_compos_constr_elim_let_subst:
  "\<forall> vari .subst_invariant expri vari subst \<Longrightarrow>
  valid_compos_constr expri c \<longrightarrow> valid_compos_constr expri (elim_let_subst subst empty c)"
apply (simp add: valid_compos_constr_def)
apply (clarsimp simp add:  interp_compos_constr_elim_let_subst)
done

lemma valid_compos_constr_elim_let_subst_iff:
  "\<forall> vari .subst_invariant expri vari subst \<Longrightarrow>
  valid_compos_constr expri c = valid_compos_constr expri (elim_let_subst subst empty c)"
apply (rule iffI)
apply (simp add: valid_compos_constr_def)
apply (clarsimp simp add:  interp_compos_constr_elim_let_subst)
apply (simp add: valid_compos_constr_def)
apply (clarsimp simp add:  interp_compos_constr_elim_let_subst)
done


lemma valid_compos_constr_valid_setconstr:
  "cnlet_free c\<Longrightarrow>
  valid_compos_constr (interp_set_rep ai) c=
  (\<forall> x \<in> set (compos_constr_to_setconstr_list c) . valid_constr ai x)"
apply (rule iffI)
apply (simp add: valid_constr_def valid_compos_constr_def)
apply (clarsimp)
apply (simp add: interp_setconstr_list_decomp [THEN sym])
apply (simp add: valid_constr_def valid_compos_constr_def)
apply (clarsimp)
apply (simp add: interp_setconstr_list_decomp [THEN sym])
done


definition
  elim_let_and_decomp:: "( ('n \<Rightarrow> ('n,'a) set_rep  option) \<Rightarrow> ('n,'a) set_rep  \<Rightarrow> ('n,'a) set_rep ) \<Rightarrow>('n,('n,'a) set_rep) compos_constr \<Rightarrow> ('n,'a) setconstr list"
where
  "elim_let_and_decomp  subst c = compos_constr_to_setconstr_list (elim_let_subst subst empty c)"


lemma valid_elim_and_decomp:
  "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep ai) vari subst \<rbrakk> \<Longrightarrow> (valid_compos_constr (interp_set_rep ai) c )=
  (\<forall> x \<in> set (elim_let_and_decomp subst c) . valid_constr ai x)"
apply (rule iffI)
prefer 2
apply (simp add: elim_let_and_decomp_def)
apply (subgoal_tac "cnlet_free (elim_let_subst subst empty c)")
prefer 2 apply (simp add: cnlet_free_elim_let_subst)
apply (subgoal_tac "valid_compos_constr (interp_set_rep ai)  (elim_let_subst subst empty c)")
prefer 2 apply (simp add: valid_compos_constr_elim_let_subst)
apply (simp add:  valid_compos_constr_valid_setconstr)
apply (simp add: valid_compos_constr_elim_let_subst_iff [THEN sym])

apply (simp add: elim_let_and_decomp_def)
apply (subgoal_tac "cnlet_free (elim_let_subst subst empty c)")
prefer 2 apply (simp add: cnlet_free_elim_let_subst)
apply (subgoal_tac "valid_compos_constr (interp_set_rep ai)  (elim_let_subst subst empty c)")
prefer 2 apply (simp add: valid_compos_constr_elim_let_subst)
apply (simp add:  valid_compos_constr_valid_setconstr)
done



lemma is_elementary_elim_and_decomp:
  "(\<forall> x \<in> set (elim_let_and_decomp subst c) . is_elementary_constr x)"
by (simp add: elim_let_and_decomp_def is_elementary_setconstr_list_decomp)

fun 
  simplif_constr :: "('n,'a) setconstr \<Rightarrow> ('n,'a) setconstr"
where
  "simplif_constr (Lesseq l r) = (Lesseq (simplif_set_rep l)(simplif_set_rep r))"

lemma nf_constr_impl_is_nf_simplif_constr:
  "is_nf c \<longrightarrow> is_nf (simplif_constr c)"
apply (induct c)
apply (simp add: is_dnf_def is_cnf_def is_distrib_simplif)
done

lemma valid_simplif_constr:
  "valid_constr ai c \<Longrightarrow> valid_constr ai (simplif_constr c)"
apply (simp add: valid_constr_def)
apply (induct c)
apply (clarsimp)
apply (simp add: simplif_set_rep_equiv)
done




lemma valid_simplif_constr_iff:
  "valid_constr ai c = valid_constr ai (simplif_constr c)"
apply (rule iffI)
apply (simp add: valid_simplif_constr)
apply (simp add: valid_constr_def)
apply (clarsimp)
apply (induct c)
apply (clarsimp)
apply (simp add: simplif_set_rep_equiv)
done

definition 

  elim_let_decomp_simplif:: "( ('n \<Rightarrow> ('n,'a) set_rep  option) \<Rightarrow> ('n,'a) set_rep  \<Rightarrow> ('n,'a) set_rep ) \<Rightarrow>('n,('n,'a) set_rep) compos_constr \<Rightarrow> ('n,'a) setconstr list"
where
  "elim_let_decomp_simplif subst c = (map (\<lambda> x. simplif_constr x) ( simp_elementary_constr_list_exec (elim_let_and_decomp subst c)))"




lemma valid_elim_decomp_simplif:
  "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep ai) vari subst \<rbrakk> \<Longrightarrow>valid_compos_constr (interp_set_rep ai) c =
  (\<forall> x \<in> set (elim_let_decomp_simplif subst c) . valid_constr  ai x)"
apply (rule iffI)
apply (simp add: elim_let_decomp_simplif_def)
apply (simp add: simp_elementary_constr_list_exec_def)
apply (subgoal_tac "(\<forall> x \<in> set (elim_let_and_decomp subst c) . is_elementary_constr x)")
prefer 2
apply (simp add: is_elementary_elim_and_decomp)
apply (clarsimp)
apply (simp add:   simp_elementary_constr_exec_adeq)
apply (subgoal_tac "(\<forall> x \<in> set (elim_let_and_decomp subst c). valid_constr ai (simp_elementary_constr x))")
prefer 2
apply (simp add:   valid_simp_elementary_constr valid_elim_and_decomp)
apply (simp add: valid_simplif_constr)
apply (simp add: elim_let_decomp_simplif_def)
apply (simp add: simp_elementary_constr_list_exec_def)
apply (simp add: valid_elim_and_decomp valid_simplif_constr_iff[THEN sym] simp_elementary_constr_exec_adeq)
apply (subgoal_tac "(\<forall> x \<in> set (elim_let_and_decomp subst c) . is_elementary_constr x)")
prefer 2
apply (simp add: is_elementary_elim_and_decomp)
apply (clarsimp)
apply (simp add: valid_simp_elementary_constr)
done

definition 
list_switch_and_nf ::" ('n,'a) setconstr list \<Rightarrow>  ('n,'a) setconstr list" where
  "list_switch_and_nf l = (map (\<lambda> x . (switch_and_nf x)) l)"

definition 
  elim_decomp_simplif_switch :: "( ('n \<Rightarrow> ('n,'a) set_rep  option) \<Rightarrow> ('n,'a) set_rep  \<Rightarrow> ('n,'a) set_rep ) \<Rightarrow>('n,('n,'a) set_rep) compos_constr \<Rightarrow> ('n,'a) setconstr list"
where
  "elim_decomp_simplif_switch subst c = (map (\<lambda> x. (simplif_constr (switch_and_nf x))) (elim_let_decomp_simplif subst c))"


lemma valid_decomp_simplif_switch:
    "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep ai) vari subst ;valid_compos_constr (interp_set_rep ai) c \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (elim_decomp_simplif_switch subst c) . valid_constr  ai x)"


by (simp add: elim_decomp_simplif_switch_def valid_simplif_constr valid_constr_impl_valid_switch_constr [THEN sym] valid_elim_decomp_simplif)


lemma valid_decomp_simplif_switch_iff:
    "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep ai) vari subst \<rbrakk> \<Longrightarrow>valid_compos_constr (interp_set_rep ai) c =
  (\<forall> x \<in> set (elim_decomp_simplif_switch subst c) . valid_constr  ai x)"
apply (rule iffI)
apply (simp add: valid_decomp_simplif_switch)
apply (simp add: valid_elim_decomp_simplif)
apply (clarsimp)
apply (simp add: elim_decomp_simplif_switch_def valid_constr_impl_valid_switch_constr [THEN sym])
apply (simp add: valid_simplif_constr_iff  [THEN sym])
apply (simp add: valid_constr_impl_valid_switch_constr [THEN sym])
done



lemma no_vars_simplif_set_rep[simp]:
"vars s = {} \<longrightarrow> vars (simplif_set_rep s)={}"
apply (induct s)
apply (simp_all)
apply (intro conjI impI)
apply (simp_all add: pair_dom_def)
apply (simp_all split: bop.split)
apply (clarsimp)
apply (intro conjI impI)
apply (fastforce simp add: pair_union_def)+
done


lemma no_var_constr_simplif_constr:
  "vars_constr c = {} \<longrightarrow> vars_constr (simplif_constr c)={}"
apply (case_tac c)
apply (fastforce simp only : no_vars_simplif_set_rep simplif_constr.simps vars_constr.simps Un_empty)
done

lemma vars_elim_and_decomp_simplif:
  "(\<forall> x \<in> set (elim_let_decomp_simplif subst c) . vars_constr x={})"
apply ( simp add: elim_let_decomp_simplif_def) 
apply (simp add: simp_elementary_constr_list_exec_def simp_elementary_constr_exec_adeq)
apply (simp add: no_var_constr_simplif_constr)
done

 lemma vars_comp_distrib[simp]:
   "vars s = {} \<longrightarrow> vars (comp_distrib s)={}"
apply (induct s rule: comp_distrib.induct)
apply (simp_all)
done

lemma vars_distrib1[simp]:
  "(vars s1 = {} \<and>vars s2 ={}) \<longrightarrow> vars (distrib1 bop s1 s2)={}"
apply (induct s2)
apply (simp_all)
done

lemma vars_distrib2[simp]:
  "(vars s1 = {} \<and>vars s2 ={}) \<longrightarrow> vars (distrib2 bop s1 s2)={}"
apply (induct s1)
apply (simp_all  del: vars.simps)
apply (fastforce)
done




lemma vars_distrib[simp]:
  "vars s = {} \<longrightarrow> vars (distrib bop s)={}"
apply (induct s)
apply (simp_all del : vars.simps)
apply (intro conjI impI)
apply (clarsimp simp del: vars.simps)
prefer 2
apply (simp)
apply (subgoal_tac "(vars (Binop bop s1 s2)={}) =(vars s1={}\<and>vars s2 = {})")
apply (simp del: vars.simps)
apply (simp)
done

lemma vars_nf_constr[simp]:
  "vars_constr c= {} \<longrightarrow> vars_constr (nf_constr c)={}"
apply (induct c)
apply (simp add: cnf_def dnf_def vars_comp_distrib vars_distrib del: vars.simps)
done

lemma vars_switch_and_nf[simp]:
  "vars_constr c={} \<longrightarrow> vars_constr (switch_and_nf c)={}"
apply (induct c)
apply (simp del : vars.simps add: switch_and_nf_def)
apply (clarsimp simp del: vars.simps)
apply (rule conjI)
apply (subgoal_tac "vars (Comp set_rep2)={}")
prefer 2
apply (simp)
prefer 2

apply (simp add: cnf_def pair_dom_def)
apply (simp add: dnf_def del: vars.simps)
done


lemma decomp_simplif_switch_impl_no_vars[simp]:
    "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep ai) vari subst ;valid_compos_constr (interp_set_rep ai) c \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (elim_decomp_simplif_switch subst c) .(vars_constr x)={})"
apply (simp add: elim_decomp_simplif_switch_def)
apply (clarsimp)
apply (subgoal_tac "vars_constr x ={}")
prefer 2
apply (simp add: vars_elim_and_decomp_simplif)
apply (simp only:  no_var_constr_simplif_constr vars_switch_and_nf)

done


lemma decomp_simplif_switch_impl_no_vars1[simp]:
    "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep ai) vari subst  \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (elim_decomp_simplif_switch subst c) .(vars_constr x)={})"
apply (simp add: elim_decomp_simplif_switch_def)
apply (clarsimp)
apply (subgoal_tac "vars_constr x ={}")
prefer 2
apply (simp add: vars_elim_and_decomp_simplif)
apply (simp only:  no_var_constr_simplif_constr vars_switch_and_nf)

done




(*************************************************************************)
(*                       ('n,affset_tp) set_rep                          *)
(*************************************************************************)


fun 
  list_to_union :: "affset_tp list \<Rightarrow> ('n,affset_tp) set_rep"
where
  "list_to_union [] = (Bconst Bot)"
| "list_to_union (t#r) = (Binop Sup (Atom t) (list_to_union r))"


lemma list_to_union_set_rep[simp]:
  "interp_set_rep (interp_affset) vi (list_to_union l) = interp_affset_union l"
by (induct l, simp,simp)


fun
  compl_affset_set_rep_elementary :: "('n,affset_tp) set_rep\<Rightarrow> ('n,affset_tp) set_rep"
where
  "compl_affset_set_rep_elementary (Comp a) = 
  (case a of 
    (Atom a) \<Rightarrow>  list_to_union (compl_affset_calc a)
    |(Bconst b) \<Rightarrow> (if b = Top then (Bconst Bot) else (Bconst Top))
    | x\<Rightarrow> (Comp x))"






lemma Compl_affset_set_rep:
"is_elementary a \<Longrightarrow> 
  interp_set_rep (interp_affset) vi  (Comp a)=interp_set_rep (interp_affset) vi (compl_affset_set_rep_elementary  (Comp a))"
apply (induct a, simp_all add: interp_compl2)
apply (case_tac bconst)
apply (simp)
apply (simp)
done



fun 
compl_affset_set_rep_elem_compos :: "('n,affset_tp) set_rep\<Rightarrow> ('n,affset_tp) set_rep"
where

  "compl_affset_set_rep_elem_compos   (Atom a) = (Atom a)"
| "compl_affset_set_rep_elem_compos   (Bconst b)=(Bconst b)"
| "compl_affset_set_rep_elem_compos  (Var n) =(Var n)"
| "compl_affset_set_rep_elem_compos   (Comp a) = (compl_affset_set_rep_elementary (Comp a))"
| "compl_affset_set_rep_elem_compos   (Binop bop a b)=
     (Binop bop (compl_affset_set_rep_elem_compos  a) (compl_affset_set_rep_elem_compos  b))"


fun 
no_compl_set_rep ::"('n,'a) set_rep \<Rightarrow> bool"
where 
  "no_compl_set_rep (Comp a) = False"
  |"no_compl_set_rep (Binop bop a b)=((no_compl_set_rep a)\<and> (no_compl_set_rep b))"
  |"no_compl_set_rep _ = True"


fun 
  no_vars_set_rep :: "('n,'a)set_rep \<Rightarrow> bool"
  where
  "no_vars_set_rep (Var x)=False"
  |"no_vars_set_rep (Comp a)=(no_vars_set_rep a)"
  |"no_vars_set_rep (Binop bop a b) = ((no_vars_set_rep a) \<and> (no_vars_set_rep b))"
  |"no_vars_set_rep _ = True"


lemma No_vars_set_rep:
  "(vars s = {}) = no_vars_set_rep s"
apply (induct s)
apply (simp_all)
apply (simp add: pair_dom_def)+
done


lemma no_compl_set_rep_comp_distrib [simp]:
  "no_compl_set_rep s \<longrightarrow> (no_compl_set_rep (comp_distrib  s))"
apply (induct s)
apply (fastforce)
apply (simp_all)
done

lemma no_compl_set_rep_distrib1[simp]:
  "no_compl_set_rep s1 \<and> no_compl_set_rep s2
 \<longrightarrow> (no_compl_set_rep (distrib1 bop  s1 s2))"
apply (induct s2)
apply (simp_all)
done


lemma no_compl_set_rep_distrib2[simp]:
  "no_compl_set_rep s1 \<and> no_compl_set_rep s2
 \<longrightarrow> (no_compl_set_rep (distrib2 bop  s1 s2))"
apply (induct s1)
apply (simp_all)
done

lemma no_compl_set_rep_distrib[simp]:
  "no_compl_set_rep s
 \<longrightarrow> (no_compl_set_rep (distrib bop  s))"
apply (induct s)
apply (simp_all)
done


lemma no_compl_set_rep_list_to_union:
  "no_compl_set_rep (list_to_union a)"
apply (induct a)
apply (simp_all)
done

lemma no_vars_list_to_union:
  "vars (Atom a) ={} \<longrightarrow> vars (list_to_union  a)={}"
apply (induct a)
apply (clarsimp)
apply (simp add :pair_dom_def)
apply (simp)
apply (clarsimp simp add: pair_dom_def)
done

fun
  no_compl_constr ::"('n,'a) setconstr \<Rightarrow> bool"
where
  "no_compl_constr (Lesseq f g) = ((no_compl_set_rep f) \<and> (no_compl_set_rep g))"

lemma interp_compl_affset:
  "  interp_set_rep (interp_affset) vi (compl_affset_set_rep_elem_compos  s) = interp_set_rep (interp_affset) vi s"
apply (induct s)
apply (simp_all)
apply (simp split: set_rep.split)
apply (intro conjI impI)
apply (clarsimp)
apply (simp add: interp_compl2)
apply (clarsimp)
apply (case_tac bconst)
apply (simp_all add: Compl_empty_eq)
done





lemma normal_form_interp_compl_affset:
  "is__distrib bop  a\<and>is_comp_distrib a \<longrightarrow>
  interp_set_rep (interp_affset) vi (compl_affset_set_rep_elem_compos  s) = interp_set_rep (interp_affset) vi s"
apply (induct s)

apply  (simp_all  split: set_rep.split add : interp_compl2)+
apply (clarsimp)
apply (case_tac bconst)
apply (simp_all)
done




lemma is_elementary_composition_impl_no_compl_affset [rule_format,simp]:
  "((is_elementary_composition bop s)  \<and> (vars s = {})) \<longrightarrow>
  no_compl_set_rep (compl_affset_set_rep_elem_compos s)"
apply (induct s)
apply (simp_all)+
apply (simp split: set_rep.split)
apply (intro conjI impI)
apply (simp add: no_compl_set_rep_list_to_union)
apply (simp add: pair_dom_def)
done

lemma nf_no_compl_affset_set_rep[simp]:
  "((is_distrib bop s)  \<and> (vars s = {}) \<and>is_comp_distrib s) \<longrightarrow>
  no_compl_set_rep (compl_affset_set_rep_elem_compos s)"
apply (induct s)
apply (simp_all del : vars.simps split: set_rep.split)
apply (intro conjI impI)
apply (clarsimp)
apply (fastforce simp add : no_compl_set_rep_list_to_union pair_dom_def swap_def)+
apply (clarsimp)
done







fun 
  compl_affset_constr :: "('n,affset_tp) setconstr\<Rightarrow> ('n,affset_tp) setconstr"
where 
  "compl_affset_constr (Lesseq f g) = nf_constr (Lesseq (compl_affset_set_rep_elem_compos  f) (compl_affset_set_rep_elem_compos g))"

lemma elementary_constr_impl_no_compl: 
  "(is_elementary_constr c \<and> (vars_constr c = {})) \<longrightarrow> 
  no_compl_constr (compl_affset_constr c)"
apply (case_tac c)
apply (subgoal_tac "((vars set_rep1) \<union> (vars set_rep2)= {}) = ((vars set_rep1={} \<and> vars set_rep2 = {}))")
prefer 2
apply (simp)
apply (simp del: vars.simps)
apply (clarsimp simp add: dnf_def cnf_def)
done

lemma nf_constr_compl_affset:
  "is_nf c \<Longrightarrow> is_nf (compl_affset_constr c)"
by (case_tac c)
 (simp)

lemma interp_compl_affset_constr:
  "interp_constr (interp_affset) vi (compl_affset_constr c)=interp_constr (interp_affset) vi c"
apply (induct c)
apply (simp add: interp_compl_affset)
done


lemma valid_compl_affset_constr:
  "valid_constr (interp_affset) c \<Longrightarrow> valid_constr (interp_affset) (compl_affset_constr c)"
apply (simp add: valid_constr_def)
apply (clarsimp simp add: interp_compl_affset_constr)
done



lemma valid_compl_affset_constr_iff:
  "valid_constr (interp_affset) c = valid_constr (interp_affset) (compl_affset_constr c)"
apply (rule iffI)
apply (simp add: valid_compl_affset_constr)
apply (simp add: valid_constr_def)
apply (clarsimp simp add: interp_compl_affset_constr)
done


definition 
  affset_elim_decomp_simplif_switch :: 
"( ('n \<Rightarrow> ('n,affset_tp) set_rep  option) \<Rightarrow> ('n,affset_tp) set_rep  \<Rightarrow> ('n,affset_tp) set_rep ) \<Rightarrow>('n,('n,affset_tp) set_rep) compos_constr \<Rightarrow> ('n,affset_tp) setconstr list"
where
  "affset_elim_decomp_simplif_switch subst c = map (\<lambda> x. compl_affset_constr x) (elim_decomp_simplif_switch subst c)"




lemma no_compl_affset_constr_compl_affset_constr:
  "vars_constr c = {} \<and> is_nf c\<longrightarrow> no_compl_constr (compl_affset_constr c)"
apply (induct c)
apply (simp del: vars.simps)
apply (simp add: dnf_def cnf_def nf_no_compl_affset_set_rep del: vars.simps)
apply (clarsimp simp del: vars.simps)
apply (simp add: is_dnf_def is_cnf_def del:vars.simps)
apply (rule conjI)
apply (subgoal_tac "no_compl_set_rep (compl_affset_set_rep_elem_compos set_rep1)")
prefer 2
apply (subgoal_tac "is_comp_distrib set_rep1")

apply (simp  del: vars.simps)
apply (simp add: distrib_impl_comp_distrib)
apply (clarsimp simp add: nf_no_compl_affset_set_rep no_compl_set_rep_comp_distrib no_compl_set_rep_distrib)


apply (subgoal_tac "is_comp_distrib set_rep2")

apply (simp  del: vars.simps)
apply (simp add: distrib_impl_comp_distrib)
done


lemma valid_affset_elim_simplif_switch:
  "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst \<rbrakk> \<Longrightarrow>
  valid_compos_constr (interp_set_rep interp_affset ) c =
  (\<forall> x \<in> set (affset_elim_decomp_simplif_switch subst c) . valid_constr  (interp_affset) x)"
apply (rule iffI)
apply (simp add: affset_elim_decomp_simplif_switch_def)
apply (clarsimp)
apply (subgoal_tac "valid_constr (interp_affset)x")
apply (simp add: valid_compl_affset_constr)
apply (simp add: valid_decomp_simplif_switch)


apply (simp add: affset_elim_decomp_simplif_switch_def)
apply (simp add: valid_elim_decomp_simplif)
apply (clarsimp)
apply (clarsimp simp only: valid_compl_affset_constr_iff [THEN sym]) 
apply (simp add: elim_decomp_simplif_switch_def)
apply (simp add: valid_simplif_constr_iff [THEN sym])
apply (simp add: valid_constr_impl_valid_switch_constr [THEN sym])
done



lemma no_comp_affset_elim_simplif_switch:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst ;valid_compos_constr (interp_set_rep interp_affset) c \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_elim_decomp_simplif_switch subst c) .no_compl_constr  x)"
apply (simp add: affset_elim_decomp_simplif_switch_def)
apply (clarsimp)
apply (subgoal_tac "vars_constr x= {}")
prefer 2
apply (simp)
apply (subgoal_tac "is_nf x")
apply (simp add: no_compl_affset_constr_compl_affset_constr)
apply (simp)
apply (clarsimp simp add: is_comp_distrib_simplif is_distrib_simplif elim_decomp_simplif_switch_def)
apply (simp add: switch_and_nf_def nf_constr_impl_is_nf_simplif_constr  is_nf_nf_constr)
done

lemma no_comp_affset_elim_simplif_switch1:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_elim_decomp_simplif_switch subst c) .no_compl_constr  x)"
apply (simp add: affset_elim_decomp_simplif_switch_def)
apply (clarsimp)
apply (subgoal_tac "vars_constr x= {}")
prefer 2
apply (simp)
apply (subgoal_tac "is_nf x")
apply (simp add: no_compl_affset_constr_compl_affset_constr)
apply (simp)
apply (clarsimp simp add: is_comp_distrib_simplif is_distrib_simplif elim_decomp_simplif_switch_def)
apply (simp add: switch_and_nf_def nf_constr_impl_is_nf_simplif_constr  is_nf_nf_constr)
done



lemma no_vars_compl_affset_set_rep:
  "vars s = {} \<longrightarrow> vars (compl_affset_set_rep_elem_compos s)={}"
apply (induct s)
apply (simp)
apply (simp)
apply (simp)
prefer 2
apply (simp)
apply (rule impI)
apply (clarsimp simp del: vars.simps split: set_rep.split)
apply (intro conjI impI)
prefer 2
apply (fastforce)
prefer 2
apply (fastforce)
prefer 2 apply (fastforce)
prefer 2 apply (fastforce)
apply (clarsimp simp del: vars.simps)
apply (subgoal_tac "vars (Atom (compl_affset_calc a))={}")
prefer 2
apply (simp)
apply (simp add: no_vars_list_to_union del: vars.simps)
done
lemma no_var_constr_affset_constr:
  "vars_constr c = {} \<longrightarrow> vars_constr (compl_affset_constr c)={}"
apply (induct c)
apply (simp del: vars.simps add :no_vars_compl_affset_set_rep  dnf_def cnf_def)
done

lemma no_vars_affset_elim_simplif_switch:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst ;valid_compos_constr (interp_set_rep interp_affset) c \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_elim_decomp_simplif_switch subst c) .vars_constr  x={})"
apply (simp add: affset_elim_decomp_simplif_switch_def)
apply (clarsimp)
apply (subgoal_tac "vars_constr x={}")
prefer 2
apply (simp)
apply (simp add: no_var_constr_affset_constr  del: vars.simps)
done


lemma no_vars_affset_elim_simplif_switch1:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst  \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_elim_decomp_simplif_switch subst c) .vars_constr  x={})"
apply (simp add: affset_elim_decomp_simplif_switch_def)
apply (clarsimp)
apply (subgoal_tac "vars_constr x={}")
prefer 2
apply (simp)
apply (simp add: no_var_constr_affset_constr  del: vars.simps)
done


lemma no_right_constr_simplif:
  "no_right_constr c \<longrightarrow> no_right_constr (simplif_constr c)"
apply (induct c)
apply (simp)
done


lemma no_right_elim_simplif_switch:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst ;valid_compos_constr (interp_set_rep interp_affset) c \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (elim_decomp_simplif_switch subst c) .no_right_constr  x)"
apply (simp add: _elim_decomp_simplif_switch_def)
apply (clarsimp)

apply (simp add: no_right_switch_and_nf no_right_constr_simplif)
done

lemma no_right_elim_simplif_switch1:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst  \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (elim_decomp_simplif_switch subst c) .no_right_constr  x)"
apply (simp add: _elim_decomp_simplif_switch_def)
apply (clarsimp)

apply (simp add: no_right_switch_and_nf no_right_constr_simplif)
done


lemma no_right_constr_nf_constr:
  "no_right_constr c \<longrightarrow> no_right_constr (nf_constr c)"
apply (case_tac c)
apply (simp)
apply (simp add: cnf_def)
done

lemma no_right_compl_affset_constr:
  "no_right_constr c \<longrightarrow> no_right_constr (compl_affset_constr c)"
apply (case_tac c)
apply (simp add: cnf_def)
done


lemma no_right_affset_elim_simplif_switch:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst ;valid_compos_constr (interp_set_rep interp_affset) c \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_elim_decomp_simplif_switch subst c) .no_right_constr  x)"
apply (simp add: affset_elim_decomp_simplif_switch_def)
apply (clarsimp )
apply (simp add: no_right_constr_nf_constr no_right_elim_simplif_switch no_right_compl_affset_constr)
done

lemma no_right_affset_elim_simplif_switch1:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_elim_decomp_simplif_switch subst c) .no_right_constr  x)"
apply (simp add: affset_elim_decomp_simplif_switch_def)
apply (clarsimp )
apply (simp add: no_right_constr_nf_constr no_right_elim_simplif_switch1 no_right_compl_affset_constr)
done




(**********************************************************************************************************)
definition
  affset_decompose ::
"( ('n \<Rightarrow> ('n,affset_tp) set_rep  option) \<Rightarrow> ('n,affset_tp) set_rep  \<Rightarrow> ('n,affset_tp) set_rep ) \<Rightarrow>('n,('n,affset_tp) set_rep) compos_constr \<Rightarrow> ('n,affset_tp) setconstr list"
where
  "affset_decompose subst c = decomp_setconstr_list (affset_elim_decomp_simplif_switch subst c)"

lemma is_elementary_affset_decompose:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst ;valid_compos_constr (interp_set_rep interp_affset) c \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose subst c) .is_elementary_constr   x)"
apply (simp add: affset_decompose_def)
apply (simp add: decomp_setconstr_list_is_elementary_constr) 
done

lemma is_elementary_affset_decompose1:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst  \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose subst c) .is_elementary_constr   x)"
apply (simp add: affset_decompose_def)
apply (simp add: decomp_setconstr_list_is_elementary_constr) 
done



lemma no_right_setconstr_decompose_left:
  "no_right_constr c \<longrightarrow> (\<forall> x\<in> set (setconstr_decompose_left c) . no_right_constr x)"
apply (induct c)
apply (simp)
done


lemma no_right_setconstr_decompose_right:
  "no_right_constr c \<longrightarrow> (\<forall> x\<in> set (setconstr_decompose_right c) . no_right_constr x)"
apply (induct c)
apply (simp)
done

lemma no_right_setconstr_to_list:
  "no_right_constr c \<longrightarrow> (\<forall> x\<in> set (setconstr_to_list c) . no_right_constr x)"
apply (simp add: setconstr_to_list_def)
apply (clarsimp)
apply (subgoal_tac "no_right_constr a")
apply (subgoal_tac "no_right_constr x")
prefer 2
apply (simp add: no_right_setconstr_decompose_right)
prefer 2
apply (simp add: no_right_setconstr_decompose_left)
apply (simp)
done


lemma no_compl_decompose:
  "no_compl_set_rep s \<longrightarrow> (\<forall> x \<in> set (decompose bop s) .no_compl_set_rep x)"
apply (induct s)
apply (simp)
apply (simp)
apply (simp)
apply (simp (no_asm_use) only: decompose.simps)
apply (rule impI)
apply (simp (no_asm_use))
apply (simp (no_asm_use) only :decompose.simps)
apply (simp (no_asm_use) split: split_if)
apply (blast)
done



lemma no_compl_decompose_left:
"no_compl_constr c \<longrightarrow> (\<forall> x \<in> set (setconstr_decompose_left c) . no_compl_constr x)"
apply (induct c)
apply (simp)
apply (clarsimp)
apply (simp add: no_compl_decompose)
done


lemma no_compl_decompose_right:
"no_compl_constr c \<longrightarrow> (\<forall> x \<in> set (setconstr_decompose_right c) . no_compl_constr x)"
apply (induct c)
apply (simp)
apply (clarsimp)
apply (simp add: no_compl_decompose)
done

lemma no_compl_setconstr_to_list:
"no_compl_constr c \<longrightarrow> (\<forall> x \<in> set (setconstr_to_list c) . no_compl_constr x)"
apply (simp add: setconstr_to_list_def)
apply (clarsimp)
apply (subgoal_tac "no_compl_constr c")
apply (subgoal_tac "no_compl_constr a")
apply (simp only: no_compl_decompose_right)
apply (simp add: no_compl_decompose_left)
apply (simp  add: no_compl_decompose)
done

lemma valid_constr_setconstr_to_list1:
  "valid_constr ai c \<longrightarrow> (\<forall> x\<in>set  (setconstr_to_list c) .( valid_constr ai x))"

apply ( simp add: valid_constr_def)+
apply (clarsimp)
apply (subgoal_tac "(\<forall> x \<in> set (setconstr_to_list c) .interp_constr ai vi x) =( interp_constr ai vi c)")
apply (fastforce)
apply (simp add: interp_constr_setconstr_to_list)
done




lemma valid_constr_setconstr_to_list2:
  " (\<forall> x\<in>set  (setconstr_to_list c) .( valid_constr ai x)) \<longrightarrow> valid_constr ai c"

apply (simp add: valid_constr_def)
apply (clarsimp)
apply (subgoal_tac "(\<forall> x \<in> set (setconstr_to_list c) .interp_constr ai vi x) =( interp_constr ai vi c)")
apply (fastforce)
apply (simp only: interp_constr_setconstr_to_list)
done

lemma no_vars_decompose:
  "vars  s={} \<longrightarrow> (\<forall> x \<in> set (decompose bop s) . vars x={})"
apply (induct s)
apply (simp)
apply (simp)
apply (simp)
apply (simp (no_asm_use) only: decompose.simps)
apply (rule impI)
apply (simp (no_asm_use))
apply ( simp (no_asm_use) only: decompose.simps)
apply (rule impI)
apply (simp (no_asm_use) split: split_if del : vars.simps)
apply (intro conjI impI)
prefer 2
apply (assumption)
apply (subgoal_tac "vars (Binop bopa s1 s2)={} \<longrightarrow> (vars s1 ={} \<and> vars s2 = {})")
prefer 2
apply (simp (no_asm_use))
apply (simp (no_asm_use))
apply (fastforce)
done




lemma no_vars_setconstr_decompose_right:
  "vars_constr c = {} \<longrightarrow> (\<forall> x \<in> set (setconstr_decompose_right c) . vars_constr x = {})"
apply (case_tac c)
apply (simp del: vars.simps)
apply (intro conjI impI)
apply (subgoal_tac "vars set_rep1 = {}")
apply (clarsimp simp del: vars.simps)
apply (simp add:  no_vars_decompose del: vars.simps)+
done




lemma no_vars_setconstr_decompose_left:
  "vars_constr c = {} \<longrightarrow> (\<forall> x \<in> set (setconstr_decompose_left c) . vars_constr x = {})"
apply (case_tac c)
apply (simp del: vars.simps)
apply (intro conjI impI)
apply (subgoal_tac "vars set_rep1 = {}")
apply (clarsimp simp del: vars.simps)
apply (simp add:  no_vars_decompose del: vars.simps)+
done


lemma no_vars_setconstr_to_list[simp]:
  "vars_constr c ={} \<longrightarrow> (\<forall> x \<in> set (setconstr_to_list c) . (vars_constr x = {}))"
apply (simp only: setconstr_to_list_def)
apply (rule impI)
apply (simp add: concat.simps)
apply (clarsimp)
apply (subgoal_tac "vars_constr a={}")
apply (simp add: no_vars_setconstr_decompose_right)
apply (simp add: no_vars_setconstr_decompose_left)
done

lemma no_vars_affset_decompose:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst ;valid_compos_constr (interp_set_rep interp_affset) c \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose subst c) .vars_constr   x={})"
apply (simp add: affset_decompose_def)
apply (simp add: decomp_setconstr_list_def)
apply (clarsimp)
apply (subgoal_tac "vars_constr a= {}")
apply (subgoal_tac "vars_constr (nf_constr a)={}")
apply (simp only: no_vars_setconstr_to_list)
apply (simp add: no_vars_affset_elim_simplif_switch)
apply (simp add: no_vars_affset_elim_simplif_switch)
done

lemma no_vars_affset_decompose1:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst  \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose subst c) .vars_constr   x={})"
apply (simp add: affset_decompose_def)
apply (simp add: decomp_setconstr_list_def)
apply (clarsimp)
apply (subgoal_tac "vars_constr a= {}")
apply (subgoal_tac "vars_constr (nf_constr a)={}")
apply (simp only: no_vars_setconstr_to_list)
apply (simp add: no_vars_affset_elim_simplif_switch)
apply (simp add: no_vars_affset_elim_simplif_switch1)
done


lemma no_compl_affset_decompose:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst ;valid_compos_constr (interp_set_rep interp_affset) c \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose subst c) .no_compl_constr   x)"
apply (simp add: affset_decompose_def decomp_setconstr_list_def)
apply (clarsimp)
apply (subgoal_tac  "no_compl_constr a")
apply (subgoal_tac "no_compl_constr (nf_constr a)")
apply (simp add: no_compl_setconstr_to_list)
prefer 2
apply (simp add: no_comp_affset_elim_simplif_switch)
apply (case_tac a)
apply (simp)
apply (simp add: dnf_def cnf_def)
done


lemma no_compl_affset_decompose1:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose subst c) .no_compl_constr   x)"
apply (simp add: affset_decompose_def decomp_setconstr_list_def)
apply (clarsimp)
apply (subgoal_tac  "no_compl_constr a")
apply (subgoal_tac "no_compl_constr (nf_constr a)")
apply (simp add: no_compl_setconstr_to_list)
prefer 2
apply (simp add: no_comp_affset_elim_simplif_switch1)
apply (case_tac a)
apply (simp)
apply (simp add: dnf_def cnf_def)
done

lemma valid_affset_decompose:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst \<rbrakk> \<Longrightarrow> valid_compos_constr (interp_set_rep interp_affset) c =
  (\<forall> x \<in> set (affset_decompose subst c) .valid_constr (interp_affset) x)"
apply (rule iffI)
apply (simp add: affset_decompose_def)
apply (simp add: decomp_setconstr_list_def)

apply (clarsimp)
apply (subgoal_tac "valid_constr (interp_affset)  a")
apply (subgoal_tac "valid_constr (interp_affset) (nf_constr a)")
apply (simp add: valid_constr_setconstr_to_list1)
apply (simp add: valid_constr_def)

apply (simp add: valid_affset_elim_simplif_switch)



apply (simp add: affset_decompose_def)
apply (simp add: decomp_setconstr_list_def)
apply (subgoal_tac "(\<forall>a\<in>set (affset_elim_decomp_simplif_switch subst c).valid_constr (interp_affset) a)")
apply (simp add: valid_affset_elim_simplif_switch)
apply (clarsimp)
apply (subgoal_tac "valid_constr (interp_affset) (nf_constr a)=valid_constr (interp_affset) a")
apply (simp add: valid_constr_setconstr_to_list2)
apply (simp add: valid_constr_def)
done


fun 
no_bconst_in_binop_constr :: "('n,'a) setconstr \<Rightarrow> bool"
where
  "no_bconst_in_binop_constr (Lesseq f g) = (no_bconst_in_binop f \<and>no_bconst_in_binop g)"

(*********************************************************************************************)
lemma no_bconst_in_binop_constr_simplif:
  "no_bconst_in_binop_constr (simplif_constr c)"
apply (case_tac c)
apply (simp add:no_bconst_in_binop_set_rep)
done


definition
affset_decompose_simplif::
"( ('n \<Rightarrow> ('n,affset_tp) set_rep  option) \<Rightarrow> ('n,affset_tp) set_rep  \<Rightarrow> ('n,affset_tp) set_rep ) \<Rightarrow>('n,('n,affset_tp) set_rep) compos_constr \<Rightarrow> ('n,affset_tp) setconstr list"
where
  "affset_decompose_simplif subst c = map (\<lambda> x . simplif_constr x) (affset_decompose subst c)"



lemma no_bconst_in_binop_affset_decompose_simplif:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst ;valid_compos_constr (interp_set_rep interp_affset) c \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose_simplif subst c) .no_bconst_in_binop_constr  x)"
apply (simp add: affset_decompose_simplif_def)
apply (simp add: no_bconst_in_binop_constr_simplif)
done


lemma no_bconst_in_binop_affset_decompose_simplif1:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose_simplif subst c) .no_bconst_in_binop_constr  x)"
apply (simp add: affset_decompose_simplif_def)
apply (simp add: no_bconst_in_binop_constr_simplif)
done

lemma valid_affset_decompose_simplif:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst \<rbrakk> \<Longrightarrow>valid_compos_constr (interp_set_rep interp_affset) c =
  (\<forall> x \<in> set (affset_decompose_simplif subst c) .valid_constr (interp_affset)  x)"
apply (rule iffI)
apply (simp add: affset_decompose_simplif_def)
apply (clarsimp)
apply (subgoal_tac "valid_constr (interp_affset) x")
apply (subgoal_tac "valid_constr (interp_affset) (simplif_constr x)")

apply (simp only:)
apply (simp only: valid_simplif_constr)
apply (simp add: valid_affset_decompose)
apply (simp add: affset_decompose_simplif_def)
apply (simp add: valid_affset_decompose)
apply (clarsimp)
apply (subgoal_tac "valid_constr (interp_affset) (simplif_constr x)=valid_constr (interp_affset) x")
apply (simp add: valid_affset_decompose)
apply (simp add: valid_simplif_constr_iff [THEN sym])
done

lemma no_compl_simplif_set_rep:
  "no_compl_set_rep s \<longrightarrow> no_compl_set_rep (simplif_set_rep s)"
apply (induct s)
apply (simp_all)
apply (rule conjI impI)
apply (simp_all split: bop.split)
done

lemma no_compl_constr_simplif:
  "no_compl_constr c \<longrightarrow> no_compl_constr (simplif_constr c)"
apply (case_tac c)
apply (simp add: no_compl_simplif_set_rep)
done

lemma no_compl_affset_decompose_simplif:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst ;valid_compos_constr (interp_set_rep interp_affset) c \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose_simplif subst c) .no_compl_constr  x)"
apply (simp add: affset_decompose_simplif_def)
apply (clarsimp)
apply (subgoal_tac "no_compl_constr x")
apply (subgoal_tac "no_compl_constr (simplif_constr x)")
apply (simp only:)
prefer 2
apply (simp add: no_compl_affset_decompose)
apply (simp add: no_compl_constr_simplif)
done

lemma no_compl_affset_decompose_simplif1:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose_simplif subst c) .no_compl_constr  x)"
apply (simp add: affset_decompose_simplif_def)
apply (clarsimp)
apply (subgoal_tac "no_compl_constr x")
apply (subgoal_tac "no_compl_constr (simplif_constr x)")
apply (simp only:)
prefer 2
apply (simp add: no_compl_affset_decompose1)
apply (simp add: no_compl_constr_simplif)
done



lemma no_vars_affset_decompose_simplif:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst ;valid_compos_constr (interp_set_rep interp_affset) c \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose_simplif subst c) .vars_constr  x={})"
apply (simp add: affset_decompose_simplif_def)
apply (clarsimp)
apply (subgoal_tac "vars_constr x={}")
apply (subgoal_tac "vars_constr (simplif_constr x)={}")

apply (simp only:)
prefer 2
apply (simp add: no_vars_affset_decompose)
apply (case_tac x)
apply (simp del: vars.simps)
done



lemma no_vars_affset_decompose_simplif1:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose_simplif subst c) .vars_constr  x={})"
apply (simp add: affset_decompose_simplif_def)
apply (clarsimp)
apply (subgoal_tac "vars_constr x={}")
apply (subgoal_tac "vars_constr (simplif_constr x)={}")

apply (simp only:)
prefer 2
apply (simp add: no_vars_affset_decompose1)
apply (case_tac x)
apply (simp del: vars.simps)
done

 


lemma no_right_affset_decompose_simplif:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst ;valid_compos_constr (interp_set_rep interp_affset) c \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose_simplif subst c) .no_right_constr  x)"
apply (simp add: affset_decompose_simplif_def)
apply (clarsimp)
apply (simp add: affset_decompose_def)
apply (simp add: decomp_setconstr_list_def)
apply (clarsimp)
apply (subgoal_tac "no_right_constr a")
apply (subgoal_tac "no_right_constr x")
apply (simp add: no_right_constr_simplif)
apply (subgoal_tac "no_right_constr (nf_constr a)")
apply (simp only: no_right_setconstr_to_list)
apply (simp add: no_right_constr_nf_constr)
apply (simp add: no_right_affset_elim_simplif_switch)
done

lemma no_right_affset_decompose_simplif1:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose_simplif subst c) .no_right_constr  x)"
apply (simp add: affset_decompose_simplif_def)
apply (clarsimp)
apply (simp add: affset_decompose_def)
apply (simp add: decomp_setconstr_list_def)
apply (clarsimp)
apply (subgoal_tac "no_right_constr a")
apply (subgoal_tac "no_right_constr x")
apply (simp add: no_right_constr_simplif)
apply (subgoal_tac "no_right_constr (nf_constr a)")
apply (simp only: no_right_setconstr_to_list)
apply (simp add: no_right_constr_nf_constr)
apply (simp add: no_right_affset_elim_simplif_switch1)
done



lemma is_elementary_constr_affset_decompose_simplif:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst ;valid_compos_constr (interp_set_rep interp_affset) c \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose_simplif subst c) .is_elementary_constr   x)"

apply (simp add: affset_decompose_simplif_def)
apply (clarsimp)
apply (subgoal_tac "is_elementary_constr x")
prefer 2
apply (simp add: is_elementary_affset_decompose)
apply (case_tac x)
apply (simp)
apply (rule conjI)
apply (simp_all add: elementary_compos_impl_elementary_compo_simplif )
done


lemma is_elementary_constr_affset_decompose_simplif1:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst  \<rbrakk>\<Longrightarrow>
  (\<forall> x \<in> set (affset_decompose_simplif subst c) .is_elementary_constr   x)"

apply (simp add: affset_decompose_simplif_def)
apply (clarsimp)
apply (subgoal_tac "is_elementary_constr x")
prefer 2
apply (simp add: is_elementary_affset_decompose1)
apply (case_tac x)
apply (simp)
apply (rule conjI)
apply (simp_all add: elementary_compos_impl_elementary_compo_simplif )
done





(**************************************************************************)
(*                              affset_inter_constr                       *)
(**************************************************************************)



fun 
  no_binop :: "('n,'a) set_rep \<Rightarrow> bool"
  where
  "no_binop (Binop bop a b)=False"
  |"no_binop _ = True"

fun 
  affset_inter_set_rep_atom :: "('n,affset_tp) set_rep \<Rightarrow> ('n,affset_tp) set_rep \<Rightarrow>
  ('n,affset_tp) set_rep"
where 
  "affset_inter_set_rep_atom (Atom a) (Atom b) = (Atom (affset_inter_calc a b))" 



lemma no_binop_affset_inter_set_rep_atom:
"(no_vars_set_rep a \<and>no_compl_set_rep a \<and> no_binop a \<and> \<not> (is_bconst a) \<and>
  no_vars_set_rep b \<and>no_compl_set_rep b \<and> no_binop b \<and> \<not> (is_bconst b)) \<longrightarrow>
  no_binop (affset_inter_set_rep_atom a b)"
apply (case_tac a)
apply (simp_all)
apply (case_tac b)
apply (simp_all)
done

lemma no_vars_affset_inter_set_rep_atom:
"(no_vars_set_rep a \<and>no_compl_set_rep a \<and> no_binop a \<and> \<not> (is_bconst a) \<and>
  no_vars_set_rep b \<and>no_compl_set_rep b \<and> no_binop b \<and> \<not> (is_bconst b)) \<longrightarrow>
  no_vars_set_rep (affset_inter_set_rep_atom a b)"
apply (case_tac a)
apply (simp_all)
apply (case_tac b)
apply (simp_all)
done


lemma affset_inter_set_rep_atom_simp:
"(no_vars_set_rep a \<and>no_compl_set_rep a \<and> no_binop a \<and> \<not> (is_bconst a) \<and>
  no_vars_set_rep b \<and>no_compl_set_rep b \<and> no_binop b \<and> \<not> (is_bconst b)) \<longrightarrow>
  interp_set_rep (interp_affset) vi (affset_inter_set_rep_atom a b)=
  interp_set_rep (interp_affset) vi (Binop Inf a b)"
apply (case_tac a)

apply (simp_all)
apply (case_tac b)
apply (simp_all add: affset_inter_calc_correct)
done
fun
 affset_inter_set_rep :: "('n,affset_tp) set_rep \<Rightarrow> ('n,affset_tp) set_rep"
  where
  "affset_inter_set_rep (Atom a) = (Atom a)"
  |"affset_inter_set_rep (Bconst b)= (Bconst b)"
  |"affset_inter_set_rep (Binop Inf a b)=
  (let a' = affset_inter_set_rep a in
  (let b' = affset_inter_set_rep b in
  (affset_inter_set_rep_atom a' b')))"







lemma affset_inter_set_rep_is_atom:
"  

 no_vars_set_rep a2 \<and>   no_compl_set_rep a2\<and> no_bconst_in_binop a2 \<and>
        is_elementary_composition bop.Inf a2 \<and> \<not> (is_bconst a2)
       \<longrightarrow> (\<exists>c2. affset_inter_set_rep a2 = Atom c2)"
by (induct a2) (clarsimp)+







lemma affset_inter_set_rep_is_bconst:
"  

 no_vars_set_rep a2 \<and>   no_compl_set_rep a2\<and> no_bconst_in_binop a2 \<and>
        is_elementary_composition bop.Inf a2 \<and>  (is_bconst a2)
       \<longrightarrow> (\<exists>c2. affset_inter_set_rep a2 = Bconst  c2)"
by (induct a2) (clarsimp)+






lemma affset_inters_set_rep_correct_aux:
  "(no_vars_set_rep a \<and> (no_compl_set_rep a) \<and> (is_elementary_composition Inf a)) 
  \<and> no_bconst_in_binop a
  \<longrightarrow>
  interp_set_rep (interp_affset) vi (affset_inter_set_rep a)=interp_set_rep (interp_affset) vi a"
apply (induct a)
apply (simp_all)
apply (clarsimp)
apply (subgoal_tac " (\<exists> c1 . (affset_inter_set_rep a1)=(Atom c1))")
apply (subgoal_tac " (\<exists> c2 . (affset_inter_set_rep a2)=(Atom c2))")
apply (clarsimp)
apply (simp add: affset_inter_set_rep_atom_simp affset_inter_calc_correct)
apply (clarsimp simp add: affset_inter_set_rep_is_atom)+
done




lemma no_vars_interp_v1_v2:
  "(no_vars_set_rep a \<and>no_compl_set_rep a\<and> (is_elementary_composition Inf a) \<and>no_bconst_in_binop a) \<longrightarrow>
  interp_set_rep ai v1 a = interp_set_rep ai v2 a"
apply (induct a)
apply (simp_all)
done

lemma no_vars_interp_v1_v2_ [rule_format]:
  "(no_vars_set_rep a) \<longrightarrow>
  interp_set_rep ai v1 a = interp_set_rep ai v2 a"
apply (induct a)
apply (simp_all)
done




lemma affset_inters_set_rep_correct_all:
  "(no_vars_set_rep a \<and> (no_compl_set_rep a) \<and> (is_elementary_composition Inf a)) 
  \<and> no_bconst_in_binop  a \<Longrightarrow>
 (\<forall> vi . interp_set_rep  (interp_affset) vi  (affset_inter_set_rep a) =  interp_set_rep  (interp_affset) vi  a)"
by (clarsimp simp add: affset_inters_set_rep_correct_aux)


lemma affset_inters_set_rep_correct:
  "(vars a={} \<and> (no_compl_set_rep a) \<and> (is_elementary_composition Inf a)) 
  \<and> no_bconst_in_binop a
  \<longrightarrow>
  interp_set_rep (interp_affset) vi (affset_inter_set_rep a)=interp_set_rep (interp_affset) vi a"
apply (simp only: No_vars_set_rep)
apply (simp add:  affset_inters_set_rep_correct_aux)
done


fun
  affset_inters_constr ::"('n,affset_tp) setconstr\<Rightarrow>bool"
where
  "affset_inters_constr (Lesseq f g) = 
  (((affset_inter_set_rep f)=(Atom AffsetEmpty)) \<or> ((affset_inter_set_rep f)=(Bconst Bot)))"


lemma interp_empty_affset_empty_iff:
  "(interp_affset a = {}) = (a = AffsetEmpty)"
apply (induct a)
apply (fastforce simp add: affset_def)+
done

lemma interp_bconst_empty_bot_iff:
  "(interp_bconst b= {}) = (b = Bot)"
by (case_tac b)
 (simp_all)





lemma affset_inter_set_rep_bconst_or_atom[rule_format]:
  "(no_compl_set_rep a \<and> no_vars_set_rep a  \<and> is_elementary_composition Inf a \<and> no_bconst_in_binop a ) \<longrightarrow>
  (interp_set_rep (interp_affset) vi (affset_inter_set_rep a)  = {})=
  (((affset_inter_set_rep a) = (Atom AffsetEmpty)) \<or>
  (affset_inter_set_rep a) = (Bconst Bot))"
apply (rule impI)
apply (case_tac "\<not> (is_bconst a)")
apply (subgoal_tac "(\<exists> a'. (affset_inter_set_rep a)=(Atom a'))")
apply (clarsimp)
apply (simp add: interp_empty_affset_empty_iff)
apply (simp add: affset_inter_set_rep_is_atom)
apply (subgoal_tac "(\<exists> b. (affset_inter_set_rep a)=(Bconst b))")
prefer 2
apply (simp add: affset_inter_set_rep_is_bconst)
apply (clarsimp simp add: interp_bconst.simps interp_bconst_empty_bot_iff)
done



lemma affset_inter_set_rep_bconst_or_atom1[rule_format]:
  "\<lbrakk>no_compl_set_rep a ; no_vars_set_rep a  ; is_elementary_composition Inf a ; no_bconst_in_binop a \<rbrakk> \<Longrightarrow>
  (interp_set_rep (interp_affset) vi (affset_inter_set_rep a)  = {})=
  (((affset_inter_set_rep a) = (Atom AffsetEmpty)) \<or>
  (affset_inter_set_rep a) = (Bconst Bot))"
by (simp add: affset_inter_set_rep_bconst_or_atom)




lemma interp_affset_inters_constr[rule_format]:
  "\<lbrakk>no_right_constr c \<and> vars_constr c = {} ; is_elementary_constr c ; no_bconst_in_binop_constr c ;no_compl_constr c\<rbrakk>\<Longrightarrow>
 ( (affset_inters_constr c) =  ((interp_constr (interp_affset) vi c)))"
apply (case_tac c)
apply (simp  only:)
apply (simp  del: vars.simps)
apply (simp only :No_vars_set_rep)

apply (clarsimp)
apply (subgoal_tac "(interp_set_rep (interp_affset) vi set_rep1)=(interp_set_rep (interp_affset) vi (affset_inter_set_rep set_rep1))")
prefer 2
apply (simp add: affset_inters_set_rep_correct_aux)
apply (clarsimp)
apply (simp add: affset_inter_set_rep_bconst_or_atom)
done


lemma interp_valid_eq[rule_format]:
  " vars_constr c = {}\<longrightarrow>
  interp_constr ai vi c = valid_constr ai c"
apply (induct c)
apply (simp del: vars.simps)
apply (simp add: valid_constr_def del: vars.simps)
apply (simp only: No_vars_set_rep)
apply (clarsimp)
apply (rule iffI)
apply (clarsimp)
apply (subgoal_tac "interp_set_rep ai vi set_rep1  =interp_set_rep ai via set_rep1 ")
apply (subgoal_tac "interp_set_rep ai vi set_rep2  =interp_set_rep ai via set_rep2 ")
apply (simp)
apply (rule no_vars_interp_v1_v2_)
apply (assumption)
apply (rule no_vars_interp_v1_v2_)
apply (assumption)
apply (simp)
done



lemma valid_affset_inters_constr:
  "(no_right_constr c \<and> vars_constr c = {} \<and> is_elementary_constr c \<and> no_bconst_in_binop_constr c \<and> no_compl_constr c )\<Longrightarrow>
  valid_constr (interp_affset) c = affset_inters_constr c "
apply (clarsimp)
apply (induct c)
apply (simp del: vars.simps)
apply (simp only: No_vars_set_rep)
apply (simp add: valid_constr_def)
apply (clarsimp)
apply (rule iffI)
prefer 2
apply (clarsimp)
apply (subgoal_tac "interp_set_rep (interp_affset) vi set_rep1=
interp_set_rep (interp_affset) vi (affset_inter_set_rep set_rep1)")
apply (simp)
apply (simp add:  affset_inter_set_rep_bconst_or_atom)
apply (simp add: affset_inters_set_rep_correct_aux)
apply (frule  allE)
apply (simp)

apply (subgoal_tac "
(interp_set_rep interp_affset (?x6 set_rep1) (affset_inter_set_rep set_rep1) = {})=
(affset_inter_set_rep set_rep1 = (Atom AffsetEmpty) \<or> affset_inter_set_rep set_rep1 =( Bconst Bot))")
prefer 2
apply (simp add:   affset_inter_set_rep_bconst_or_atom1)
apply (simp)
apply (subgoal_tac "(interp_set_rep interp_affset (?x15 set_rep1 set_rep1) (affset_inter_set_rep set_rep1) = {})")
apply (fastforce)
apply (subgoal_tac "(interp_set_rep interp_affset (?x31 set_rep1 set_rep1 set_rep1) (affset_inter_set_rep set_rep1)) =
  (interp_set_rep interp_affset (?x31 set_rep1 set_rep1 set_rep1) set_rep1)")
apply (simp)
apply (simp add: affset_inters_set_rep_correct_aux)
done
(**********************************************************************)
(*                           Solver                                   *)
(**********************************************************************)


fun
  affset_inters_constr_list :: "('n,affset_tp) setconstr list \<Rightarrow>bool"
where
  "affset_inters_constr_list [] = True"
| "affset_inters_constr_list (t#r) = ((affset_inters_constr t)\<and> (affset_inters_constr_list r))"




lemma valid_affset_inters_constr_list:
"(\<forall> x \<in> set l . ((vars_constr  x={})  \<and>no_bconst_in_binop_constr x \<and> no_compl_constr x 
  \<and> is_elementary_constr  x \<and> no_right_constr x)\<and>
  valid_constr (interp_affset) x) \<longrightarrow> affset_inters_constr_list l"
apply (induct l)
apply (simp)
apply (clarsimp)
apply (fastforce simp add: valid_affset_inters_constr)
done



lemma valid_affset_inters_constr_list_iff:
"(\<forall> x \<in> set l . ((vars_constr  x={})  \<and>no_bconst_in_binop_constr x \<and> no_compl_constr x 
  \<and> is_elementary_constr  x \<and> no_right_constr x))\<longrightarrow>
  affset_inters_constr_list l  = (\<forall> x \<in> set l . (valid_constr (interp_affset) x)) "
apply (intro impI iffI)
apply (induct l)
apply (simp)
apply (clarsimp)
apply (fastforce simp add: valid_affset_inters_constr)+
apply (simp add: valid_affset_inters_constr_list)
done


definition
solver ::
"( ('n \<Rightarrow> ('n,affset_tp) set_rep  option) \<Rightarrow> ('n,affset_tp) set_rep  \<Rightarrow> ('n,affset_tp) set_rep ) \<Rightarrow>('n,('n,affset_tp) set_rep) compos_constr \<Rightarrow> bool"
where
  "solver subst c = affset_inters_constr_list (affset_decompose_simplif subst c)"





lemma valid_solver_iff[rule_format]:
 "\<lbrakk>\<forall> vari. subst_invariant (interp_set_rep interp_affset) vari subst \<rbrakk> \<Longrightarrow>valid_compos_constr (interp_set_rep interp_affset) c = solver subst c"
apply (simp add: solver_def)
apply (rule iffI)
apply (subgoal_tac "(\<forall> x \<in> set (affset_decompose_simplif subst c).valid_constr (interp_affset) x)")
apply (subgoal_tac "(\<forall> x \<in> set (affset_decompose_simplif subst c) .
(vars_constr x =  {}) \<and> no_compl_constr x \<and> no_bconst_in_binop_constr x \<and> is_elementary_constr x \<and>no_right_constr x
\<and> valid_constr (interp_affset) x)")
apply (simp_all  add: valid_affset_inters_constr_list)
prefer 2

apply (simp add: valid_affset_decompose_simplif)
apply (clarsimp)
apply (simp_all  add: valid_affset_decompose_simplif no_right_affset_decompose_simplif1 
   is_elementary_constr_affset_decompose_simplif1  no_bconst_in_binop_affset_decompose_simplif1 
   no_compl_affset_decompose_simplif1 no_vars_affset_decompose_simplif1 valid_affset_inters_constr_list_iff)
done






lemma subst_inv_subst_set_rep:
 "(\<forall> vi .subst_invariant (interp_set_rep interp_affset) vi subst_set_rep)"
apply (clarsimp simp add: subst_invariant_def)
apply (induct_tac e)
apply simp_all
apply (simp add: override_fun_with_map_def split add: option.split)
done 

definition 
  "solve c == solver (subst_set_rep) c"


lemma valid_solve:
  "solve c = valid_compos_constr (interp_set_rep interp_affset) c"
apply (simp add: solve_def)
apply (rule  valid_solver_iff [THEN sym])
apply (simp_all add: subst_inv_subst_set_rep)
done


end
