
theory Clockcalc imports Canonlang Clocklang

begin


  (* ------------------------------------------------------------ *)
subsection {* Extraction of clock expressions / statements *}

text {*
  Extraction for canonized expressions
*}

definition
  clock_combin :: "['n, ('n,'v) canon_expr] \<Rightarrow> ('n,'a) clstmt"
where  "clock_combin y e == 
  clConjs (map (\<lambda> x. (ClEq (ClOf y) (ClOf x))) (fv_canon_expr e))"

primrec
  clForall_closure :: "[('n * ('n, 'a) clexpr) list, ('n,'a) clstmt] \<Rightarrow> ('n,'a) clstmt"
where 
  "clForall_closure [] c = c" |
  "clForall_closure (p # ps) c = 
    (let (par, e) = p in 
     (ClLet par e (clForall_closure ps c)))"

text {*
  For a clock specification with parameters (p1 .. pn)  and body c
  and an argument list (a1 .. an), construct the formula
  \<forall> p1. p1 = a1 \<longrightarrow> \<dots> \<forall> pn.  pn = an \<longrightarrow> c
  This is equivalent to carrying out a substitution: c[p1:=a1, .. pn:=an]
*}
primrec
  clock_param_form :: "[('n,'a) clspec, 'n list] \<Rightarrow> ('n,'a) clstmt"
where 
  "clock_param_form (ClSpec params c) args = 
  clForall_closure (zip params (map ClOf args)) c"


lemma clock_param_form_clForall_closure: "length parexprs = length args \<Longrightarrow>
  (clock_param_form (reduce_to_clspec (ClSpecEq parexprs)) args) = 
  (clForall_closure (zip (map fst parexprs) (map ClOf args))
                 (clConjs (map (\<lambda>(p, y). ClEq (ClOf p) y) parexprs)))"
by simp


primrec
  clock_canon_stmt :: 
  "['n \<Rightarrow> 'a, ('n,'a) clock_env, 
        ('n,'v, (('n,'a) clexpr) inter_typ) canon_stmt] \<Rightarrow> ('n,'a) clstmt"
where
  "clock_canon_stmt abstr clenv CanEmptyStmt = ClTrue"

| "clock_canon_stmt abstr clenv (CanCombin y e) = clock_combin y e"

| "clock_canon_stmt abstr clenv (CanPre y v x) = (ClEq (ClOf y) (ClOf x))"

| "clock_canon_stmt abstr clenv (CanWhen y x b) = 
        ClEq (ClOf y) (ClInter (ClOf x) (ClWhen (abstr b)))"

| "clock_canon_stmt abstr clenv (CanDefault y x z) = 
        ClEq (ClOf y) (ClUnion (ClOf x) (ClOf z))"

| "clock_canon_stmt abstr clenv (CanPar c c') =
        ClConj (clock_canon_stmt abstr clenv c) (clock_canon_stmt abstr clenv c')"

| "clock_canon_stmt abstr clenv (CanPCall cn args) =
     clock_param_form (reduce_to_clspec (the (clenv (cname_header cn)))) args"

| "clock_canon_stmt abstr clenv (CanLetv v clk c) =
    ClLet v (get_inter_typ_clocktp clk) (clock_canon_stmt abstr clenv c)"


lemma dom_interp_expr_fv_INT: "
  dom (interp_canon_expr error_fns_flat tr e) = 
  (\<Inter>x\<in>set (fv_canon_expr e). dom (trace_proj tr x))"
by (induct e) (simp_all add: INT_Un)


lemma dom_interp_canon_expr_fv_UN: "
  dom (interp_canon_expr error_fns_lifted tr e)
  = (\<Union>x\<in>set (fv_canon_expr e). dom (trace_proj tr x))"
by (induct e) (simp_all add: INT_Un)

(*
lemma dom_interp_expr_sub_expr_INT: "
  dom (interp_expr tr (emb_canon_expr e)) = 
  (\<Inter>e'\<in>set (canon_sub_expr e). dom (interp_expr tr (emb_canon_expr e')))"
apply (induct e)
apply (simp_all add: INT_Un)
*)

lemma dom_trace_proj [rule_format]:
  "error_free_signal
         (interp_canon_expr error_fns_lifted tr e)
  \<longrightarrow>
  (\<forall> x \<in> set (fv_canon_expr e). 
  dom (trace_proj tr x) = dom (interp_canon_expr error_fns_lifted tr e))"
apply (induct e)
apply simp
apply clarsimp
apply (frule error_free_signal_signalpair_error)
apply (frule error_free_signal_signalpair_dom)
apply clarsimp

apply (erule disjE)
apply simp_all
done

lemma error_free_signal_impl_sync_vars:
  "error_free_signal
         (interp_canon_expr error_fns_lifted tr e)
  \<Longrightarrow> sync_vars tr (set (fv_canon_expr e))"
by (clarsimp simp add: sync_vars_def sync_signals_def)
   (blast dest: dom_trace_proj)


lemma sync_signals_impl_error_free_signal [rule_format]:
  "sync_signals (interp_canon_expr error_fns_lifted tr ` 
  (set (canon_sub_expr e))) \<longrightarrow>
  (\<forall> e' \<in> set (canon_sub_expr e). 
      error_free_signal (interp_canon_expr error_fns_lifted tr e'))"
apply (induct e)
apply (simp add: sync_signals_def)

apply clarify
apply (clarsimp)
apply (drule sync_signals_insertD)

apply (clarsimp simp add: image_Un)
apply (frule sync_signals_unionD, simp)
apply (fastforce intro: error_free_signal_sync_signals sync_signals_subset)

apply clarsimp
apply (drule sync_signals_insertD)
apply simp
done


lemma sync_vars_sync_signals [rule_format]:
  "sync_vars tr (set (fv_canon_expr e))
  \<longrightarrow>  sync_signals (interp_canon_expr error_fns_lifted tr ` 
  (set (canon_sub_expr e)))"
apply (simp add: sync_signals_def)
apply (clarsimp simp add: dom_interp_canon_expr_fv_UN)
apply (simp add: sync_vars_def sync_signals_def)
apply (subgoal_tac "\<forall> e' \<in> set (canon_sub_expr e). set (fv_canon_expr e') \<subseteq> set (fv_canon_expr e)")
apply (subgoal_tac "\<exists> v. v \<in> set (fv_canon_expr s)")
apply (subgoal_tac "\<exists> v'. v' \<in> set (fv_canon_expr s')")
apply (blast dest: bspec)
apply (rule fv_canon_expr_productive)+
apply (rule ballI)
apply (erule fv_canon_sub_expr)
done

lemma sync_vars_impl_error_free_signal:
  "sync_vars tr (set (fv_canon_expr e)) \<Longrightarrow>
  error_free_signal
         (interp_canon_expr error_fns_lifted tr e)"
apply (frule sync_vars_sync_signals)
apply (erule sync_signals_impl_error_free_signal)
apply simp
done

lemma error_free_signal_sync_vars:
  "error_free_signal
         (interp_canon_expr error_fns_lifted tr e) = 
  sync_vars tr (set (fv_canon_expr e))"
by (fast intro: error_free_signal_impl_sync_vars sync_vars_impl_error_free_signal)


  (* ------------------------------------------------------------ *)
subsection {* Clock interpretation as safety check *}

  (* 
  The equality  between the interpretations of the specification and the
  type of the body is required:
  direction \<longrightarrow> for interp_clstmt_remap_fun_clock_param_form
  direction \<longleftarrow> for interp_clstmt_clock_param_form_remap_fun
 *)

  (* the parameter 'concr' has been added to the def of wf_clspec 
  in order to make wf_clspec explicitly depend on the type parameter 't, 
  which otherwise occurs only on the RHS of the def., if concr is 
  universally quantified as in (\<forall> concr abstr tr. \<dots>)
  Therefore, this parameter has to be dragged around in other
  definitions (wf_clock_env, \<dots>)
 *)

definition
  wf_clspec :: 
  "[('t::order,'n,'v) trace \<Rightarrow> 'a \<Rightarrow> 't set, 'n \<Rightarrow> 'a, 
    ('n, 'a) clock_env, 'n list, 
    ('n,'v, (('n,'a) clexpr) inter_typ) canon_stmt, ('n, 'a) clspec] 
  \<Rightarrow> bool" where
 "wf_clspec concr abstr clenv pars bd clsp =
  (\<exists> clbd. clsp = (ClSpec pars clbd) \<and> 
     set (fv_clstmt clbd) \<subseteq> set pars \<and> 
    (\<forall> tr. 
      ((interp_clstmt tr concr clbd) \<longrightarrow>
       (interp_clstmt tr concr (clock_canon_stmt abstr clenv bd)))))"


  (* specialized predicate for clock specs of the form ClSpecEq *)
primrec
  wf_clspeceq :: 
  "[('t::order,'n,'v) trace \<Rightarrow> 'a \<Rightarrow> 't set, 'n \<Rightarrow> 'a,
    ('n, 'a) clock_env, 'n list, 
    ('n,'v, (('n,'a) clexpr) inter_typ) canon_stmt, ('n, 'a) clspec] 
  \<Rightarrow> bool" where
"wf_clspeceq concr abstr clenv pars bd (ClSpecEq parexprs) = 
  (pars = map fst parexprs \<and> 
   (\<forall> e \<in> set (map snd parexprs).  (set (fv_clexpr e)) \<subseteq> set pars) \<and> 
    (\<forall> tr. 
     (interp_clstmt tr concr (clForall_closure parexprs (clock_canon_stmt abstr clenv bd)))))"


(*
lemma "interp_clstmt tr concr (clForall_closure parexprs c) = 
((interp_clstmt tr concr clbd) \<longrightarrow>
       (interp_clstmt tr concr (clock_canon_stmt abstr clenv bd)))"

*)


lemma wf_clspeceq_impl_wf_clspec_aux [rule_format]: "interp_clstmt tr concr (clForall_closure parexprs c)
\<longrightarrow> ((\<forall>x\<in>set parexprs.
               dom (trace_proj tr (fst x)) = interp_clexpr tr concr (snd x)) \<longrightarrow>
           interp_clstmt tr concr c)"
apply (induct parexprs)
apply simp
apply clarsimp
apply (drule spec) apply (drule mp) apply assumption
apply clarsimp 
done


lemma wf_clspeceq_impl_wf_clspec:
  "wf_clspeceq concr abstr clenv pars bd (ClSpecEq parexprs) \<Longrightarrow>
  wf_clspec concr abstr clenv pars bd (reduce_to_clspec (ClSpecEq parexprs))"
apply (simp add: wf_clspec_def)
apply clarsimp
apply (clarsimp simp add: split_def)
apply (rule conjI) apply fast
apply clarsimp
apply (fast intro: wf_clspeceq_impl_wf_clspec_aux)
done


definition
  wf_clock_env :: 
  "[('t::order,'n,'v) trace \<Rightarrow> 'a \<Rightarrow> 't set, 'n \<Rightarrow> 'a,
    ('n,'v, (('n,'a) clexpr) inter_typ) canon_proc_env, ('n, 'a) clock_env] 
  \<Rightarrow> bool"
  where "wf_clock_env concr abstr cpenv clenv =
  (\<forall> pn pn' tpars bd. 
     cpenv pn = Some (CanProc pn' tpars bd) \<longrightarrow> 
         (\<exists> clsp. clenv pn = Some clsp \<and> 
                  wf_clspec concr abstr clenv (var_types_names tpars) bd (reduce_to_clspec clsp)))"

definition
  wf_clock_env_proc_list :: 
  "[('t::order,'n,'v) trace \<Rightarrow> 'a \<Rightarrow> 't set, 'n \<Rightarrow> 'a,
    ('n,'v, (('n,'a) clexpr) inter_typ) canon_proc list, ('n, 'a) clock_env] 
  \<Rightarrow> bool"
where "wf_clock_env_proc_list concr abstr procs clenv =
  (list_all (\<lambda> pr. case pr of (CanProc pn tpars bd) \<Rightarrow> 
                  (\<exists> clsp. clenv pn = Some clsp \<and> 
                    wf_clspec concr abstr clenv (var_types_names tpars) bd (reduce_to_clspec clsp)))
     procs)"



definition
  dom_indifferent :: "(('t,'n, 'v) trace \<Rightarrow> 'a \<Rightarrow> 't set) \<Rightarrow> bool"
where  "dom_indifferent concr == \<forall> tr tr'. 
  (\<forall>x. dom (trace_proj tr x) = dom (trace_proj tr' x)) \<longrightarrow>
        concr tr = concr tr'"



lemma dom_indifferent_error_fns_flat [rule_format]:
  "(\<forall>x. dom (trace_proj tr x) = dom (trace_proj tr' x)) \<longrightarrow>
        dom (interp_canon_expr error_fns_flat tr e) =
        dom (interp_canon_expr error_fns_flat tr' e)"
by (induct e) simp_all


lemma dom_indifferent_error_fns_lifted [rule_format]:
  "(\<forall>x. dom (trace_proj tr x) = dom (trace_proj tr' x)) \<longrightarrow>
        dom (interp_canon_expr error_fns_lifted tr e) =
        dom (interp_canon_expr error_fns_lifted tr' e)"
by (induct e) simp_all


lemma interp_clexpr_eq_dom:
 "\<lbrakk> \<forall> x. dom (trace_proj tr x) = dom (trace_proj tr' x);
  dom_indifferent concr \<rbrakk>
  \<Longrightarrow> interp_clexpr tr concr e = interp_clexpr tr' concr e"
apply (induct e)
apply simp_all
apply (simp add: dom_indifferent_def fun_eq_iff)
done


lemma interp_clstmt_eq_dom [rule_format]: "
  dom_indifferent concr  
  \<Longrightarrow> 
  \<forall> tr tr'. 
  (\<forall> x. dom (trace_proj tr x) = dom (trace_proj tr' x))
  \<longrightarrow> interp_clstmt tr concr c = interp_clstmt tr' concr c"
apply (induct c)
apply simp

  (* ClEq *)
apply clarsimp
apply (rule iffI)
apply (fast dest: interp_clexpr_eq_dom)
apply (fast dest: interp_clexpr_eq_dom)

  (* ClConj *)
apply clarsimp+
apply blast

  (* ClImpl
apply clarsimp
apply (rule iffI)
apply clarsimp
apply (subgoal_tac "(\<forall>x. dom (trace_proj (trace_upd tr n s) x) =
             dom (trace_proj (trace_upd tr' n s) x))")
apply blast apply simp
apply clarsimp 
apply (subgoal_tac "(\<forall>x. dom (trace_proj (trace_upd tr n s) x) =
             dom (trace_proj (trace_upd tr' n s) x))")
apply blast apply simp
 *)
  (* ClForall
apply clarsimp
apply (rule iffI)
apply clarsimp
apply (subgoal_tac "(\<forall>x. dom (trace_proj (trace_upd tr n s) x) =
             dom (trace_proj (trace_upd tr' n s) x))")
apply blast apply simp
apply clarsimp 
apply (subgoal_tac "(\<forall>x. dom (trace_proj (trace_upd tr n s) x) =
             dom (trace_proj (trace_upd tr' n s) x))")
apply blast apply simp
 *)

  (* ClLet *)
apply clarsimp
apply (rule iffI)
apply clarsimp
apply (drule_tac x=s in spec)
apply (simp add: interp_clexpr_eq_dom)
apply (subgoal_tac "(\<forall>x. dom (trace_proj (trace_upd tr n s) x) =
             dom (trace_proj (trace_upd tr' n s) x))")
apply blast apply simp
apply clarsimp 
apply (drule_tac x=s in spec)
apply (simp add: interp_clexpr_eq_dom)
apply (subgoal_tac "(\<forall>x. dom (trace_proj (trace_upd tr n s) x) =
             dom (trace_proj (trace_upd tr' n s) x))")
apply blast apply simp
done

lemma interp_clstmt_clForall_closure [rule_format]:
  "\<lbrakk> length pars = length args; dom_indifferent concr \<rbrakk>
  \<Longrightarrow> distinct pars \<longrightarrow> (set args \<inter> set pars = {}) \<longrightarrow>
      (\<forall> tr. interp_clstmt
         (\<lambda>a. remap_fun (tr a) pars args) concr
         clbd
       \<longrightarrow> interp_clstmt tr concr
           (clForall_closure (zip pars (map ClOf args)) clbd))"
apply (induct rule: list_induct2)
apply (simp add: remap_fun_def)

apply (clarsimp simp add: remap_fun_def)
apply (drule spec, drule mp) prefer 2 apply assumption

apply (rule interp_clstmt_eq_dom [THEN iffD1]) apply assumption
defer
apply assumption

apply (simp add: trace_upd_def fun_upd_map_upd [THEN sym])
apply (simp add: trace_proj_fun_upd_map)
apply (simp add: trace_proj_def)
done

lemma interp_clstmt_remap_fun [rule_format]:
  "\<lbrakk> length pars = length args; dom_indifferent concr \<rbrakk>
  \<Longrightarrow> distinct pars \<longrightarrow> (set args \<inter> set pars = {}) \<longrightarrow>
      (\<forall> tr. interp_clstmt tr concr
           (clForall_closure (zip pars (map ClOf args)) clbd)
          \<longrightarrow> interp_clstmt
              (\<lambda>a. remap_fun (tr a) pars args) concr clbd)"
apply (induct rule: list_induct2)
apply (simp add: remap_fun_def)

apply (clarsimp simp add: remap_fun_def)
apply (drule spec, drule mp) apply (rule refl)
apply (drule spec, drule mp) apply assumption
apply (rule interp_clstmt_eq_dom [THEN iffD1]) apply assumption
prefer 2 apply assumption

apply (simp add: trace_upd_def fun_upd_map_upd [THEN sym])
apply (simp add: trace_proj_fun_upd_map)
apply (simp add: trace_proj_def)
done



lemma interp_clstmt_clock_param_form_remap_fun:
  "\<lbrakk>
  interp_clstmt tr concr (clock_param_form  (reduce_to_clspec (the (clenv pn))) args);
  (CanProc pn tpars bd) \<in> set procs;
  wf_clock_env_proc_list concr abstr procs clenv;
  dom_indifferent concr;
  length tpars = length args;
  distinct (var_types_names tpars);
  set args \<inter> set (var_types_names tpars) = {} \<rbrakk>
  \<Longrightarrow> interp_clstmt
         (\<lambda>a. remap_fun (tr a) (var_types_names tpars) args) concr
         (clock_canon_stmt abstr clenv bd)"
apply (clarsimp simp add: wf_clock_env_proc_list_def list_all_iff)
apply (drule bspec) apply assumption
(* apply ((drule spec)+, drule mp, assumption)*)
apply (clarsimp simp add: wf_clspec_def)

apply (drule spec)+ apply (drule mp) prefer 2 apply assumption
apply (rule interp_clstmt_remap_fun)
apply simp+
done


  (* ------------------------------------------------------------ *)
subsection {* Clock interpretation ensures error-freeness *}

text {* Show that satisfiability of a clock statement ensures error-freeness of all expressions in the statement. Consequence: clock well-typed statements can use error-free semantics
 *}

lemma interp_clstmt_clock_combin: "interp_clstmt tr concr (clock_combin n e) \<Longrightarrow>
  error_free_signal (interp_canon_expr error_fns_lifted tr e)"
apply (clarsimp simp add: error_free_signal_sync_vars)
apply (simp add: clock_combin_def sync_vars_def)
apply (clarsimp simp add: sync_signals_def)
apply blast
done

lemma interp_canon_stmt_Blocking_imp_Permissive:
  "interp_canon_stmt Blocking tr penv tpi c 
  \<Longrightarrow> interp_canon_stmt Permissive tr penv tpi c"
apply (induct rule: interp_canon_stmt.induct)
apply simp_all
apply fastforce
done


  (* ------------------------------------------------------------ *)


  (* from List.thy *)
lemma map_eq_imp_length_eq:
  assumes "map f xs = map g ys"
  shows "length xs = length ys"
using assms proof (induct ys arbitrary: xs)
  case Nil then show ?case by simp
next
  case (Cons y ys) then obtain z zs where xs: "xs = z # zs" by auto
  from Cons xs have "map f zs = map g ys" by simp
  moreover with Cons have "length zs = length ys" by blast
  with xs show ?case by simp
qed


lemma wf_clock_env_solve_impl_wf_clock_env_aux [rule_format]:
  " map_of (map (\<lambda>pr. (get_can_proc_header pr, pr)) procs) pn = Some (CanProc pn' tpars bd)
  \<longrightarrow> pn = pn'"
by (induct procs) auto


lemma clock_type_soundness_aux [rule_format]: " 
  interp_canon_stmt efs tr cpenv tpi cc
  \<Longrightarrow>
  cpenv = (constr_canon_proc_env procs) \<longrightarrow>
  dom_indifferent concr \<longrightarrow>
  wf_clock_env_proc_list concr abstr procs clenv  \<longrightarrow>
  wtpd_canon_proc_list ienv procs \<longrightarrow> 
  efs = Permissive \<longrightarrow> 
  (\<exists> tenv. wtpd_canon_stmt ienv tenv cc) \<longrightarrow>
  interp_clstmt tr concr (clock_canon_stmt abstr clenv cc) \<longrightarrow> 
  tpi = (\<lambda> tr' clk. interp_clexpr tr' concr (get_inter_typ_clocktp clk)) \<longrightarrow>
        interp_canon_stmt Blocking tr cpenv tpi cc"
apply (induct rule: interp_canon_stmt.induct)

-- CanEmptyStmt
apply simp

-- CanCombin
apply (clarsimp simp add: interp_clstmt_clock_combin)

-- CanPre
apply simp

-- CanWhen
apply simp

-- CanDefault
apply simp

-- CanPar
apply fastforce


prefer 2
-- CanLetv
apply fastforce


-- PCall
apply clarsimp
apply (subgoal_tac "(\<exists>tenv. wtpd_canon_stmt ienv tenv bd)")


apply (frule wf_clock_env_solve_impl_wf_clock_env_aux) 
apply simp

apply (subgoal_tac "(CanProc pn' tpars bd) \<in> set procs")

apply (frule interp_clstmt_clock_param_form_remap_fun)
  apply assumption (* prefer 7 *) prefer 6 apply simp (* TODO: var_types_names vs. map get_var_types_name *)apply fast 
  apply simp_all
  apply (clarsimp split add: proc_header.split_asm option.split_asm interf.split_asm)
  apply (rule sym) apply (rule map_eq_imp_length_eq) apply (rule trans) apply assumption

(*
  apply (fastforce dest: wtpd_canon_proc_env_distinct)
apply (simp add: set_empty [THEN sym])
apply (fast intro: wtpd_canon_proc_env_wtpd_canon_stmt)
done
*)
sorry

  (* should be the precondition for the realizability statement *)
definition
  concr_abstr_correspond :: 
  "(('t::order,'n,'v::vals_cl) trace \<Rightarrow> 'a \<Rightarrow> 't set) \<Rightarrow> ('n \<Rightarrow> 'a) \<Rightarrow> bool"
where  "concr_abstr_correspond concr abstr ==
   (\<forall> tr x. (concr tr (abstr x)) =
            {t . t \<in> dom (trace_proj tr x) \<and> true_val (the (trace_proj tr x t))})"

theorem clock_type_soundness:
  "\<lbrakk>wtpd_canon_proc_list ienv procs; wf_clock_env_proc_list concr abstr procs clenv; 
  wtpd_canon_stmt ienv tenv c;
  dom_indifferent concr;
  cpenv = (constr_canon_proc_env procs);
  interp_clstmt tr concr (clock_canon_stmt abstr clenv c);
  tpi = (\<lambda> tr' clk. interp_clexpr tr' concr (get_inter_typ_clocktp clk))
  \<rbrakk>
  \<Longrightarrow> interp_canon_stmt Blocking tr cpenv tpi c
    = interp_canon_stmt Permissive tr cpenv tpi c"
by (fast intro: interp_canon_stmt_Blocking_imp_Permissive
  dest: clock_type_soundness_aux)

end
