theory Clocksolver imports Clockcalc External
 Efficient_Nat 
  "../Setsolvers/Gencode"

begin

type_synonym  ('n,'a) clock_env_local = "'n \<Rightarrow> ('n,'a) clexpr option"

type_synonym  ('n,'a) set_rep_env_local = "'n \<Rightarrow> ('n,'a) set_rep option"

primrec
  cnConjs :: "('n,'a) compos_constr list \<Rightarrow> ('n,'a) compos_constr"
where
  "cnConjs [] = CnTrue" |
  "cnConjs (c # cs) = CnConj c (cnConjs cs)"

primrec
  extract_set_rep :: "('n,'a) clexpr \<Rightarrow> ('n, 'a) set_rep"
where
   "extract_set_rep (ClOf n) = (Var n)"
|  "extract_set_rep (ClWhen a) = (Atom a)"
|  "extract_set_rep ClTop = Bconst Top"
|  "extract_set_rep (ClInter e e') = (Binop Inf (extract_set_rep e) (extract_set_rep e'))"
|  "extract_set_rep (ClUnion e e') = (Binop Sup (extract_set_rep e) (extract_set_rep e'))"
|  "extract_set_rep (ClCompl e) = (Comp (extract_set_rep e))"


definition
  clock_combin_set_rep :: "['n \<Rightarrow> 'a option, 'n, ('n,'v) canon_expr] \<Rightarrow> ('n, 'a) compos_constr" where
  "clock_combin_set_rep lenv y e == 
  cnConjs (map (\<lambda> x. (CnEq ((the (lenv y))) ((the (lenv x))))) (fv_canon_expr e))"


primrec
  canon_stmt_to_set_rep :: 
  "[('n, 'a) clock_env, ('n,'a) set_rep_env_local, ('n,'v, (('n,'a) clexpr) inter_typ) canon_stmt]
  \<Rightarrow> ('n, ('n, 'a) set_rep) compos_constr"
where
  "canon_stmt_to_set_rep genv lenv CanEmptyStmt = CnTrue"

| "canon_stmt_to_set_rep genv lenv (CanCombin y e) = clock_combin_set_rep lenv y e"

| "canon_stmt_to_set_rep genv lenv (CanPre y v x) = (CnEq ((the (lenv y))) ( (the (lenv x))))"

| "canon_stmt_to_set_rep genv lenv (CanWhen y x b) = 
        CnEq ( (the (lenv y))) (Binop Inf ( (the (lenv x))) (Var b))"

| "canon_stmt_to_set_rep genv lenv (CanDefault y x z) = 
        CnEq ( (the (lenv y))) (Binop Sup ((the (lenv x))) ( (the (lenv z))))"

| "canon_stmt_to_set_rep genv lenv (CanPar c c') =
        CnConj (canon_stmt_to_set_rep genv lenv c) (canon_stmt_to_set_rep genv lenv c')"

| "canon_stmt_to_set_rep genv lenv (CanLetv v clk c) =
  (let sr = (extract_set_rep (get_inter_typ_clocktp clk)) in 
    CnLet v sr (canon_stmt_to_set_rep genv (lenv(v\<mapsto>sr)) c))"

term map_of

definition parexprs_to_clock_env_local :: "('n * (('n,'a) clexpr)) list \<Rightarrow> ('n,'a) set_rep_env_local"
where
"parexprs_to_clock_env_local parexprs = map_of (map (\<lambda> (x, e). (x, extract_set_rep e)) parexprs)"


  (* genv: global clock environment *)
  (* slv is the solver *)
definition
  wf_clspeceq_solve :: 
  "[('n, 'a) clock_env, 'n list, 
    ('n,'v, (('n, 'a) clexpr) inter_typ) canon_stmt, ('n * (('n,'a) clexpr)) list,
  ('n, ('n, 'a) set_rep) compos_constr \<Rightarrow> bool] 
  \<Rightarrow> bool" where
  "wf_clspeceq_solve genv pars bd parexprs slv = 
  (pars = map fst parexprs \<and> 
   (list_all (\<lambda> e. list_subset (fv_clexpr e) pars) (map snd parexprs)) \<and> 
   (slv (canon_stmt_to_set_rep genv (parexprs_to_clock_env_local parexprs) bd)))"

definition
  wf_clock_env_solve :: 
  "[('n,'v, (('n, 'a) clexpr) inter_typ) canon_proc list, ('n, 'a) clock_env,
  ('n, ('n, 'a) set_rep) compos_constr \<Rightarrow> bool] 
  \<Rightarrow> bool"
where "wf_clock_env_solve procs clenv slv =
  (list_all (\<lambda> cp. 
     (case cp of (CanProc ph tpars bd) \<Rightarrow> 
     (case clenv ph of 
        None \<Rightarrow> False
      | Some clsp \<Rightarrow> 
        (case clsp of 
           ClSpec params c \<Rightarrow> False
         | ClSpecEq parexprs \<Rightarrow> 
                    wf_clspeceq_solve clenv (var_types_names tpars) bd parexprs slv))))
 procs)"


definition constr_clspeceq_from_var_types_list ::
    "('n, ('n, 'a) clexpr inter_typ) var_types list \<Rightarrow> ('n, 'a) clspec"
where
 "constr_clspeceq_from_var_types_list vts = 
  ClSpecEq (map (\<lambda> vt. (get_var_types_name vt, (get_inter_typ_clocktp (get_var_types_type vt))))  vts)"

definition constr_clock_env :: "('n,'v, ('n, 'a) clexpr inter_typ) canon_proc list \<Rightarrow> ('n,'a) clock_env"
where
  "constr_clock_env procs = 
  map_of 
  (map (\<lambda> pr. (get_can_proc_header pr, 
              (constr_clspeceq_from_var_types_list (get_can_proc_var_types pr))))
        procs)"



primrec wtpd_canon_program :: 
  "[('n, vals, (('n, affset_tp) clexpr) inter_typ) interf_env, ('n, vals, (('n, affset_tp) clexpr) inter_typ) canon_program] \<Rightarrow> bool"
where
  "wtpd_canon_program ienv (CanProgram procs c) = 
  (distinct (map get_can_proc_header procs) \<and>
   wtpd_canon_proc_list ienv procs \<and>
   wf_clock_env_solve procs (constr_clock_env procs) solve
)"


primrec wtpd_canon_program_ext :: 
  "[('n, vals, (('n, affset_tp) clexpr) inter_typ) canon_program_ext] \<Rightarrow> bool"
where
  "wtpd_canon_program_ext (CanProgramExt interfs procs c) = 
   (wtpd_canon_program (constr_interf_env interfs) (constr_canon_program (CanProgramExt interfs procs c)))"


export_code wtpd_canon_program_ext (* wtpd_canon_stmt_exec *) wtpd_canon_stmt wf_canon_program_ext
   map_of vals_tp fv_clstmt InterTyp
  in OCaml module_name Gencode file "../../Code/Gencode/gencode.ml"


end
