
theory Clocklang imports List_extended Stream_fcts Stretching

begin



subsection {* Clock expressions *}


datatype ('n,'a) clexpr
  = ClOf 'n
  | ClWhen 'a
  | ClTop
  | ClInter "(('n,'a) clexpr)" "(('n,'a) clexpr)"
  | ClUnion "(('n,'a) clexpr)" "(('n,'a) clexpr)"
  | ClCompl "(('n,'a) clexpr)"


datatype ('n,'a) clstmt
  = ClTrue
  | ClEq "(('n,'a) clexpr)" "(('n,'a) clexpr)"  
  | ClConj "(('n,'a) clstmt)" "(('n,'a) clstmt)"
  | ClLet 'n "(('n,'a) clexpr)" "(('n,'a) clstmt)"



primrec
  clConjs :: "('n,'a) clstmt list \<Rightarrow> ('n,'a) clstmt"
where
  "clConjs [] = ClTrue" |
  "clConjs (c # cs) = ClConj c (clConjs cs)"

text {*
  A clock specification is of the general form ClSpec 
  (with a list of variable names and a clock statement that may only contain these names)
  or of an equational form ClSpecEq [(v1, e1) ... (vn, en)], interpreted as
  ClOf(v1) = e1 \<and> ...  ClOf(vn) = en.
  Only the latter form can currently be handled in wf_clspec.
*}
datatype ('n,'a) clspec = 
   ClSpec "('n list)" "('n,'a) clstmt"
 | ClSpecEq "('n * (('n,'a) clexpr)) list"

primrec reduce_to_clspec :: "('n,'a) clspec \<Rightarrow> ('n,'a) clspec"
where
  "reduce_to_clspec (ClSpec params c) = ClSpec params c"
| "reduce_to_clspec (ClSpecEq parexprs) = 
      ClSpec (map fst parexprs) (clConjs (map (\<lambda> (p, e). ClEq (ClOf p) e) parexprs))"

type_synonym  ('n,'a) clock_env = "'n proc_header \<Rightarrow> ('n,'a) clspec option"


subsection {* Interpretation of clock expressions / statements *}

primrec
  interp_clexpr :: 
  "[('t,'n,'v) trace, ('t,'n,'v) trace \<Rightarrow> 'a \<Rightarrow> 't set, ('n,'a) clexpr]
    \<Rightarrow> 't set"
where
  "interp_clexpr tr concr (ClOf x) = dom (trace_proj tr x)"|
  "interp_clexpr tr concr (ClWhen a) = (concr tr a)"|
  "interp_clexpr tr concr ClTop = UNIV"|
  "interp_clexpr tr concr (ClInter ce ce') = 
     (interp_clexpr tr concr ce) \<inter> (interp_clexpr tr concr ce')"|
  "interp_clexpr tr concr (ClUnion ce ce') = 
     (interp_clexpr tr concr ce) \<union> (interp_clexpr tr concr ce')"|
  "interp_clexpr tr concr (ClCompl ce) = - (interp_clexpr tr concr ce)"

primrec
  interp_clstmt :: 
  "[('t::order, 'n, 'v) trace,  ('t,'n, 'v) trace \<Rightarrow> 'a \<Rightarrow> 't set, 
    ('n,'a) clstmt] \<Rightarrow> bool"
where
  "interp_clstmt tr concr ClTrue = True"|
  "interp_clstmt tr concr (ClEq e e') = 
     ((interp_clexpr tr concr e) = (interp_clexpr tr concr e'))"|
  "interp_clstmt tr concr (ClConj c c') = 
     (interp_clstmt tr concr c \<and> interp_clstmt tr concr c')"|
(*
  "interp_clstmt tr concr (ClImpl c c') = 
     (interp_clstmt tr concr c \<longrightarrow> interp_clstmt tr concr c')"
  "interp_clstmt tr concr (ClForall v c) = 
     (\<forall> s. interp_clstmt (trace_upd tr v s) concr c)"
*)
 "interp_clstmt tr concr (ClLet v e c) = 
  (\<forall>s. dom s = interp_clexpr tr concr e \<longrightarrow>
         interp_clstmt (trace_upd tr v s) concr c)"


lemma interp_clstmt_clConjs [simp]:
 "interp_clstmt tr concr (clConjs cs) = 
  (\<forall> c \<in> set cs. interp_clstmt tr concr c)"
by (induct cs) auto



(* Non-free variables do not influence clock interpretation - the same as for expression interpretation *)

primrec
  fv_clexpr :: "('n,'a) clexpr \<Rightarrow> 'n list"

where
  "fv_clexpr (ClOf x) = [x]"
| "fv_clexpr (ClWhen a) = []"
| "fv_clexpr ClTop = []"
| "fv_clexpr (ClInter ce ce') = list_union (fv_clexpr ce) (fv_clexpr ce')"
| "fv_clexpr (ClUnion ce ce') = list_union (fv_clexpr ce) (fv_clexpr ce')"
| "fv_clexpr (ClCompl ce) = (fv_clexpr ce)"

primrec
  fv_clstmt :: "('n,'a) clstmt \<Rightarrow> 'n list"
where
  "fv_clstmt ClTrue = []"
| "fv_clstmt (ClEq ce ce') = list_union (fv_clexpr ce) (fv_clexpr ce')"
| "fv_clstmt (ClConj c c') = list_union (fv_clstmt c) (fv_clstmt c')"
| "fv_clstmt (ClLet v e c) = list_union (fv_clexpr e) (list_diff (fv_clstmt c) [v])"


lemma fv_clstmt_clConjs [simp]:
  "set (fv_clstmt (clConjs cs)) =  (\<Union>c \<in> set cs. set (fv_clstmt c))"
by (induct cs) auto

end
