theory Clocksolver imports Clockcalc 
  "../Setsolvers/Gencode"
(* before, only:
  "../Setsolvers/Affset_constr"
*)

begin

datatype 'n istrue = IsTrue 'n

primrec name_of_istrue :: "'n istrue \<Rightarrow> 'n" where
 "name_of_istrue (IsTrue n) = n"

primrec
  extract_set_rep :: "('n \<Rightarrow> 'a option) \<Rightarrow> ('n, 'n istrue) clexpr \<Rightarrow> ('n, 'a) set_rep"
where
   "extract_set_rep env (ClOf n) = (Atom (the (env n)))"
|  "extract_set_rep env (ClWhen a) = (Var (name_of_istrue a))"
|  "extract_set_rep env ClTop = Bconst Top"
|  "extract_set_rep env (ClInter e e') = (Binop Inf (extract_set_rep env e) (extract_set_rep env e'))"
|  "extract_set_rep env (ClUnion e e') = (Binop Sup (extract_set_rep env e) (extract_set_rep env e'))"
|  "extract_set_rep env (ClCompl e) = (Comp (extract_set_rep env e))"


primrec
  extract_compos_constr :: "('n \<Rightarrow> 'a option) \<Rightarrow> ('n, 'n istrue) clstmt \<Rightarrow> ('n, ('n, 'a) set_rep) compos_constr "
where
  "extract_compos_constr env ClTrue = CnTrue"
| "extract_compos_constr env (ClEq e e') = CnEq (extract_set_rep env e) (extract_set_rep env e')"
| "extract_compos_constr env (ClConj c c') = CnConj (extract_compos_constr env c) (extract_compos_constr env c')"
| "extract_compos_constr env (ClLet v e c) = (CnLet v (extract_set_rep env e) (extract_compos_constr env c))"



lemma dom_trace_proj_upd: 
  "(\<lambda>na. dom (if n = na then s else trace_proj tr na)) = (dom \<circ> (trace_proj tr))(n := dom s)"
by simp

lemma characteristic_stream: "\<exists>s. (dom s = x)"
by (rule_tac x="\<lambda> v. (if v \<in> x then Some arbitrary else None)" in exI)
   (simp add:  dom_def fun_eq_iff)



  (* ---------- Instantiation to Affset ---------- *)

(*
lemma interp_clexpr_interp_setexpr:
  "interp_set_rep interp_affset
                  (dom \<circ> (trace_proj tr)) (extract_set_rep clexpr) =
  interp_clexpr tr (\<lambda>tr. interp_affset) clexpr"
by (induct clexpr) simp_all

lemma compose_trace_upd: "(f \<circ> trace_proj (trace_upd tr nm s)) = ((f \<circ> trace_proj tr)(nm := f s))"
by (simp add: fun_upd_def)


lemma interp_clstmt_interp_compos_constr [rule_format]: 
  "concr = (\<lambda> tr. interp_affset) \<Longrightarrow>
   \<forall> tr. 
   interp_clstmt tr concr clstmt =
   interp_compos_constr (dom \<circ> (trace_proj tr)) (interp_set_rep interp_affset) (extract_compos_constr clstmt)"
apply (induct clstmt)
apply simp 			-- "ClTrue"
apply (clarsimp simp add: interp_clexpr_interp_setexpr)  -- "ClEq"
apply simp			-- "ClConj"
   -- "ClLet"
apply (clarsimp simp add: interp_clexpr_interp_setexpr)
apply (clarsimp simp add: compose_trace_upd characteristic_stream)
done


lemma dom_indifferent_interp_affset [simp]:
  "dom_indifferent (\<lambda> tr. interp_affset)"
by (clarsimp simp add: dom_indifferent_def fun_eq_iff)


lemma clock_type_soundness_solve:
  "\<lbrakk>concr = (\<lambda> tr. interp_affset);
  wtpd_canon_proc_list ienv procs; 
  wf_clock_env_proc_list concr abstr procs clenv;
  wtpd_canon_stmt ienv tenv c;
  cpenv = (constr_canon_proc_env procs);
  solve (extract_compos_constr (clock_canon_stmt abstr clenv c));
  tpi = (\<lambda> tr' clk. interp_clexpr tr' concr (get_inter_typ_clocktp clk))
  \<rbrakk>
  \<Longrightarrow>   interp_canon_stmt Blocking tr cpenv tpi c
    = interp_canon_stmt Permissive tr cpenv tpi c"
apply (rule clock_type_soundness)
apply assumption+
apply (fastforce simp add: interp_clstmt_interp_compos_constr valid_solve valid_compos_constr_def)+
done


lemma valid_compos_constr_interp_clstmt:  
  "\<lbrakk>concr = (\<lambda>tr. interp_affset); 
     valid_compos_constr (interp_set_rep interp_affset) (extract_compos_constr ccs)\<rbrakk>
    \<Longrightarrow> interp_clstmt tr concr ccs"
by (fastforce simp add: interp_clstmt_interp_compos_constr valid_compos_constr_def)

lemma solve_interp_clstmt: "concr = (\<lambda>tr. interp_affset) \<Longrightarrow> 
  solve (extract_compos_constr clock_cstmt) \<Longrightarrow> 
  interp_clstmt tr concr clock_cstmt"
by (fastforce simp add: valid_solve intro: valid_compos_constr_interp_clstmt)
*)

term "\<lambda> parexprs clenv bd. 
(extract_compos_constr empty
  (clForall_closure parexprs (clock_canon_stmt IsTrue clenv bd)))
"

primrec
  wf_clspeceq_solve :: 
  "[('n, affset_tp) clock_env, 'n list, 
    ('n,'v, (('n, affset_tp) clexpr) inter_typ) canon_stmt, ('n, affset_tp) clspec] 
  \<Rightarrow> bool" where
  "wf_clspeceq_solve clenv pars bd (ClSpecEq parexprs) = 
  (pars = map fst parexprs \<and> 
   (list_all (\<lambda> e. list_subset (fv_clexpr e) pars) (map snd parexprs)) \<and> 
   (solve (extract_compos_constr empty
             (clForall_closure parexprs (clock_canon_stmt IsTrue clenv bd)))))"

(*
lemma wf_clspeceq_solve_impl_wf_clspeceq:
  "concr = (\<lambda>tr. interp_affset) \<Longrightarrow> 
  wf_clspeceq_solve abstr clenv pars bd (ClSpecEq parexprs) \<Longrightarrow>
  wf_clspeceq concr abstr clenv pars bd (ClSpecEq parexprs)"
apply (simp add: wf_clspeceq_solve_def wf_clspeceq_def list_all_iff)
apply (fastforce intro: solve_interp_clstmt)
done
*)

definition constr_clspeceq_from_var_types_list ::
    "('n, ('n, 'a) clexpr inter_typ) var_types list \<Rightarrow> ('n, 'a) clspec"
where
 "constr_clspeceq_from_var_types_list vts = 
  ClSpecEq (map (\<lambda> vt. (get_var_types_name vt, (get_inter_typ_clocktp (get_var_types_type vt))))  vts)"

definition constr_clock_env :: "('n,'v, ('n, 'a) clexpr inter_typ) canon_proc list \<Rightarrow> ('n,'a) clock_env"
where
  "constr_clock_env procs = 
  map_of 
  (map (\<lambda> pr. (get_can_proc_header pr, 
              (constr_clspeceq_from_var_types_list (get_can_proc_var_types pr))))
        procs)"


definition
  wf_clock_env_solve :: 
  "[('n,'v, (('n, affset_tp) clexpr) inter_typ) canon_proc list, ('n, affset_tp) clock_env]
  \<Rightarrow> bool"
where "wf_clock_env_solve procs clenv =
  (list_all (\<lambda> cp. 
     (case cp of (CanProc ph tpars bd) \<Rightarrow> 
     (case clenv ph of 
        None \<Rightarrow> False
      | Some clsp \<Rightarrow> 
        (case clsp of 
           ClSpec params c \<Rightarrow> False
         | ClSpecEq parexprs \<Rightarrow> 
                    wf_clspeceq_solve clenv (var_types_names tpars) bd (ClSpecEq parexprs))))) 
 procs)"

lemma wtpd_canon_proc_env_list_equivs_aux2 [rule_format]:
  "distinct (map get_can_proc_header procs) \<longrightarrow> 
   ( ((map_of (map (\<lambda>pr. (get_can_proc_header pr, pr)) procs) ph) = Some x) \<longrightarrow> (x \<in> set procs))"
apply (induct procs)
apply simp
apply clarify
apply simp
apply (case_tac "ph = get_can_proc_header a")
apply simp+
done

lemma wtpd_canon_proc_env_list_equivs_aux3 [rule_format]:
  "distinct (map get_can_proc_header procs) \<Longrightarrow>
   ( ((map_of (map (\<lambda>pr. (get_can_proc_header pr, pr)) procs) (get_can_proc_header x)) = Some x)
   = (x \<in> set procs))"
apply (rule iffI)
apply (fast intro: wtpd_canon_proc_env_list_equivs_aux2)
apply (fast intro: wtpd_canon_proc_env_list_equivs_aux)
done

lemma wf_clock_env_solve_impl_wf_clock_env: 
  "concr = (\<lambda>tr. interp_affset) \<Longrightarrow> 
  distinct (map get_can_proc_header procs) \<Longrightarrow>
  wf_clock_env_solve abstr procs clenv \<Longrightarrow>
  wf_clock_env concr abstr (constr_canon_proc_env procs) clenv"
apply (clarsimp simp add: wf_clock_env_solve_def wf_clock_env_def list_all_iff)
apply (frule wtpd_canon_proc_env_list_equivs_aux2, assumption)
apply (drule bspec, assumption)

apply (clarsimp 
  simp del: wf_clspeceq_solve.simps reduce_to_clspec.simps 
  split add: option.split_asm clspec.split_asm)

apply (fast intro: wf_clspeceq_impl_wf_clspec wf_clspeceq_solve_impl_wf_clspeceq
dest: wf_clock_env_solve_impl_wf_clock_env_aux
)
done


lemma wf_clock_env_solve_impl_wf_clock_env_proc_list: 
  "concr = (\<lambda>tr. interp_affset) \<Longrightarrow> 
  distinct (map get_can_proc_header procs) \<Longrightarrow>
  wf_clock_env_solve abstr procs clenv \<Longrightarrow>
  wf_clock_env_proc_list concr abstr procs clenv"
apply (clarsimp simp add: wf_clock_env_solve_def wf_clock_env_proc_list_def list_all_iff)
apply (drule bspec, assumption)

apply (clarsimp 
  simp del: wf_clspeceq_solve.simps reduce_to_clspec.simps 
  split add: canon_proc.split_asm  option.split_asm clspec.split_asm)
apply (fast intro: wf_clspeceq_impl_wf_clspec wf_clspeceq_solve_impl_wf_clspeceq)
done


lemma clock_type_soundness_interp_compos_constr [rule_format]:
  "\<lbrakk>concr = (\<lambda> tr. interp_affset);
  distinct (map get_can_proc_header procs);
  wtpd_canon_proc_list ienv procs;
  wf_clock_env_solve abstr procs clenv;
  wtpd_canon_stmt ienv tenv c;
  solve (extract_compos_constr (clock_canon_stmt abstr clenv c));
  tpi = (\<lambda> tr' clk. interp_clexpr tr' concr (get_inter_typ_clocktp clk))
  \<rbrakk>
  \<Longrightarrow>   interp_canon_stmt Blocking tr (constr_canon_proc_env procs) tpi c
    = interp_canon_stmt Permissive tr (constr_canon_proc_env procs) tpi c"
apply (rule clock_type_soundness_solve)
apply assumption+
apply (fast intro: wf_clock_env_solve_impl_wf_clock_env_proc_list)
apply simp_all
done

  (* TODO: what was 'clenv' above must not be 'empty' here, but rather the clockspec 
  gathered from the interfaces.
 *)
primrec wtpd_canon_program :: "('n \<Rightarrow> affset_tp) \<Rightarrow> ('n,vals,('n, affset_tp) clexpr inter_typ) canon_program \<Rightarrow> bool"
where
  "wtpd_canon_program abstr (CanProgram procs c) = 
  (distinct (map get_can_proc_header procs) \<and>
   wtpd_canon_proc_list procs \<and>
   wf_clock_env_solve abstr procs (constr_clock_env procs) \<and>
   wtpd_canon_stmt (constr_canon_proc_env procs) empty c \<and>
   solve (extract_compos_constr (clock_canon_stmt abstr (constr_clock_env procs) c)))"


theorem clock_type_soundness_program: 
  "wtpd_canon_program abstr prg \<Longrightarrow> 
  tpi = (\<lambda> tr' clk. interp_clexpr tr' (\<lambda> tr. interp_affset) (get_inter_typ_clocktp clk)) \<Longrightarrow>
  interp_canon_program Blocking tr tpi prg = 
  interp_canon_program Permissive tr tpi prg"
apply (case_tac prg)
apply (simp  only: interp_canon_program.simps wtpd_canon_program.simps)
apply (rule clock_type_soundness_interp_compos_constr)
apply (rule refl)
apply fast+
done


end

