theory List_extended imports Main

begin

text {* Additional functions for lists, in particular in view of code generation. 
  Still unsatisfactory since equality of list members is not correctly taken into account.
*}


definition "list_insert x xs = (if (List.member xs x) then xs else x # xs)"

lemma set_list_insert [simp]: "set (list_insert x xs) = insert x (set xs)"
apply (induct xs)
apply (clarsimp simp add: list_insert_def List.member_def)+ 
apply blast
done


definition "list_subset xs ys = List.list_all (List.member ys) xs"

lemma set_list_subset [simp]: "(list_subset xs ys) = (set xs \<subseteq> set ys)"
by (simp add: list_subset_def list_all_iff member_def) fastforce

definition "list_equal xs ys = (list_subset xs ys \<and> list_subset ys xs)"

lemma set_list_equal [simp]: "(list_equal xs ys) = (set xs = set ys)"
by (fastforce simp add: list_equal_def)

definition "list_inter xs ys = (List.filter (List.member xs) ys)"

lemma set_list_inter [simp]: "set (list_inter xs ys) = (set xs) \<inter> (set ys)"
by (simp add: list_inter_def List.member_def) blast

definition "list_union xs ys = (List.foldr list_insert xs ys)"

lemma set_list_union [simp]: "set (list_union xs ys) = (set xs) \<union> (set ys)"
by (induct xs) (simp add: list_union_def )+

definition "list_diff xs ys = (List.filter (\<lambda> x. \<not> (List.member ys x)) xs)"

lemma set_list_diff [simp]: "set (list_diff xs ys) = (set xs) - (set ys)"
by (simp add: list_diff_def List.member_def) blast


definition  list_UNION :: "'a list \<Rightarrow> ('a \<Rightarrow> 'b list) \<Rightarrow> 'b list" where
  "list_UNION A B = (List.foldr (list_union \<circ> B) A [])"

lemma set_list_UNION [simp]: "set (list_UNION A B) = (\<Union>a \<in> set A. set (B a))"
by (induct A) (simp add: list_UNION_def)+

end

