theory Stretching imports Streams

begin


subsection {* Stretching *}

text {* 

Stretching (also -- maybe inappropriately -- called ``densification'' in the
literature) allows to stretch the temporal domain by a stretching
function. A stretching function maps an instant of time $t$ to a
larger instant $f t$. 

These definitions use Isabelle's axiomatic typeclasses -- the type
@{text "'t"} is constrained to be a preorder.

*}



subsection {* Stretching functions *}


(*
locale increasing =
  fixes f
  assumes increasing: "x \<le> (f x)"
*)

definition
  increasing :: "['t::order \<Rightarrow> 't] \<Rightarrow> bool" where
  "increasing f == \<forall> x. x \<le> (f x)"

lemma increasing_comp:
  "\<lbrakk>increasing f1; increasing f2\<rbrakk> \<Longrightarrow> increasing (f1 \<circ> f2)"
apply (clarsimp simp add: increasing_def)
apply (rule_tac y="f2 x" in  order_trans) apply fast+
done

definition
  stretch_fun :: "['t::order \<Rightarrow> 't] \<Rightarrow> bool" where
  "stretch_fun f == (inj f \<and> mono f \<and> increasing f)"


lemma mono_id [simp]: "mono id"
by (simp add: mono_def)

lemma mono_comp [simp]: "\<lbrakk>mono f1; mono f2\<rbrakk> \<Longrightarrow> mono (f1 \<circ> f2)"
by (simp add: mono_def)

lemma stretch_fun_id [simp]: "stretch_fun id"
by (simp add: stretch_fun_def increasing_def)

lemma stretch_fun_comp: 
  "\<lbrakk> stretch_fun f1; stretch_fun f2\<rbrakk> \<Longrightarrow> stretch_fun (f1 \<circ> f2)"
apply (clarsimp simp add: stretch_fun_def)
apply (rule conjI)
apply (rule comp_inj_on) apply simp+
apply (rule subset_inj_on) apply assumption apply simp
apply (simp add: increasing_comp)
done

definition
  -- "tr can be stretched to tr' by function f"
  stretch_fun_leq :: 
  "['n set, ('t::order,'n,'v) trace, ('t,'n,'v) trace, 't \<Rightarrow> 't] \<Rightarrow> bool"
where
  "stretch_fun_leq A tr tr' f == 
       stretch_fun f \<and> 
       (\<forall> x \<in> A. (trace_proj tr x) = (trace_proj tr' x) \<circ> f) \<and>
       (\<forall> x \<in> A. (dom (trace_proj tr' x) = f ` dom (trace_proj tr x)))"

lemma stretch_fun_leq_refl [simp]: "stretch_fun_leq A tr tr id"
by (clarsimp simp add: stretch_fun_leq_def)

lemma stretch_fun_leq_trans:
  "\<lbrakk> stretch_fun_leq A tr tr' f1; stretch_fun_leq A tr' tr'' f2 \<rbrakk> \<Longrightarrow>
  stretch_fun_leq A tr tr'' (f2 \<circ> f1)"
by (clarsimp simp add: stretch_fun_leq_def stretch_fun_comp image_compose)


lemma stretch_fun_leq_funs_eq_on:
  "(stretch_fun_leq A tr tr' id) = (\<forall> t. funs_eq_on (tr t) (tr' t) A)"
by (fastforce 
  simp add: fun_eq_iff stretch_fun_leq_def funs_eq_on_def trace_proj_def)

lemma stretch_fun_leq_empty [simp]: 
  "stretch_fun_leq {} tr tr sf = stretch_fun sf"
by (clarsimp simp add: stretch_fun_leq_def)

lemma stretch_fun_leq_union: 
  "stretch_fun_leq (A \<union> B) tr tr' f = 
  (stretch_fun_leq A tr tr' f \<and> stretch_fun_leq B tr tr' f)"
by (fastforce simp add: stretch_fun_leq_def)

lemma stretch_fun_leq_insert: 
  "stretch_fun_leq (insert x A) tr tr' f = 
  (stretch_fun_leq {x} tr tr' f \<and> stretch_fun_leq A tr tr' f)"
by (simp add: stretch_fun_leq_union [THEN sym])


lemma stretch_fun_leq_subseteq:
 "\<lbrakk> stretch_fun_leq A tr trext sf; B \<subseteq> A \<rbrakk> \<Longrightarrow> stretch_fun_leq B tr trext sf"
by (fastforce simp add: stretch_fun_leq_def)

lemma stretch_fun_leq_id_sym: 
  "stretch_fun_leq A tr' tr id \<Longrightarrow> stretch_fun_leq A tr tr' id"
by (simp (no_asm_use) add: stretch_fun_leq_def) fastforce

lemma stretch_fun_leq_override_on_split:
  "(stretch_fun_leq A tr (\<lambda>t. override_on (tr' t) (tr'' t) B) sf)
  =  ((stretch_fun_leq (A \<inter> B) tr tr'' sf) \<and> 
     (stretch_fun_leq (A - B) tr tr' sf))"
apply (simp add: stretch_fun_leq_def trace_proj_def)
apply (rule iffI)
apply (clarsimp)

apply (intro conjI)
apply clarsimp
apply (drule bspec, assumption)+
apply (simp add: image_dom)
apply clarsimp
apply (drule bspec, assumption)+
apply (simp add: image_dom)

apply clarsimp
apply (intro conjI)
apply clarsimp
apply (case_tac "x \<in> B")
apply (simp add: image_dom)
apply clarsimp+
apply (case_tac "x \<in> B")

apply (simp add: image_dom)
apply clarsimp+
done

-- "just an instance of the above, but better proved separately"
lemma stretch_fun_leq_override_on:
  "A \<subseteq> B \<Longrightarrow> 
  (stretch_fun_leq A tr (\<lambda>t. override_on (tr' t) (tr'' t) B) sf)
=  (stretch_fun_leq A tr tr'' sf)"
by (fastforce simp add: stretch_fun_leq_def override_on_def trace_proj_def)


  -- "the trace obtained by stretching tr with stretching function sf"
definition
  stretched_trace :: "[('t, 'n, 'v) trace, ('t \<Rightarrow> 't)] \<Rightarrow> ('t, 'n, 'v) trace"
where  "stretched_trace tr sf == 
    (\<lambda> t x. (if t \<in> range sf then (tr (inv sf t) x) else None))"

lemma trace_proj_stretched_trace:
  "(trace_proj (stretched_trace tr sf) x) = 
  (override_on empty (trace_proj tr x \<circ> inv sf) (range sf))"
by (simp add: stretched_trace_def override_on_def trace_proj_def)

lemma stretch_fun_leq_stretched_trace:
  "stretch_fun sf \<Longrightarrow>
  stretch_fun_leq A tr (stretched_trace tr sf) sf"
apply (simp add: stretch_fun_leq_def)
apply (clarsimp simp add: stretch_fun_def)
apply (simp add: image_dom trace_proj_stretched_trace)
done


definition
  stretch_signal_leq :: 
  "[('t::order,'v) signal, ('t,'v) signal, 't \<Rightarrow> 't] \<Rightarrow> bool"
where
  "stretch_signal_leq s s' f == 
       stretch_fun f \<and> (s = s' \<circ> f) \<and> (dom s' = f ` dom s)"

  -- "the signal obtained by stretching s with stretching function sf"
definition
  stretched_signal :: "[('t, 'v) signal, ('t \<Rightarrow> 't)] \<Rightarrow> ('t, 'v) signal"
where  "stretched_signal s sf == 
    (\<lambda> t. (if t \<in> range sf then s (inv sf t) else None))"


lemma stretched_signal_override:
  "(stretched_signal s sf) = (override_on empty (s \<circ> inv sf) (range sf))"
by (simp add: stretched_signal_def override_on_def)



definition
  stretch_leq :: "['n set, ('t::order,'n,'v) trace, ('t,'n,'v) trace] \<Rightarrow> bool"
where  "stretch_leq A tr tr' == (\<exists> f. stretch_fun_leq A tr tr' f)"

definition  stretch_eq :: "['n set, ('t::order,'n,'v) trace, ('t,'n,'v) trace] \<Rightarrow> bool"
where  "stretch_eq A tr tr' == (stretch_leq A tr tr' \<and> stretch_leq A tr' tr)"

lemma stretch_leq_refl: "stretch_leq A tr tr"
apply (simp add: stretch_leq_def)
apply (rule_tac x=id in exI)
apply simp
done

lemma stretch_leq_trans: "\<lbrakk>stretch_leq A tr tr'; stretch_leq A tr' tr'' \<rbrakk> 
  \<Longrightarrow> stretch_leq A tr tr''"
apply (clarsimp simp add: stretch_leq_def)
apply (rename_tac f1 f2)
apply (rule_tac x="(f2 \<circ> f1)" in exI)
apply (simp add: stretch_fun_leq_trans)
done

definition
  undef_slice :: "['n set, ('t, 'n, 'v) trace, 't] \<Rightarrow> bool"
where  "undef_slice A tr n == (\<forall> x \<in> A. tr n x = None)"


(* The "inverse" of the following function is the stretching function
  required to minimize a trace.
  def_pos computes the number of defined elements up to the given position.
  *)
primrec
  def_pos :: "['n set, (nat,'n,'v) trace, nat] \<Rightarrow> nat"
where
  "def_pos A tr 0 = 0"|
  "def_pos A tr (Suc n) = (if (\<forall> k > n. undef_slice A tr k)
                         then (def_pos A tr n) + 1
                         else if (undef_slice A tr n)
                         then (def_pos A tr n)
                         else (def_pos A tr n) + 1)"

(* the stream containing no undefined values in between, 
  but only at the end *)
definition
  stretch_reduced :: "['n set, (nat, 'n, 'v) trace] \<Rightarrow> bool"
where  "stretch_reduced A tr == 
  \<forall> n. undef_slice A tr n \<longrightarrow> (\<forall> k \<ge> n. undef_slice A tr k)"



definition
  trace_rel :: "['n set, (nat, 'n, 'v) trace] 
  \<Rightarrow> (('n, 'v) event * ('n, 'v) event) set"
where  "trace_rel A tr == {(e,e') . 
  \<exists> i j. i\<le>j \<and> e=(tr i) \<and> e'=(tr j) \<and> 
         \<not> undef_slice A tr i \<and> \<not> undef_slice A tr j \<and>
         (\<forall> k. i < k \<longrightarrow> k < j \<longrightarrow> undef_slice A tr k)}"


subsection {* Equivalence of traces *}

text {* As a special case of stretching (with stretching function id) *}

definition
  traces_eq_on :: "[('t,'n,'v) trace, ('t,'n,'v) trace, 'n set] \<Rightarrow> bool" where
  "traces_eq_on tr tr' A == \<forall> t. funs_eq_on (tr t) (tr' t) A"

lemma traces_eq_on_refl [simp]: "traces_eq_on tr tr A"
by (simp add: traces_eq_on_def)

lemma traces_eq_on_sym: 
  "traces_eq_on tr' tr A \<Longrightarrow> traces_eq_on tr tr' A"
by (fastforce simp add: traces_eq_on_def funs_eq_on_def) 

lemma traces_eq_on_trans: 
  "\<lbrakk> traces_eq_on tr tr' A; traces_eq_on tr' tr'' A\<rbrakk>
  \<Longrightarrow> traces_eq_on tr tr'' A"
by (fastforce simp add: traces_eq_on_def funs_eq_on_def) 

lemma  traces_eq_on_UNIV [simp]:
  "traces_eq_on tr tr' UNIV = (tr = tr')"
by (auto simp add: traces_eq_on_def)

lemma traces_eq_on_empty [simp]:
  "traces_eq_on tr tr' {}"
by (simp add: traces_eq_on_def funs_eq_on_def)

lemma traces_eq_on_mono: 
  "\<lbrakk>traces_eq_on tr tr' A; B \<subseteq> A\<rbrakk> \<Longrightarrow> traces_eq_on tr tr' B"
by (simp add: traces_eq_on_def) (fast intro: funs_eq_on_mono)


lemma traces_eq_on_trace_proj: 
  "\<lbrakk> traces_eq_on tr tr' A; n \<in> A \<rbrakk> \<Longrightarrow> trace_proj tr n = trace_proj tr' n"
by (simp add: traces_eq_on_def funs_eq_on_def trace_proj_def)


lemma stretch_fun_leq_traces_eq:
 "stretch_fun_leq A tr tr' id = traces_eq_on tr tr' A"
by (simp add: stretch_fun_leq_funs_eq_on traces_eq_on_def)


lemma traces_eq_on_override_on_split:
  "(traces_eq_on tr (\<lambda>t. override_on (tr' t) (tr'' t) B) A)
  =  ((traces_eq_on tr tr'' (A \<inter> B)) \<and> 
     (traces_eq_on tr tr' (A - B)))"
by (fastforce simp add: traces_eq_on_def funs_eq_on_override_on_split)

lemma traces_eq_on_override_on1:
  "A \<subseteq> - B \<Longrightarrow> 
  (traces_eq_on tr (\<lambda> t. override_on (tr' t) (tr'' t) B) A) 
  = (traces_eq_on tr tr' A)"
by (fastforce simp add: traces_eq_on_def funs_eq_on_override_on1)

lemma traces_eq_on_override_on2:
  "A \<subseteq> B \<Longrightarrow> 
  (traces_eq_on tr (\<lambda>t. override_on (tr' t) (tr'' t) B) A)
  = (traces_eq_on tr tr'' A)"
by (fastforce simp add: traces_eq_on_def funs_eq_on_override_on2)

lemma traces_eq_on_union [simp]:
  "traces_eq_on tr tr' (A \<union> B) =
  (traces_eq_on tr tr' A \<and> traces_eq_on tr tr' B)"
by (fastforce simp add: traces_eq_on_def)


end
