theory Tp imports Aux Sstring 

begin


(* ---------------------------------------------------------------------- *)
subsection {* Types and their properties *}


datatype val_tp 
  = UnitT
  | BoolT
  | NatT
  | PairT val_tp val_tp




definition tp_pair :: "val_tp option \<Rightarrow> val_tp option \<Rightarrow> val_tp option" where
  "tp_pair t1 t2 \<equiv> 
  (case t1 of 
     None \<Rightarrow> None
   | Some t1' \<Rightarrow> 
    (case t2 of 
       None \<Rightarrow> None
     | Some t2' \<Rightarrow> Some (PairT t1' t2')))"

definition tp_compatible :: "val_tp option \<Rightarrow> val_tp option \<Rightarrow> bool" where
  "tp_compatible t1 t2 \<equiv> 
  (case t1 of 
     None \<Rightarrow> False
   | Some t1' \<Rightarrow> 
    (case t2 of 
       None \<Rightarrow> False
     | Some t2' \<Rightarrow> t1' = t2'))"

(* ---------------------------------------------------------------------- *)
subsection {* Values and their types *}

  (* constants *)
datatype cst
  = UnitC
  | BoolC bool
  | NatC nat

  (* Specific names and identifiers *)
datatype specialf 
  = AffsetF nat nat

datatype calcf 
  = PlusF 			(* Plus on nat *)
  
  (* Function identifiers *)
datatype funid 
  = ConstF cst
  | SpecialF specialf
  | CalcF calcf



definition tp_fun :: "funid \<Rightarrow> val_tp option \<Rightarrow> val_tp option" where
  "tp_fun fid tp \<equiv> (case fid of CalcF PlusF \<Rightarrow> Some NatT
                               |_ \<Rightarrow> None)"
  



  (* the definitions (primrec) of these functions can be found
  elsewhere (currently: test.thy) *)
consts interp_cst :: "cst \<Rightarrow> 'v \<Rightarrow> 'v"
consts interp_specialf :: "specialf \<Rightarrow> 'v \<Rightarrow> 'v"

consts interp_calcf :: "calcf \<Rightarrow> 'v \<Rightarrow> 'v"

consts interp_funid :: "funid \<Rightarrow> 'v \<Rightarrow> 'v"

(* has now been parameterized
types proc_interfname = sstring
types proc_implname = sstring

datatype proc_header = ProcHeader proc_interfname proc_implname -- "procedure header"
*)

datatype 'n proc_header = ProcHeader 'n 'n -- "procedure header"

fun get_proc_interf_name :: "'n proc_header \<Rightarrow> 'n"
where "get_proc_interf_name (ProcHeader a b) =  a"

fun get_proc_impl_name :: "'n proc_header \<Rightarrow> 'n"
where "get_proc_impl_name (ProcHeader a b) =  b"



text {* Procedure interfaces with procedure name, parameters *}
datatype ('n, 'pn) proc_interf
  = ProcInterf "'pn" "('n list)" 



text {* In a procedure call, the procedure name has to be tagged with an id
 (yielding the "call name") that is unique within the procedure body, so that 
  different calls to the same procedure can be distinguished. This tagging 
  can be done during preprocessing by the parser. 
  This is necessary for building up a data structure with different instances 
  for different function call invocations.
*}

datatype 'n cname = Cname "('n proc_header)" 'n -- "call name"

primrec
  cname_header :: "'n cname \<Rightarrow> 'n proc_header"
where
  "cname_header (Cname pn n) = pn"


end

