theory Values imports Tp

begin

  (* type + *)
class vals_cl =
  fixes true_val :: "'a \<Rightarrow> bool"
  fixes pair_val :: "'a * 'a \<Rightarrow> 'a"
  fixes proj1_val :: "'a \<Rightarrow> 'a"
  fixes proj2_val :: "'a \<Rightarrow> 'a"
  assumes proj1_pair_val [simp]: "\<forall> v. proj1_val (pair_val v) = (fst v)"
  assumes proj2_pair_val [simp]: "\<forall> v. proj2_val (pair_val v) = (snd v)"


lemma projs_pair_val:
  "x = proj1_val (pair_val(x, y)) \<and> y = proj2_val (pair_val(x, y))"
by simp


datatype vals 
  = UnitV
  | BoolV bool
  | NatV nat
  | PairV vals vals


primrec
  vals_tp :: "vals \<Rightarrow> val_tp"
where
  "vals_tp UnitV = UnitT" |
  "vals_tp (BoolV b) = BoolT"|
  "vals_tp (NatV n) = NatT"|
  "vals_tp (PairV v v') = PairT (vals_tp v) (vals_tp v')"

primrec
  proj1_val_pr :: "vals \<Rightarrow> vals"
where
  "proj1_val_pr (PairV v v') = v "

primrec 
  proj2_val_pr :: "vals \<Rightarrow> vals"
where
  "proj2_val_pr (PairV v v') = v' "


instantiation vals :: vals_cl
begin

(*
constdefs
  true_val :: "vals \<Rightarrow> bool"
*)

definition
  true_val_def:  "true_val v == (v = BoolV True)"

(*
constdefs 
  pair_val :: "vals * vals \<Rightarrow> vals"
*)
definition
  pair_val_def: "pair_val == \<lambda> (v, v'). PairV v v'"

definition
  proj1_val_def: "proj1_val = proj1_val_pr"
definition
  proj2_val_def: "proj2_val = proj2_val_pr"

  instance
  by (intro_classes) (simp add: pair_val_def proj1_val_def  proj2_val_def)+
end

(*
lemma proj1_pair_val: "proj1_val (pair_val v) = (fst v)"
by (case_tac v) (simp add: pair_val_def)

lemma proj2_pair_val: "proj2_val (pair_val v) = (snd v)"
by (case_tac v) (simp add: pair_val_def)

lemma projs_pair_val:
  "x = proj1_val (pair_val(x, y)) \<and> y = proj2_val (pair_val(x, y))"
by (simp add: proj1_pair_val proj2_pair_val)
*)

end
